/*
 * USB Driver for ALi m5602 based webcams
 *
 * Copyright (C) 2008 Erik Andrén
 * Copyright (C) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (C) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver sofwtare,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#include <linux/vmalloc.h>
#include <linux/usb.h>
#include <media/v4l2-dev.h>

#include "m5602_ov9650.c"
#include "m5602_s5k83a.c"
#include "m5602_s5k4aa.c"
#include "m5602_mt9m111.c"
#include "m5602_unknown_0xdc.c"

#include "m5602_v4l2.c"

static struct usb_device_id m5602_table[] = {
	{USB_DEVICE(0x0402, 0x5602)},
	{}
};

static char *m5602_name[2] = {
	"BisonCam",
	NULL
};

MODULE_DEVICE_TABLE(usb, m5602_table);

/*static void m5602_create_sysfs_files(struct video_device *vdev);
static void m5602_remove_sysfs_files(struct video_device *vdev);*/

/* Reads a byte from the m5602 */
static int m5602_read_bridge(struct m5602_camera *cam, u8 address, u8 *i2c_data)
{
	int err;
	err = usb_control_msg(cam->udev, usb_rcvctrlpipe(cam->udev, 0),
			      0x04, 0xc0, 0x14,
			      0x8100 + address, cam->control_buffer,
			      1, M5602_URB_MSG_TIMEOUT);
	*i2c_data = cam->control_buffer[0];

	PDEBUG(DBG_TRACE, "Reading bridge register 0x%x containing 0x%x",
	       address, *i2c_data);

	/* usb_control_msg(...) returns the number of bytes sent upon success,
	mask that and return zero upon success instead*/
	return (err < 0) ? err : 0;
}

/* Writes a byte to to the m5602 */
static int m5602_write_bridge(struct m5602_camera *cam, u8 address, u8 i2c_data)
{
	int err;
	PDEBUG(DBG_TRACE, "Writing bridge register 0x%x with 0x%x",
	       address, i2c_data);

	memcpy(cam->control_buffer, bridge_urb_skeleton,
	       sizeof(bridge_urb_skeleton));
	cam->control_buffer[1] = address;
	cam->control_buffer[3] = i2c_data;

	err = usb_control_msg(cam->udev, usb_sndctrlpipe(cam->udev, 0),
				0x04, 0x40, 0x19,
				0x0000, cam->control_buffer,
				4, M5602_URB_MSG_TIMEOUT);

	/* usb_control_msg(...) returns the number of bytes sent upon success,
	   mask that and return zero upon success instead*/
	return (err < 0) ? err : 0;
}

/* Dump all the registers of the m5602 bridge,
   unfortunately this breaks the camera until it's power cycled */
static void m5602_dump_bridge(struct m5602_camera *cam)
{
	int i;
	for (i = 0; i < 0x80; i++) {
		unsigned char val = 0;
		m5602_read_bridge(cam, i, &val);
		info("ALi m5602 address 0x%x contains 0x%x", i, val);
	}
	info("Warning: The camera probably won't work until it's power cycled");
}

static int m5602_initialize_camera(struct m5602_camera *cam)
{
	PDEBUG(DBG_INIT, "m5602_initialize_camera");

	if (!(cam->state & DEV_INITIALIZED)) {
		init_waitqueue_head(&cam->open);

		mutex_init(&cam->dev_mutex);
		mutex_init(&cam->fileop_mutex);

		spin_lock_init(&cam->queue_lock);

		init_waitqueue_head(&cam->wait_frame);
		init_waitqueue_head(&cam->wait_stream);

		cam->state |= DEV_INITIALIZED;
	}

	if (dump_bridge)
		m5602_dump_bridge(cam);

	/* Now we should be able to probe for the sensor */
	if (m5602_probe_sensor(cam))
		goto init_fail;

	/* Ok, sensor is probed and attached, now init it */
	if (cam->sensor->init(cam))
		goto init_fail;

	return 0;

init_fail:
	info("Initialization of the ALi m5602 webcam failed");
	return -ENODEV;
}

static void m5602_release_resources(struct m5602_camera *cam)
{
	/* down(&m5602_sysfs_lock); */

	info("Video device %d releases its resources", cam->vdev->minor);
	video_set_drvdata(cam->vdev, NULL);
	video_unregister_device(cam->vdev);
	kfree(cam->control_buffer);

	/* m5602_remove_sysfs_files(cam->vdev); */
	/* up(&m5602_sysfs_lock); */
}

struct file_operations v4l_m5602_fops = {
	.owner	  	= THIS_MODULE,
	.open	  	= m5602_v4l_open,
	.release  	= m5602_v4l_release,
	.read	  	= NULL,
	.poll	  	= NULL,
	.mmap	  	= m5602_v4l_mmap,
	.ioctl	  	= m5602_v4l_ioctl,
	.compat_ioctl 	= v4l_compat_ioctl32,
	.llseek   	= no_llseek
};

static int m5602_probe_sensor(struct m5602_camera *cam)
{
	/* Try the mt9m111 sensor */
	cam->sensor = &mt9m111;
	if (!cam->sensor->probe(cam))
		return 0;

	/* Try the s5k4aa */
	cam->sensor = &s5k4aa;
	if (!cam->sensor->probe(cam))
		return 0;

	/* Try the ov9650 */
	cam->sensor = &ov9650;
	if (!cam->sensor->probe(cam))
		return 0;

	/* Try the s5k83a */
	cam->sensor = &s5k83a;
	if (!cam->sensor->probe(cam))
		return 0;

	/* Try a unidentified sensor */
	cam->sensor = &unknown_0xdc;
	if (!cam->sensor->probe(cam))
		return 0;

	/* More sensor probe function goes here */
	info("Failed to find a sensor");
	cam->sensor = NULL;
	return -ENODEV;
}

/* We only have one interface in our webcam. */
static int m5602_usb_probe(struct usb_interface *intf,
			   const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(intf);
	struct m5602_camera *cam;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	char *camera_name = NULL;
	size_t buffer_size;
	int i, error = 0;

	for (i = 0; i < ARRAY_SIZE(m5602_table); i++) {
		if ((le16_to_cpu(udev->descriptor.idVendor)
				   == m5602_table[i].idVendor) &&
		    (le16_to_cpu(udev->descriptor.idProduct)
				   == m5602_table[i].idProduct)) {
			camera_name = m5602_name[i];
			info("%s webcam found", m5602_name[i]);
		}
	}

	if (!camera_name) {
		err("Vendor/Product ID doesn't match");
		return -ENODEV;
	}

	cam = kzalloc(sizeof(struct m5602_camera), GFP_KERNEL);
	if (!cam)
		return -ENOMEM;

	PDEBUG(DBG_TRACE, "cam 0x%p", cam);

	cam->udev = udev;
	memcpy(&cam->dev, &udev->dev, sizeof(struct device));

	cam->name = camera_name;
	cam->interface = intf;

	cam->control_buffer = kmalloc(256, GFP_KERNEL);

	if (!cam->control_buffer) {
		err("kmalloc() failed");
		error = -ENOMEM;
		goto fail;
	}

	PDEBUG(DBG_PROBE, "Num altsettings: %d", intf->num_altsetting);

	usb_set_interface(cam->udev, 0, 2);

	iface_desc = intf->cur_altsetting;

	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if (!cam->int_in_endpointAddr &&
			((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
				   == USB_DIR_IN) &&
			((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
				   == USB_ENDPOINT_XFER_INT)) {

			/* we found an int in endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			cam->int_in_size = buffer_size;
			cam->int_in_endpointAddr = endpoint->bEndpointAddress;
			PDEBUG(DBG_V4L2, "INTR Endpoint: 0x%x, "
					 "wMaxPacketSize: 0x%zx",
					 cam->int_in_endpointAddr,
					 cam->int_in_size);
		}

		if (!cam->isoc_in_endpointAddr &&
			((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
				   == USB_DIR_IN) &&
			((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
				   == USB_ENDPOINT_XFER_ISOC)) {

			/* we found an isoc in endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			cam->isoc_in_size = (buffer_size & 0x7ff) * 3;
			cam->isoc_in_endpointAddr = endpoint->bEndpointAddress;

			PDEBUG(DBG_V4L2, "ISOC Endpoint: 0x%x, "
					 "wMaxPacketSize: 0x%zx",
					 cam->isoc_in_endpointAddr,
					 cam->isoc_in_size);
		}
	}

	if (!(cam->int_in_endpointAddr && cam->isoc_in_endpointAddr)) {
		err("Could not find both int-in and isoc-in endpoints");
		error = -ENODEV;
		goto fail;
	}

	cam->vdev = video_device_alloc();
	if (!cam->vdev) {
		error = -ENOMEM;
		goto fail;
	}

	error = m5602_initialize_camera(cam);
	if (error)
		goto fail;

	strcpy(cam->vdev->name, "BisonCam");
	cam->vdev->owner    = THIS_MODULE;
	cam->vdev->type     = VID_TYPE_CAPTURE;
	cam->vdev->fops     = &v4l_m5602_fops;
	cam->vdev->release  = video_device_release;
	cam->vdev->minor    = -1;

	video_set_drvdata(cam->vdev, cam);

	mutex_lock(&cam->dev_mutex);

	error = video_register_device(cam->vdev, VFL_TYPE_GRABBER, -1);
	if (error) {
		err("video_register_device failed");
		mutex_unlock(&cam->dev_mutex);
		goto fail;
	}

	info("ALi m5602 webcam driver is now controlling video device %d",
	     cam->vdev->minor);

	/* m5602_create_sysfs_files(cam->vdev); */
	usb_set_intfdata(intf, cam);

	mutex_unlock(&cam->dev_mutex);
	return 0;

fail:
	if (cam) {
		kfree(cam->control_buffer);
		cam->control_buffer = NULL;
		if (cam->vdev)
			video_device_release(cam->vdev);

		kfree(cam);
	}
	return error;
}

static int m5602_suspend(struct usb_interface *intf, pm_message_t message)
{
	struct m5602_camera *cam = usb_get_intfdata(intf);
	int err = 0;

	PDEBUG(DBG_TRACE, "Suspending m5602 webcam");

	if (cam->stream == STREAM_ON) {
		PDEBUG(DBG_TRACE, "Halting stream");
		err = m5602_stop_transfer(cam);
	}

	return err;
}

static int m5602_resume(struct usb_interface *intf)
{
	struct m5602_camera *cam = usb_get_intfdata(intf);
	int err = 0;

	PDEBUG(DBG_TRACE, "Resuming m5602 webcam");

	if (cam->stream == STREAM_ON) {
		PDEBUG(DBG_TRACE, "Restarting stream");
		err = m5602_start_transfer(cam);
	}

	return err;
}

static void m5602_usb_disconnect(struct usb_interface *intf)
{
	struct m5602_camera *cam = usb_get_intfdata(intf);

	if (!cam)
		return;

	down_write(&m5602_disconnect);
	mutex_lock(&cam->dev_mutex);

	cam->state |= DEV_DISCONNECTED;

	if (cam->users) {
		PDEBUG(DBG_TRACE, "Device /dev/video%d is still open",
		       cam->vdev->minor);

		cam->state |= DEV_MISCONFIGURED;
		m5602_stop_transfer(cam);
		cam->state |= DEV_DISCONNECTED;

		wake_up_interruptible(&cam->wait_frame);
		wake_up_interruptible(&cam->wait_stream);
	}

	mutex_unlock(&cam->dev_mutex);

	wake_up_interruptible_all(&cam->open);

	PDEBUG(DBG_TRACE, "Waiting for all applications to release device /dev/video%d",
	       cam->vdev->minor);

	wait_event_interruptible_exclusive(cam->open, (cam->users == 0));

	if (!cam->users) {
		PDEBUG(DBG_TRACE, "All applications has released device /dev/video%d",
		       cam->vdev->minor);

		m5602_release_resources(cam);
		cam->state |= DEV_DISCONNECTED;

		PDEBUG(DBG_TRACE, "Freeing cam struct");
		kfree(cam);

		info("ALi m5602 webcam disconnected");
	}
	up_write(&m5602_disconnect);
}

static struct usb_driver m5602_usb_driver = {
	.name 	    = "m5602",
	.probe	    = m5602_usb_probe,
	.disconnect = m5602_usb_disconnect,
	.id_table   = m5602_table,
	.suspend    = m5602_suspend,
	.resume     = m5602_resume
};

static int __init usb_m5602_init(void)
{
	int result;

	info("ALi m5602 webcam driver startup");

	if (debug)
		info("debug level defined to 0x%x", debug);

	/* register this driver with the USB subsystem */
	result = usb_register(&m5602_usb_driver);

	if (result)
		err("usb_register failed with error %d", result);

	return result;
}

static void __exit usb_m5602_exit(void)
{
	info("ALi m5602 webcam driver shutdown");

	/* deregister this driver with the USB subsystem */
	usb_deregister(&m5602_usb_driver);
}

module_init(usb_m5602_init);
module_exit(usb_m5602_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
module_param(debug, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(debug, "toggles debug on/off");

module_param(force_sensor, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(force_sensor,
		 "force detection of sensor, "
		 "1 = OV9650, 2 = S5K83A, 3 = S5K4AA, 4 = MT9M111, 5 = UNKNOWN_0xDC");

module_param(dump_bridge, bool, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(dump_bridge, "Dumps all usb bridge registers at startup");

module_param(dump_sensor, bool, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(dump_sensor, "Dumps all usb sensor registers "
		"at startup providing a sensor is found");
