/*
 * Driver for the mt9m111 sensor
 *
 * Copyright (C) 2008 Erik Andrén
 * Copyright (C) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (C) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver sofwtare,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#include "m5602_mt9m111.h"

static int mt9m111_probe(struct m5602_camera *cam)
{
	u8 data[2] = {0x00, 0x00};
	int i;

	info("Probing for a mt9m111 sensor");

	/* Do the preinit */
	for (i = 0; i < ARRAY_SIZE(preinit_mt9m111); i++) {
		u8 data[2];

		if (preinit_mt9m111[i][0] == BRIDGE) {
			m5602_write_bridge(cam,
				preinit_mt9m111[i][1],
				preinit_mt9m111[i][2]);
		} else {
			data[0] = preinit_mt9m111[i][2];
			data[1] = preinit_mt9m111[i][3];
			mt9m111_write_sensor(cam,
				preinit_mt9m111[i][1], data, 2);
		}
	}

	if (mt9m111_read_sensor(cam, MT9M111_SC_CHIPVER, data, 2))
		return -ENODEV;

	if ((data[0] == 0x14) && (data[1] == 0x3a)) {
		info("Detected a mt9m111 sensor");
		return 0;
	}

	if (force_sensor == MT9M111_SENSOR) {
		info("Forcing a %s sensor", mt9m111.name);
		return 0;
	}

	return -ENODEV;
}

static int mt9m111_init(struct m5602_camera *cam)
{
	int i, err = 0;

	/* Init the sensor */
	for (i = 0; i < ARRAY_SIZE(init_mt9m111); i++) {
		u8 data[2];

		if (init_mt9m111[i][0] == BRIDGE) {
			err = m5602_write_bridge(cam,
				init_mt9m111[i][1],
				init_mt9m111[i][2]);
		} else {
			data[0] = init_mt9m111[i][2];
			data[1] = init_mt9m111[i][3];
			err = mt9m111_write_sensor(cam,
				init_mt9m111[i][1], data, 2);
		}
	}

	if (dump_sensor)
		mt9m111_dump_registers(cam);

	return (err < 0) ? err : 0;
}

static int mt9m111_get_ctrl(struct m5602_camera *cam, struct v4l2_control *ctrl)
{
	int tmp, err = 0;
	u8 data[2] = {0x00, 0x00};
	
	switch (ctrl->id) {
	case V4L2_CID_GAIN:
		mt9m111_read_sensor(cam, MT9M111_SC_GLOBAL_GAIN, data, 2);
		tmp = ((data[1] << 8) | data[0]);

		if (tmp >= 0)
			ctrl->value = ((tmp & (1 << 10)) * 2)
			    | ((tmp & (1 << 9)) * 2)
			    |  (tmp & 0x2f);
		else
			ctrl->value = tmp;

		PDEBUG(DBG_V4L2, "Read gain %d", ctrl->value);
		break;

	case V4L2_CID_VFLIP:
		mt9m111_read_sensor(cam, MT9M111_SC_R_MODE_CONTEXT_B, data, 2);
		ctrl->value = data[0] & MT9M111_RMB_MIRROR_ROWS;
		PDEBUG(DBG_V4L2, "Read vflip %d", ctrl->value);
		break;

	case V4L2_CID_HFLIP:
		mt9m111_read_sensor(cam, MT9M111_SC_R_MODE_CONTEXT_B, data, 2);
		ctrl->value = data[0] & MT9M111_RMB_MIRROR_COLS;
		PDEBUG(DBG_V4L2, "Read hflip %d", ctrl->value);
		break;

	default:
		PDEBUG(DBG_V4L2, "Unknown V4L2, id 0x%x", ctrl->value);
		return -EINVAL;
	}
	return (err < 0) ? err : 0;
}

static int mt9m111_set_ctrl(struct m5602_camera *cam,
		     const struct v4l2_control *ctrl)
{
	int err = 0;
	u8 data[2] = {0x00, 0x00};
	int tmp;

	switch (ctrl->id) {
	case V4L2_CID_VFLIP:
		PDEBUG(DBG_V4L2, "Setting vflip to %d", ctrl->value);

		mt9m111_read_sensor(cam, MT9M111_SC_R_MODE_CONTEXT_B, data, 2);
		data[0] = (data[0] & 0xfe) | ctrl->value;
		mt9m111_write_sensor(cam, MT9M111_SC_R_MODE_CONTEXT_B, data, 2);
		break;

	case V4L2_CID_HFLIP:
		PDEBUG(DBG_V4L2, "Setting hflip to %d", ctrl->value);

		mt9m111_read_sensor(cam, MT9M111_SC_R_MODE_CONTEXT_B, data, 2);
		data[0] = (data[0] & 0xfd) | ((ctrl->value << 1) & 0x02);
		mt9m111_write_sensor(cam, MT9M111_SC_R_MODE_CONTEXT_B, data, 2);
		break;

	case V4L2_CID_GAIN:
		tmp = ctrl->value;

		if (tmp > 63 * 2 * 2)
	               return -EINVAL;

		if ((tmp >= 64 * 2) && (tmp < 63 * 2 * 2))
			tmp = (1 << 10) | (1 << 9) | (tmp / 4);
		else if ((tmp >= 64) && (tmp < 64 * 2))
			tmp = (1 << 9) | (tmp / 2);

		data[1] = (tmp & 0xff00) >> 8;
		data[0] = (tmp & 0xff);
		PDEBUG(DBG_V4L2, "tmp=%d, data[1]=%d, data[0]=%d", tmp, data[1], data[0]);

		PDEBUG(DBG_V4L2, "Set gain to %d", tmp);
		mt9m111_write_sensor(cam, MT9M111_SC_GLOBAL_GAIN, data, 2);
		break;
	}
	return (err < 0) ? err : 0;
}

static int mt9m111_read_sensor(struct m5602_camera *cam, const u8 address,
			       u8 *i2c_data, const u8 len) {
	int err, i;

	do {
		err = m5602_read_bridge(cam, M5602_XB_I2C_STATUS, i2c_data);
	} while ((*i2c_data & I2C_BUSY) && !err);

	m5602_write_bridge(cam, M5602_XB_I2C_DEV_ADDR,
			   cam->sensor->i2c_slave_id);
	m5602_write_bridge(cam, M5602_XB_I2C_REG_ADDR, address);
	m5602_write_bridge(cam, M5602_XB_I2C_CTRL, 0x1a);

	for (i = 0; i < len; i++) {
		err = m5602_read_bridge(cam, M5602_XB_I2C_DATA, &(i2c_data[i]));

		PDEBUG(DBG_TRACE, "Reading sensor register "
				"0x%x containing 0x%x ", address, *i2c_data);
	}
	return (err < 0) ? err : 0;
}

static int mt9m111_write_sensor(struct m5602_camera *cam, const u8 address,
				u8 *i2c_data, const u8 len)
{
	int err, i;
	u8 *p;

	/* No sensor with a data width larger than 16 bits has yet been seen */
	if (len > 2 || !len)
		return -EINVAL;

	memcpy(cam->control_buffer, sensor_urb_skeleton,
	       sizeof(sensor_urb_skeleton));

	cam->control_buffer[11] = cam->sensor->i2c_slave_id;
	cam->control_buffer[15] = address;

	p = cam->control_buffer + 16;

	/* Copy a four byte write sequence for each byte to be written to */
	for (i = 0; i < len; i++) {
		memcpy(p, sensor_urb_skeleton + 16, 4);
		p[3] = i2c_data[i];
		p += 4;
		PDEBUG(DBG_TRACE, "Writing sensor register 0x%x with 0x%x",
		       address, i2c_data[i]);
	}

	/* Copy the tailer */
	memcpy(p, sensor_urb_skeleton + 20, 4);

	/* Set the total length */
	p[3] = 0x10 + len;

	err = usb_control_msg(cam->udev, usb_sndctrlpipe(cam->udev, 0),
			      0x04, 0x40, 0x19,
			      0x0000, cam->control_buffer,
			      20 + len * 4, M5602_URB_MSG_TIMEOUT);

	return (err < 0) ? err : 0;

}

static void mt9m111_dump_registers(struct m5602_camera *cam)
{
	u8 address, value[2] = {0x00, 0x00};

	info("Dumping the mt9m111 register state");

	info("Dumping the mt9m111 sensor core registers");
	value[1] = MT9M111_SENSOR_CORE;
	mt9m111_write_sensor(cam, MT9M111_PAGE_MAP, value, 2);
	for (address = 0; address < 0xff; address++) {
		mt9m111_read_sensor(cam, address, value, 2);
		info("register 0x%x contains 0x%x%x",
		     address, value[0], value[1]);
	}

	info("Dumping the mt9m111 color pipeline registers");
	value[1] = MT9M111_COLORPIPE;
	mt9m111_write_sensor(cam, MT9M111_PAGE_MAP, value, 2);
	for (address = 0; address < 0xff; address++) {
		mt9m111_read_sensor(cam, address, value, 2);
		info("register 0x%x contains 0x%x%x",
		     address, value[0], value[1]);
	}

	info("Dumping the mt9m111 camera control registers");
	value[1] = MT9M111_CAMERA_CONTROL;
	mt9m111_write_sensor(cam, MT9M111_PAGE_MAP, value, 2);
	for (address = 0; address < 0xff; address++) {
		mt9m111_read_sensor(cam, address, value, 2);
		info("register 0x%x contains 0x%x%x",
		     address, value[0], value[1]);
	}

	info("mt9m111 register state dump complete");
}
