/*
 * Driver for the ov9650 sensor
 *
 * Copyright (C) 2008 Erik Andrén
 * Copyright (C) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (C) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver sofwtare,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#include "m5602_ov9650.h"

static int ov9650_read_sensor(struct m5602_camera *cam, const u8 address,
		      u8 *i2c_data, const u8 len)
{
	int err, i;

	do {
		err = m5602_read_bridge(cam, M5602_XB_I2C_STATUS, i2c_data);
	} while ((*i2c_data & I2C_BUSY) && !err);

	m5602_write_bridge(cam, M5602_XB_I2C_DEV_ADDR,
			   ov9650.i2c_slave_id);
	m5602_write_bridge(cam, M5602_XB_I2C_REG_ADDR, address);
	m5602_write_bridge(cam, M5602_XB_I2C_CTRL, 0x10 + len);
	m5602_write_bridge(cam, M5602_XB_I2C_CTRL, 0x08);

	for (i = 0; i < len; i++) {
		err = m5602_read_bridge(cam, M5602_XB_I2C_DATA, &(i2c_data[i]));

		PDEBUG(DBG_TRACE, "Reading sensor register "
				"0x%x containing 0x%x ", address, *i2c_data);
	}
	return (err < 0) ? err : 0;
}

static int ov9650_write_sensor(struct m5602_camera *cam, const u8 address,
			u8 *i2c_data, const u8 len)
{
	int err, i;
	u8 *p;

	/* No sensor with a data width larger than 16 bits has yet been seen */
	if (len > 1 || !len)
		return -EINVAL;

	memcpy(cam->control_buffer, sensor_urb_skeleton,
	       sizeof(sensor_urb_skeleton));

	cam->control_buffer[11] = cam->sensor->i2c_slave_id;
	cam->control_buffer[15] = address;

	/* Special case larger sensor writes */
	p = cam->control_buffer + 16;

	/* Copy a four byte write sequence for each byte to be written to */
	for (i = 0; i < len; i++) {
		memcpy(p, sensor_urb_skeleton + 16, 4);
		p[3] = i2c_data[i];
		p += 4;
		PDEBUG(DBG_TRACE, "Writing sensor register 0x%x with 0x%x",
		       address, i2c_data[i]);
	}

	/* Copy the tailer */
	memcpy(p, sensor_urb_skeleton + 20, 4);

	/* Set the total length */
	p[3] = 0x10 + len;

	err = usb_control_msg(cam->udev, usb_sndctrlpipe(cam->udev, 0),
			      0x04, 0x40, 0x19,
			      0x0000, cam->control_buffer,
			      20 + len * 4, M5602_URB_MSG_TIMEOUT);

	return (err < 0) ? err : 0;
}

static int ov9650_probe(struct m5602_camera *cam)
{
	u8 prod_id = 0, ver_id = 0, i;

	info("Probing for an ov9650 sensor");

	/* Run the pre-init to actually probe the unit */
	for (i = 0; i < ARRAY_SIZE(preinit_ov9650); i++) {
		u8 data = preinit_ov9650[i][2];
		if (preinit_ov9650[i][0] == SENSOR)
			ov9650_write_sensor(cam,
					    preinit_ov9650[i][1], &data, 1);
		else
			m5602_write_bridge(cam, preinit_ov9650[i][1], data);
	}

	if (ov9650_read_sensor(cam, OV9650_PID, &prod_id, 1))
		return -ENODEV;

	if (ov9650_read_sensor(cam, OV9650_VER, &ver_id, 1))
		return -ENODEV;

	if ((prod_id == 0x96) && (ver_id == 0x52)) {
		info("Detected an ov9650 sensor");
		return 0;
	}

	if (force_sensor == OV9650_SENSOR) {
		info("Forcing an %s sensor", ov9650.name);
		return 0;
	}

	return -ENODEV;
}

static int ov9650_init(struct m5602_camera *cam)
{
	int i, err = 0;
	u8 data;

	if (dump_sensor)
		ov9650_dump_registers(cam);

	for (i = 0; i < ARRAY_SIZE(init_ov9650) && !err; i++) {
		data = init_ov9650[i][2];
		if (init_ov9650[i][0] == SENSOR)
			err = ov9650_write_sensor(cam, init_ov9650[i][1],
				&data, 1);
		else
			err = m5602_write_bridge(cam, init_ov9650[i][1], data);
	}

	if (!err && dmi_check_system(ov9650_vflip_dmi_table)) {
		info("vertical flip quirk active");
		data = 0x10;
		err = ov9650_write_sensor(cam, OV9650_MVFP, &data, 1);
	}

	return (err < 0) ? err : 0;
}

static int ov9650_get_ctrl(struct m5602_camera *cam, struct v4l2_control *ctrl)
{
	u8 i2c_data = 0, id1 = 0, id2 = 0;
	int err = 0;

	switch (ctrl->id) {
	case V4L2_CID_EXPOSURE:
		/* Reading OV9650_COM1 borks the webcam feed */
		/* err = read_sensor(cam, OV9650_COM1, &i2c_data);
		if (err < 0)
		break;
		id1 = i2c_data & 0x3; */

		err = ov9650_read_sensor(cam, OV9650_AECH, &i2c_data, 1);
		if (err < 0)
			break;

		id2 = i2c_data;

		err = ov9650_read_sensor(cam, OV9650_AECHM, &i2c_data, 1);
		ctrl->value = (id1 << 10) | (id2 << 2) | id1;
		PDEBUG(DBG_V4L2, "Read exposure %d", ctrl->value);
		break;

	case V4L2_CID_GAIN:
	/* FIXME: Touching the VREF register causes hangs
		ov9650_read_sensor(cam, OV9650_VREF, i2c_data);
		id1 = i2c_data & 0xC0; */
		err = ov9650_read_sensor(cam, OV9650_GAIN, &i2c_data, 1);
		ctrl->value = i2c_data;
		PDEBUG(DBG_V4L2, "Read gain %d", ctrl->value);
		break;

	case V4L2_CID_RED_BALANCE:
		err = ov9650_read_sensor(cam, OV9650_RED, &i2c_data, 1);
		ctrl->value = i2c_data;
		PDEBUG(DBG_V4L2, "Read read balance %d", ctrl->value);
		break;

	case V4L2_CID_BLUE_BALANCE:
		err = ov9650_read_sensor(cam, OV9650_BLUE, &i2c_data, 1);
		ctrl->value = i2c_data;
		PDEBUG(DBG_V4L2, "Read blue balance %d", ctrl->value);
		break;

	case V4L2_CID_HFLIP:
		err = ov9650_read_sensor(cam, OV9650_MVFP, &i2c_data, 1);
		ctrl->value = i2c_data & 0x10;
		PDEBUG(DBG_V4L2, "Read horizontal flip %d", ctrl->value);
		break;

	case V4L2_CID_VFLIP:
		err = ov9650_read_sensor(cam, OV9650_MVFP, &i2c_data, 1);
		ctrl->value = i2c_data & 0x08;
		PDEBUG(DBG_V4L2, "Read vertical flip %d", ctrl->value);
		break;

	default:
		PDEBUG(DBG_V4L2, "Unknown V4L2, id 0x%x", ctrl->value);
		return -EINVAL;
	}
	return (err < 0) ? err : 0;
}

static int ov9650_set_ctrl(struct m5602_camera *cam,
			   const struct v4l2_control *ctrl)
{
	u8 register_data = 0;
	int err = 0;

	switch (ctrl->id) {
	case V4L2_CID_EXPOSURE:
		PDEBUG(DBG_V4L2, "Set exposure to %d",
			ctrl->value & 0xffff);
		/* The 6 MSBs */
		register_data = ((ctrl->value & 0xfc00) >> 10);
		err = ov9650_write_sensor(cam, OV9650_AECHM,
					&register_data, 1);
		if (err < 0)
			break;

		/* The 8 middle bits */
		register_data = ((ctrl->value & 0x3fc) >> 2);
		err = ov9650_write_sensor(cam, OV9650_AECH,
					&register_data, 1);
		if (err < 0)
			break;

		/* The 2 LSBs */
		register_data = ctrl->value & 0x03;
		err = ov9650_write_sensor(cam, OV9650_COM1, &register_data, 1);
		break;

	case V4L2_CID_GAIN:
		PDEBUG(DBG_V4L2, "Set gain to %d", ctrl->value & 0xff);

	/* FIXME: The gain register is 10 bits wide, but touching
		(even reading) the OV9650_VREF register causes instant
		hangs of the camera stream */

		/* The 2 MSB */
	/* Read the OV9650_VREF register first to avoid
		corrupting the VREF high and low bits */
		/* ov9650_read_sensor(cam, OV9650_VREF, &register_data); */
		/* Mask away all uninteresting bits */
	/* register_data = ((ctrl->value & 0x300) >> 2)
		| (register_data & 0x3F); */

		/* The 8 LSBs */
		register_data = ctrl->value & 0xff;
		err = ov9650_write_sensor(cam, OV9650_GAIN, &register_data, 1);
		break;

	case V4L2_CID_BLUE_BALANCE:
		PDEBUG(DBG_V4L2, "Set blue balance to %d",
			ctrl->value & 0xff);

		register_data = ctrl->value & 0xff;
		err = ov9650_write_sensor(cam, OV9650_BLUE, &register_data, 1);
		break;

	case V4L2_CID_RED_BALANCE:
		PDEBUG(DBG_V4L2, "Set red balance to %d",
			ctrl->value & 0xff);

		register_data = ctrl->value & 0xff;
		err = ov9650_write_sensor(cam, OV9650_RED, &register_data, 1);
		break;

	case V4L2_CID_HFLIP:
		PDEBUG(DBG_V4L2, "Setting horizontal mirroring to %d",
			ctrl->value);
		err = ov9650_read_sensor(cam, OV9650_MVFP, &register_data, 1);
		if (err < 0)
			break;

		register_data = ((register_data & 0xdf) |
				((ctrl->value & 0x01)) << 5);

		err = ov9650_write_sensor(cam, OV9650_MVFP, &register_data, 1);
		break;

	case V4L2_CID_VFLIP:
		PDEBUG(DBG_V4L2, "Setting vertical mirroring to %d",
			ctrl->value);
		err = ov9650_read_sensor(cam, OV9650_MVFP, &register_data, 1);
		if (err < 0)
			break;

		register_data = ((register_data & 0xef)
				| ((ctrl->value & 0x01)) << 4);

		err = ov9650_write_sensor(cam, OV9650_MVFP, &register_data, 1);
		break;

	default:
		PDEBUG(DBG_V4L2, "Illegal V4L id %d, value %d",
			ctrl->id, ctrl->value);
		return -EINVAL;
	}
	return (err < 0) ? err : 0;
}

static void ov9650_dump_registers(struct m5602_camera *cam)
{
	int address;
	info("Dumping the ov9650 register state");
	for (address = 0; address < 0xa9; address++) {
		u8 value;
		ov9650_read_sensor(cam, address, &value, 1);
		info("register 0x%x contains 0x%x",
		     address, value);
	}

	info("ov9650 register state dump complete");

	info("Probing for which registers that are read/write");
	for (address = 0; address < 0xff; address++) {
		u8 old_value, ctrl_value;
		u8 test_value[2] = {0xff, 0xff};

		ov9650_read_sensor(cam, address, &old_value, 1);
		ov9650_write_sensor(cam, address, test_value, 1);
		ov9650_read_sensor(cam, address, &ctrl_value, 1);

		if (ctrl_value == test_value[0])
			info("register 0x%x is writeable", address);
		else
			info("register 0x%x is read only", address);

		/* Restore original value */
		ov9650_write_sensor(cam, address, &old_value, 1);
	}
}
