/*
 * Driver for the ov9650 sensor
 *
 * Copyright (C) 2008 Erik Andrén
 * Copyright (C) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (C) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver sofwtare,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#ifndef M5602_OV9650_H_
#define M5602_OV9650_H_

#include <linux/dmi.h>

#include "m5602.h"

/*****************************************************************************/

#define OV9650_GAIN   0x00
#define OV9650_BLUE   0x01
#define OV9650_RED    0x02
#define OV9650_VREF   0x03
#define OV9650_COM1   0x04
#define OV9650_BAVE   0x05
#define OV9650_GEAVE  0x06
#define OV9650_RSVD   0x07
#define OV9650_PID    0x0a
#define OV9650_VER    0x0b
#define OV9650_COM5   0x0e
#define OV9650_COM6   0x0f
#define OV9650_AECH   0x10
#define OV9650_CLKRC  0x11
#define OV9650_COM7   0x12
#define OV9650_COM8   0x13
#define OV9650_COM9   0x14
#define OV9650_COM10  0x15
#define OV9650_RSVD16 0x16
#define OV9650_HSTART 0x17
#define OV9650_HSTOP  0x18
#define OV9650_VSTRT  0x19
#define OV9650_VSTOP  0x1a
#define OV9650_PSHFT  0x1b
#define OV9650_MVFP   0x1e
#define OV9650_AEW    0x24
#define OV9650_AEB    0x25
#define OV9650_VPT    0x26
#define OV9650_BBIAS  0x27
#define OV9650_GbBIAS 0x28
#define OV9650_Gr_COM 0x29
#define OV9650_RBIAS  0x2c
#define OV9650_HREF   0x32
#define OV9650_CHLF   0x33
#define OV9650_ARBLM  0x34
#define OV9650_RSVD1  0x35
#define OV9650_ADC    0x37
#define OV9650_ACOM38 0x38
#define OV9650_OFON   0x39
#define OV9650_TSLB   0x3a
#define OV9650_COM12  0x3c
#define OV9650_COM13  0x3d
#define OV9650_COM15  0x40
#define OV9650_COM16  0x41
#define OV9650_LCC1   0x62
#define OV9650_LCC2   0x63
#define OV9650_LCC3   0x64
#define OV9650_LCC4   0x65
#define OV9650_LCC5   0x66
#define OV9650_HV     0x69
#define OV9650_COM21  0x8b
#define OV9650_COM22  0x8c
#define OV9650_COM24  0x8e
#define OV9650_DBLC1  0x8f
#define OV9650_RSVD94 0x94
#define OV9650_RSVD95 0x95
#define OV9650_RSVD96 0x96
#define OV9650_LCCFB  0x9d
#define OV9650_LCCFR  0x9e
#define OV9650_AECHM  0xa1
#define OV9650_COM26  0xa5
#define OV9650_ACOMA8 0xa8
#define OV9650_ACOMA9 0xa9

#define GAIN_DEFAULT 0x14
#define RED_GAIN_DEFAULT 0x70
#define BLUE_GAIN_DEFAULT 0x20
#define EXPOSURE_DEFAULT 0x5003

/*****************************************************************************/

static int ov9650_get_ctrl(struct m5602_camera *cam,
			   struct v4l2_control *ctrl);
static int ov9650_set_ctrl(struct m5602_camera *cam,
			   const struct v4l2_control *ctrl);
static int ov9650_probe(struct m5602_camera *cam);
static int ov9650_init(struct m5602_camera *cam);

static int ov9650_read_sensor(struct m5602_camera *cam, const u8 address,
		       u8 *i2c_data, const u8 len);
static int ov9650_write_sensor(struct m5602_camera *cam, const u8 address,
		       u8 *i2c_data, const u8 len);

static void ov9650_dump_registers(struct m5602_camera *cam);

static struct m5602_sensor ov9650 = {
	.name = "OV9650",
	.probe = ov9650_probe,
	.init = ov9650_init,
	.get_ctrl = ov9650_get_ctrl,
	.set_ctrl = ov9650_set_ctrl,
	.i2c_slave_id = 0x60,
	.read_sensor = ov9650_read_sensor,
	.write_sensor = &ov9650_write_sensor,

	.qctrl = {
	{
		.id = V4L2_CID_EXPOSURE,
		.type = V4L2_CTRL_TYPE_INTEGER,
		.name = "exposure",
		.minimum = 0x00,
		.maximum = 0xffff,
		.step = 0x1,
		.default_value = EXPOSURE_DEFAULT
	},
	{
		.id = V4L2_CID_GAIN,
		.type = V4L2_CTRL_TYPE_INTEGER,
		.name = "gain",
		.minimum = 0x00,
		/* FIXME: Real gain is 10 bits (0x3ff),
			we currently limit this to 8 bits
			due to hard crashes when touching the register
			containing the higher two bits */
		.maximum = 0xff,
		.step = 0x1,
		.default_value = GAIN_DEFAULT
	},
	{
		.id = V4L2_CID_RED_BALANCE,
		.type = V4L2_CTRL_TYPE_INTEGER,
		.name = "red balance",
		.minimum = 0x00,
		.maximum = 0xff,
		.step = 0x1,
		.default_value = RED_GAIN_DEFAULT
	},
	{
		.id = V4L2_CID_BLUE_BALANCE,
		.type = V4L2_CTRL_TYPE_INTEGER,
		.name = "blue balance",
		.minimum = 0x00,
		.maximum = 0xff,
		.step = 0x1,
		.default_value = BLUE_GAIN_DEFAULT
	},
	{
		.id = V4L2_CID_HFLIP,
		.type = V4L2_CTRL_TYPE_BOOLEAN,
		.name = "horizontal mirror",
		.minimum = 0,
		.maximum = 1,
		.step = 1,
		.default_value = 0
	},
	{
		.id = V4L2_CID_VFLIP,
		.type = V4L2_CTRL_TYPE_BOOLEAN,
		.name = "vertical mirror",
		.minimum = 0,
		.maximum = 1,
		.step = 1,
		.default_value = 0
	}
	},
	.cropcap = {
		.bounds = {
			.left = 0,
			.top = 0,
			.width = M5602_DEFAULT_FRAME_WIDTH,
			.height = M5602_DEFAULT_FRAME_HEIGHT
		},
		.defrect = {
			.left = 0,
			.top = 0,
			.width = M5602_DEFAULT_FRAME_WIDTH,
			.height = M5602_DEFAULT_FRAME_HEIGHT
		}
	},
	.current_fmt = {
		.width = 640,
		.height = 480,
		.pixelformat = V4L2_PIX_FMT_SBGGR8,
		.priv = 1,
		.colorspace = V4L2_COLORSPACE_SRGB
	}
};

static const unsigned char preinit_ov9650[][3] =
{
	/* [INITCAM] */
	{BRIDGE, M5602_XB_MCU_CLK_DIV, 0x02},
	{BRIDGE, M5602_XB_MCU_CLK_CTRL, 0xb0},
	{BRIDGE, M5602_XB_SEN_CLK_DIV, 0x00},
	{BRIDGE, M5602_XB_SEN_CLK_CTRL, 0xb0},
	{BRIDGE, M5602_XB_ADC_CTRL, 0xc0},
	{BRIDGE, M5602_XB_SENSOR_CTRL, 0x00},

	{BRIDGE, M5602_XB_SENSOR_TYPE, 0x08},
	{BRIDGE, M5602_XB_GPIO_DIR, 0x05},
	{BRIDGE, M5602_XB_GPIO_DAT, 0x04},
	{BRIDGE, M5602_XB_GPIO_EN_H, 0x06},
	{BRIDGE, M5602_XB_GPIO_DIR_H, 0x06},
	{BRIDGE, M5602_XB_GPIO_DAT_H, 0x00},
	{BRIDGE, M5602_XB_GPIO_DAT, 0x00},
	{BRIDGE, M5602_XB_I2C_CLK_DIV, 0x0a},
	/* Reset chip */
	{SENSOR, OV9650_COM7, 0x80},
	/* Enable double clock */
	{SENSOR, OV9650_CLKRC, 0x80},
	/* Do something out of spec with the power */
	{SENSOR, OV9650_OFON, 0x40}
};

static const unsigned char init_ov9650[][3] =
{
	/* Set QQVGA */
	{SENSOR, OV9650_COM1, 0x20},
	/* Set fast AGC/AEC algorithm with unlimited step size */
	{SENSOR, OV9650_COM8, 0xc0},

	{SENSOR, OV9650_CHLF, 0x10},
	{SENSOR, OV9650_ARBLM, 0xbf},
	{SENSOR, OV9650_ACOM38, 0x81},
	/* Turn off color matrix coefficient double option */
	{SENSOR, OV9650_COM16, 0x00},
	/* Enable color matrix for RGB/YUV, Delay Y channel,
	set output Y/UV delay to 1 */
	{SENSOR, OV9650_COM13, 0x19},
	/* Enable digital BLC, Set output mode to U Y V Y */
	{SENSOR, OV9650_TSLB, 0x0c},
	/* Limit the AGC/AEC stable upper region */
	{SENSOR, OV9650_COM24, 0x00},
	/* Enable HREF and some out of spec things */
	{SENSOR, OV9650_COM12, 0x73},
	/* Set all DBLC offset signs to positive and
	do some out of spec stuff */
	{SENSOR, OV9650_DBLC1, 0xdf},
	{SENSOR, OV9650_COM21, 0x06},
	{SENSOR, OV9650_RSVD1, 0x91},
	/* Necessary, no camera stream without it */
	{SENSOR, OV9650_RSVD16, 0x06},
	{SENSOR, OV9650_RSVD94, 0x99},
	{SENSOR, OV9650_RSVD95, 0x99},
	{SENSOR, OV9650_RSVD96, 0x04},
	/* Enable full range output */
	{SENSOR, OV9650_COM15, 0x0},
	/* Enable HREF at optical black, enable ADBLC bias,
	enable ADBLC, reset timings at format change */
	{SENSOR, OV9650_COM6, 0x4b},
	/* Subtract 32 from the B channel bias */
	{SENSOR, OV9650_BBIAS, 0xa0},
	/* Subtract 32 from the Gb channel bias */
	{SENSOR, OV9650_GbBIAS, 0xa0},
	/* Do not bypass the analog BLC and to some out of spec stuff */
	{SENSOR, OV9650_Gr_COM, 0x00},
	/* Subtract 32 from the R channel bias */
	{SENSOR, OV9650_RBIAS, 0xa0},
	/* Subtract 32 from the R channel bias */
	{SENSOR, OV9650_RBIAS, 0x0},
	{SENSOR, OV9650_COM26, 0x80},
	{SENSOR, OV9650_ACOMA9, 0x98},
	/* Set the AGC/AEC stable region upper limit */
	{SENSOR, OV9650_AEW, 0x68},
	/* Set the AGC/AEC stable region lower limit */
	{SENSOR, OV9650_AEB, 0x5c},
	/* Set the high and low limit nibbles to 3 */
	{SENSOR, OV9650_VPT, 0xc3},
	/* Set the Automatic Gain Ceiling (AGC) to 128x,
	drop VSYNC at frame drop,
	limit exposure timing,
	drop frame when the AEC step is larger than the exposure gap */
	{SENSOR, OV9650_COM9, 0x6e},
	/* Set VSYNC negative, Set RESET to SLHS (slave mode horizontal sync)
	and set PWDN to SLVS (slave mode vertical sync) */
	{SENSOR, OV9650_COM10, 0x42},
	/* Set horizontal column start high to default value */
	{SENSOR, OV9650_HSTART, 0x1a},
	/* Set horizontal column end */
	{SENSOR, OV9650_HSTOP, 0xbf},
	/* Complementing register to the two writes above */
	{SENSOR, OV9650_HREF, 0xb2},
	/* Set vertical row start high bits */
	{SENSOR, OV9650_VSTRT, 0x02},
	/* Set vertical row end low bits */
	{SENSOR, OV9650_VSTOP, 0x7e},
	/* Set complementing vertical frame control */
	{SENSOR, OV9650_VREF, 0x10},
	/* Set raw RGB output format with VGA resolution */
	{SENSOR, OV9650_COM7, 0x45},
	{SENSOR, OV9650_ADC, 0x04},
	{SENSOR, OV9650_HV, 0x40},
	/* Enable denoise, and white-pixel erase */
	{SENSOR, OV9650_COM22, 0x23},

	/* Set the high bits of the exposure value */
	{SENSOR, OV9650_AECH, ((EXPOSURE_DEFAULT & 0xff00) >> 8)},
	/* Set the low bits of the exposure value */
	{SENSOR, OV9650_COM1, (EXPOSURE_DEFAULT & 0xff)},

	{SENSOR, OV9650_GAIN, GAIN_DEFAULT},
	{SENSOR, OV9650_BLUE, BLUE_GAIN_DEFAULT},
	{SENSOR, OV9650_RED, RED_GAIN_DEFAULT},

	{BRIDGE, M5602_XB_LINE_OF_FRAME_H, 0x82},
	{BRIDGE, M5602_XB_LINE_OF_FRAME_L, 0x00},
	{BRIDGE, M5602_XB_PIX_OF_LINE_H, 0x82},
	{BRIDGE, M5602_XB_PIX_OF_LINE_L, 0x00},
	{BRIDGE, M5602_XB_SIG_INI, 0x01},
	{BRIDGE, M5602_XB_VSYNC_PARA, 0x00},
	{BRIDGE, M5602_XB_VSYNC_PARA, 0x09},
	{BRIDGE, M5602_XB_VSYNC_PARA, 0x00},
	{BRIDGE, M5602_XB_VSYNC_PARA, 0x00},
	{BRIDGE, M5602_XB_VSYNC_PARA, 0xe0},
	{BRIDGE, M5602_XB_VSYNC_PARA, 0x00},
	{BRIDGE, M5602_XB_HSYNC_PARA, 0x00},
	{BRIDGE, M5602_XB_HSYNC_PARA, 0x5e},
	{BRIDGE, M5602_XB_HSYNC_PARA, 0x02},
	{BRIDGE, M5602_XB_HSYNC_PARA, 0xde}
};

/* Vertically flips the image if matched, needed for machines
   where the sensor is mounted upside down */
static const struct dmi_system_id ov9650_vflip_dmi_table[] = {
	{
		.ident = "ASUS A6VC",
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR, "ASUSTeK Computer Inc."),
			DMI_MATCH(DMI_PRODUCT_NAME, "A6VC")
		},
	},
	{
		.ident = "ASUS A6VM",
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR, "ASUSTeK Computer Inc."),
			DMI_MATCH(DMI_PRODUCT_NAME, "A6VM")
		},
	},
	{
		.ident = "ASUS A6JC",
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR, "ASUSTeK Computer Inc."),
			DMI_MATCH(DMI_PRODUCT_NAME, "A6JC")
		},
	},
	{
		.ident = "ASUS A6Kt",
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR, "ASUSTeK Computer Inc."),
			DMI_MATCH(DMI_PRODUCT_NAME, "A6Kt")
		},
	},
	{ }
};

#endif
