/*
 * Driver for the s5k4aa sensor
 *
 * Copyright (C) 2008 Erik Andrén
 * Copyright (C) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (C) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver sofwtare,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#include "m5602_s5k4aa.h"

static int s5k4aa_read_sensor(struct m5602_camera *cam, const u8 address,
		       u8 *i2c_data, const u8 len)
{
	int err, i;

	do {
		err = m5602_read_bridge(cam, M5602_XB_I2C_STATUS, i2c_data);
	} while ((*i2c_data & I2C_BUSY) && !err);

	m5602_write_bridge(cam, M5602_XB_I2C_DEV_ADDR,
			   cam->sensor->i2c_slave_id);
	m5602_write_bridge(cam, M5602_XB_I2C_REG_ADDR, address);
	m5602_write_bridge(cam, M5602_XB_I2C_CTRL, 0x18 + len);

	do {
		err = m5602_read_bridge(cam, M5602_XB_I2C_STATUS, i2c_data);
	} while ((*i2c_data & I2C_BUSY) && !err);

	for (i = 0; i < len; i++) {
		err = m5602_read_bridge(cam, M5602_XB_I2C_DATA, &(i2c_data[i]));

		PDEBUG(DBG_TRACE, "Reading sensor register "
				  "0x%x containing 0x%x ", address, *i2c_data);
	}
	return (err < 0) ? err : 0;
}

static int s5k4aa_write_sensor(struct m5602_camera *cam, const u8 address,
			u8 *i2c_data, const u8 len)
{
	int err, i;
	u8 *p;

	/* No sensor with a data width larger than 16 bits has yet been seen */
	if (len > 2 || !len)
		return -EINVAL;

	memcpy(cam->control_buffer, sensor_urb_skeleton,
	       sizeof(sensor_urb_skeleton));

	cam->control_buffer[11] = cam->sensor->i2c_slave_id;
	cam->control_buffer[15] = address;

	/* Special case larger sensor writes */
	p = cam->control_buffer + 16;

	/* Copy a four byte write sequence for each byte to be written to */
	for (i = 0; i < len; i++) {
		memcpy(p, sensor_urb_skeleton + 16, 4);
		p[3] = i2c_data[i];
		p += 4;
		PDEBUG(DBG_TRACE, "Writing sensor register 0x%x with 0x%x",
		       address, i2c_data[i]);
	}

	/* Copy the tailer */
	memcpy(p, sensor_urb_skeleton + 20, 4);

	/* Set the total length */
	p[3] = 0x10 + len;

	err = usb_control_msg(cam->udev, usb_sndctrlpipe(cam->udev, 0),
			      0x04, 0x40, 0x19,
			      0x0000, cam->control_buffer,
			      20 + len * 4, M5602_URB_MSG_TIMEOUT);

	return (err < 0) ? err : 0;
}

static int s5k4aa_probe(struct m5602_camera *cam)
{
	u8 prod_id;
	int i, err = 0;
	
	info("Probing for a s5k4aa sensor");

	/* Preinit the sensor */
	for (i = 0; i < ARRAY_SIZE(preinit_s5k4aa) && !err; i++) {
		u8 data[2] = {0x00, 0x00};

		switch (preinit_s5k4aa[i][0]) {
			case BRIDGE:
				err = m5602_write_bridge(cam,
						 preinit_s5k4aa[i][1],
						 preinit_s5k4aa[i][2]);
				break;

			case SENSOR:
				data[0] = preinit_s5k4aa[i][2];
				err = s5k4aa_write_sensor(cam,
						preinit_s5k4aa[i][1], data, 1);
				break;

			case SENSOR_LONG:
				data[0] = preinit_s5k4aa[i][2];
				data[1] = preinit_s5k4aa[i][3];
				err = s5k4aa_write_sensor(cam,
						preinit_s5k4aa[i][1], data, 2);
				break;
			default:
				info("Invalid stream command, exiting init");
				return -EINVAL;
		}
	}

	if (force_sensor == S5K4AA_SENSOR) {
		info("Forcing a %s sensor", s5k4aa.name);
		return 0;
	}

	/* Test a read-only register, but we don't know its exact meaning yet */
	if (s5k4aa_read_sensor(cam, 0x00, &prod_id, 1))
		return -ENODEV;

	if (prod_id != 0x24)
		return -ENODEV;
	else
		info("Detected a s5k4aa sensor");

	return 0;
}

static int s5k4aa_init(struct m5602_camera *cam)
{
	int i, err = 0;

	for (i = 0; i < ARRAY_SIZE(init_s5k4aa) && !err; i++) {
		u8 data[2] = {0x00, 0x00};

		switch (init_s5k4aa[i][0]) {
		case BRIDGE:
			err = m5602_write_bridge(cam,
				init_s5k4aa[i][1],
				init_s5k4aa[i][2]);
			break;

		case SENSOR:
			data[0] = init_s5k4aa[i][2];
			err = s5k4aa_write_sensor(cam,
				init_s5k4aa[i][1], data, 1);
			break;

		case SENSOR_LONG:
			data[0] = init_s5k4aa[i][2];
			data[1] = init_s5k4aa[i][3];
			err = s5k4aa_write_sensor(cam,
				init_s5k4aa[i][1], data, 2);
			break;
		default:
			info("Invalid stream command, exiting init");
			return -EINVAL;
		}
	}

	if (dump_sensor)
		s5k4aa_dump_registers(cam);

	if (!err && dmi_check_system(s5k4aa_vflip_dmi_table)) {
		u8 data = 0x02;
		info("vertical flip quirk active");
		s5k4aa_write_sensor(cam, S5K4AA_PAGE_MAP, &data, 1);
		s5k4aa_read_sensor(cam, S5K4AA_READ_MODE, &data, 1);
		data |= S5K4AA_RM_V_FLIP;
		data &= ~S5K4AA_RM_H_FLIP;
		s5k4aa_write_sensor(cam, S5K4AA_READ_MODE, &data, 1);

		/* Decrement COLSTART to preserve color order (BGGR) */
		s5k4aa_read_sensor(cam, S5K4AA_COLSTART_LO, &data, 1);
		data--;
		s5k4aa_write_sensor(cam, S5K4AA_COLSTART_LO, &data, 1);

		/* Increment ROWSTART to preserve color order (BGGR) */
		s5k4aa_read_sensor(cam, S5K4AA_ROWSTART_LO, &data, 1);
		data++;
		s5k4aa_write_sensor(cam, S5K4AA_ROWSTART_LO, &data, 1);
	}

	return (err < 0) ? err : 0;
}

static int s5k4aa_get_ctrl(struct m5602_camera *cam, struct v4l2_control *ctrl)
{
	return 0;
}

static int s5k4aa_set_ctrl(struct m5602_camera *cam,
			   const struct v4l2_control *ctrl)
{
	return 0;
}

static void s5k4aa_dump_registers(struct m5602_camera *cam)
{
	int address;
	info("Dumping the s5k4aa register state");
	for (address = 0; address < 0xff; address++) {
		u8 value[2] = {0, 0};
		s5k4aa_read_sensor(cam, address, value, 2);
		info("register 0x%x contains 0x%x%x",
		     address, value[1], value[0]);
	}
	info("s5k4aa register state dump complete");

	info("Probing for which registers that are read/write");
	for (address = 0; address < 0xff; address++) {
		u8 old_value, ctrl_value;
		u8 test_value[2] = {0xff, 0xff};

		s5k4aa_read_sensor(cam, address, &old_value, 1);
		s5k4aa_write_sensor(cam, address, test_value, 1);
		s5k4aa_read_sensor(cam, address, &ctrl_value, 1);

		if (ctrl_value == test_value[0])
			info("register 0x%x is writeable", address);
		else
			info("register 0x%x is read only", address);

		/* Restore original value */
		s5k4aa_write_sensor(cam, address, &old_value, 1);
	}
	info("Read/write register probing complete");
}
