/*
 * Driver for the unknown_0xdc sensor
 *
 * Copyright (c) 2008 Erik Andrén
 * Copyright (c) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (c) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver sofwtare,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#include "m5602_unknown_0xdc.h"

static int unknown_0xdc_read_sensor(struct m5602_camera *cam, const u8 address, u8 *i2c_data, const u8 len)
{
	int err, i;

	do {
		err = m5602_read_bridge(cam, M5602_XB_I2C_STATUS, i2c_data);
	} while ((*i2c_data & I2C_BUSY) && !err);

	m5602_write_bridge(cam, M5602_XB_I2C_DEV_ADDR,
			   cam->sensor->i2c_slave_id);
	m5602_write_bridge(cam, M5602_XB_I2C_REG_ADDR, address);
	m5602_write_bridge(cam, M5602_XB_I2C_CTRL, 0x10 + len);
	m5602_write_bridge(cam, M5602_XB_I2C_CTRL, 0x08);

	for (i = 0; i < len; i++) {
		err = m5602_read_bridge(cam, M5602_XB_I2C_DATA, &(i2c_data[i]));

		PDEBUG(DBG_TRACE, "Reading sensor register "
				"0x%x containing 0x%x ", address, *i2c_data);
	}
	return (err < 0) ? err : 0;
}

static int unknown_0xdc_write_sensor(struct m5602_camera *cam, const u8 address, u8 *i2c_data, const u8 len)
{
	int err, i;
	u8 *p;

	/* No sensor with a data width larger than 16 bits has yet been seen */
	if (len > 1 || !len)
		return -EINVAL;

	memcpy(cam->control_buffer, sensor_urb_skeleton, sizeof(sensor_urb_skeleton));

	cam->control_buffer[11] = cam->sensor->i2c_slave_id;
	cam->control_buffer[15] = address;

	p = cam->control_buffer + 16;

	/* Copy a four byte write sequence for each byte to be written to */
	for (i = 0; i < len; i++) {
		memcpy(p, sensor_urb_skeleton + 16, 4);
		p[3] = i2c_data[i];
		p += 4;
		PDEBUG(DBG_TRACE, "Writing sensor register 0x%x with 0x%x",
		       address, i2c_data[i]);
	}

	/* Copy the footer */
	memcpy(p, sensor_urb_skeleton + 20, 4);

	/* Set the total length */
	p[3] = 0x10 + len;

	err = usb_control_msg(cam->udev, usb_sndctrlpipe(cam->udev, 0),
			      0x04, 0x40, 0x19,
			      0x0000, cam->control_buffer,
			      20 + len * 4, M5602_URB_MSG_TIMEOUT);

	return (err < 0) ? err : 0;
}


static int unknown_0xdc_probe(struct m5602_camera *cam)
{
	u8 prod_id = 0, ver_id = 0, i;

	info("Probing for a unknown sensor with i2c address 0xdc");

	/* Run the pre-init to actually probe the unit */
	for (i = 0; i < ARRAY_SIZE(preinit_unknown_0xdc); i++) {
		u8 data = preinit_unknown_0xdc[i][2];
		if (preinit_unknown_0xdc[i][0] == SENSOR)
			unknown_0xdc_write_sensor(cam,
				preinit_unknown_0xdc[i][1], &data, 1);
		else
			m5602_write_bridge(cam, preinit_unknown_0xdc[i][1], data);
	}

	if (unknown_0xdc_read_sensor(cam, 0x3, &prod_id, 1))
		return -ENODEV;

	if (unknown_0xdc_read_sensor(cam, 0x4, &ver_id, 1))
		return -ENODEV;

	if ((prod_id == 0x02) && (ver_id == 0xef)) {
		info("Detected a 0xdc sensor");
		return 0;
	}

	if (force_sensor == UNKNOWN_0xDC) {
		info("Forcing the %s sensor", unknown_0xdc.name);
		return 0;
	}

	return -ENODEV;
}

static int unknown_0xdc_init(struct m5602_camera *cam)
{
	int i, err = 0;

	/* Init the sensor */
	for (i = 0; i < ARRAY_SIZE(init_unknown_0xdc); i++) {
		u8 data[2] = {0x00, 0x00};

		switch (init_unknown_0xdc[i][0])
		{
			case BRIDGE:
				err = m5602_write_bridge(cam,
					init_unknown_0xdc[i][1],
					init_unknown_0xdc[i][2]);
				break;

			case SENSOR:
				data[0] = init_unknown_0xdc[i][2];
				err = unknown_0xdc_write_sensor(cam,
					init_unknown_0xdc[i][1], data, 1);
				break;

			case SENSOR_LONG:
				data[0] = init_unknown_0xdc[i][2];
				data[1] = init_unknown_0xdc[i][3];
				err = unknown_0xdc_write_sensor(cam,
					init_unknown_0xdc[i][1], data, 2);
				break;
			default:
				info("Invalid stream command, exiting init");
				return -EINVAL;
		}
	}

	if (dump_sensor)
		unknown_0xdc_dump_registers(cam);

	return (err < 0) ? err : 0;
}

static int unknown_0xdc_get_ctrl(struct m5602_camera *cam,
				 struct v4l2_control *ctrl)
{
	u8 i2c_data;
	int err = 0;
	
	switch (ctrl->id) {
	case V4L2_CID_GAIN:
		err = unknown_0xdc_read_sensor(cam, SENSOR_GAIN,
					       &i2c_data, 1);
		ctrl->value = i2c_data;
		PDEBUG(DBG_V4L2, "Read gain %d", ctrl->value);
		break;

	case V4L2_CID_EXPOSURE:
		err = unknown_0xdc_read_sensor(cam, SENSOR_HIGH_EXPOSURE,
					       &i2c_data, 1);
		PDEBUG(DBG_V4L2, "Read high exposure %d", ctrl->value);
		ctrl->value = (i2c_data << 8);

		err = unknown_0xdc_read_sensor(cam, SENSOR_LOW_EXPOSURE,
					       &i2c_data, 1);
		PDEBUG(DBG_V4L2, "Read low exposure %d", ctrl->value);
		ctrl->value |= i2c_data;

		PDEBUG(DBG_V4L2, "Exposure read as %d", ctrl->value);
	}

	return 0;
}

static int unknown_0xdc_set_ctrl(struct m5602_camera *cam,
		     const struct v4l2_control *ctrl)
{
	u8 register_data = 0;
	int err = 0;

	switch (ctrl->id) {
	case V4L2_CID_GAIN:
		PDEBUG(DBG_V4L2, "Set gain to %d", ctrl->value & 0xff);
		err = unknown_0xdc_write_sensor(cam, SENSOR_GAIN,
						&register_data, 1);
		break;

	case V4L2_CID_EXPOSURE:
		PDEBUG(DBG_V4L2, "Set exposure to %d", ctrl->value & 0xffff);

		register_data = ((ctrl->value & 0xff00) >> 8);
		PDEBUG(DBG_V4L2, "Set exposure to high byte to 0x%x",
		       register_data);

		err = unknown_0xdc_write_sensor(cam, SENSOR_HIGH_EXPOSURE,
			&register_data, 1);

		register_data = (ctrl->value & 0x00ff);
		PDEBUG(DBG_V4L2, "Set exposure to low byte to 0x%x",
		       register_data);
		err = unknown_0xdc_write_sensor(cam, SENSOR_LOW_EXPOSURE,
					        &register_data, 1);
		break;
	}

	return (err < 0) ? err : 0;;
}

static void unknown_0xdc_dump_registers(struct m5602_camera *cam)
{
	int address;
	u8 value = 0;

	info("Dumping the unknown_0xdc sensor core registers");
	for (address = 0; address < 0xff; address++) {
		unknown_0xdc_read_sensor(cam, address, &value, 1);
		info("register 0x%x contains 0x%x",
		     address, value);
	}
	info("unknown_0xdc register state dump complete");

	info("Probing for which registers that are read/write");
	for (address = 0; address < 0xff; address++) {
		u8 old_value, ctrl_value;
		u8 test_value[2] = {0xff, 0xff};

		unknown_0xdc_read_sensor(cam, address, &old_value, 1);
		unknown_0xdc_write_sensor(cam, address, test_value, 1);
		unknown_0xdc_read_sensor(cam, address, &ctrl_value, 1);

		if (ctrl_value == test_value[0])
			info("register 0x%x is writeable", address);
		else
			info("register 0x%x is read only", address);

		/* Restore original value */
		unknown_0xdc_write_sensor(cam, address, &old_value, 1);
	}
}
