/*
 * USB Driver for ALi m5602 based webcams
 *
 * Copyright (C) 2008 Erik Andrén
 * Copyright (C) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (C) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver sofwtare,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#include "m5602.h"

#define VIDIOC_S_CTRL_OLD _IOW('V', 28, struct v4l2_control)

static u32 m5602_request_buffers(struct m5602_camera *cam, u32 count)
{
	u32 i;
	u8 *raw = NULL;

	PDEBUG(DBG_V4L2, "m5602_request_buffers, count %d", count);

	if (!count) {
		PDEBUG(DBG_V4L2, "No buffers requested");
		return 0;
	}

	if (count > M5602_MAX_FRAMES)
		count = M5602_MAX_FRAMES;

	if (cam->nbuffers) {
		if (cam->stream == STREAM_ON)
			m5602_stream_interrupt(cam);

		m5602_stop_transfer(cam);
		m5602_empty_framequeues(cam);
		m5602_release_buffers(cam);
	}

	cam->nbuffers = count;

	/* Allocate buffers where the raw bayer encoded frames will reside */
	while (cam->nbuffers) {
		raw = vmalloc_32_user(cam->nbuffers *
				M5602_DEFAULT_FRAME_WIDTH *
				M5602_DEFAULT_FRAME_HEIGHT);
		if (raw)
			break;

		/* Buffer allocation failed, try with one buffer less */
		cam->nbuffers--;
	}

	for (i = 0; i < cam->nbuffers; i++) {
		cam->frame[i].rawdata = raw + (i *
			PAGE_ALIGN(M5602_DEFAULT_FRAME_WIDTH *
				   M5602_DEFAULT_FRAME_HEIGHT));
		memset(&cam->frame[i].buf, 0, sizeof(struct v4l2_buffer));

		cam->frame[i].buf.m.offset = i * cam->sensor->current_fmt.width
			* cam->sensor->current_fmt.height;
		cam->frame[i].buf.length = cam->sensor->current_fmt.width
			* cam->sensor->current_fmt.height;

		cam->frame[i].buf.index = i;
		cam->frame[i].buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		cam->frame[i].buf.field = V4L2_FIELD_NONE;
		cam->frame[i].buf.memory = V4L2_MEMORY_MMAP;
	}

	PDEBUG(DBG_V4L2, "Allocated %d buffers", cam->nbuffers);
	return cam->nbuffers;
}

static void m5602_release_buffers(struct m5602_camera *cam)
{
	if (!cam)
		return;

	if (!cam->nbuffers)
		return;

	PDEBUG(DBG_INIT, "Releasing %d allocated buffers", cam->nbuffers);
	if (cam->io == V4L2_MEMORY_MMAP) {
		if (cam->frame[0].rawdata)
			vfree(cam->frame[0].rawdata);
		cam->nbuffers = 0;
	}
	cam->frame_current = NULL;
}

/* Empties the frame queue and re-initializes the linked lists */
static void m5602_empty_framequeues(struct m5602_camera *cam)
{
	u32 i;

	INIT_LIST_HEAD(&cam->inqueue);
	INIT_LIST_HEAD(&cam->outqueue);

	for (i = 0; i < M5602_MAX_FRAMES; i++) {
		cam->frame[i].state = FRAME_UNUSED;
		cam->frame[i].buf.bytesused = 0;
	}
}

/* Executes when fresh URBs arrive from the bridge */
void m5602_urb_complete(struct urb *urb)
{
	struct m5602_camera *cam = urb->context;
	struct m5602_frame_t *f;
	unsigned long lock_flags;
	int i, err = 0;

	if (urb->status == -ENOENT)
		return;

	f = cam->frame_current;

	if (cam->stream == STREAM_INTERRUPT) {
		PDEBUG(DBG_V4L2, "Stream interrupted");
		cam->stream = STREAM_OFF;
		if (f)
			f->state = FRAME_QUEUED;

		wake_up_interruptible(&cam->wait_stream);
	}

	if (cam->state & DEV_DISCONNECTED) {
		PDEBUG(DBG_V4L2, "Device disconnect");
		return;
	}

	if (cam->state & DEV_MISCONFIGURED) {
		PDEBUG(DBG_V4L2, "Device misconfigured");
		wake_up_interruptible(&cam->wait_frame);
		return;
	}

	if (list_empty(&cam->inqueue) || (cam->stream == STREAM_OFF)) {
		if (list_empty(&cam->inqueue))
			PDEBUG(DBG_V4L2, "Inqueue empty, resubmitting");
		else
			PDEBUG(DBG_V4L2, "Stream is off, resubmitting");
		goto resubmit;
	}

	if (!f) {
		spin_lock_irqsave(&cam->queue_lock, lock_flags);
		/* Pull the next frame from the inqueue */
		f = list_entry(cam->inqueue.next, struct m5602_frame_t, frame);
		spin_unlock_irqrestore(&cam->queue_lock, lock_flags);
	}

	/* frame delimiter: ff bf xx xx ff ff 00 00 aa 55 aa 55 */
	for (i = 0; i < urb->number_of_packets; i++) {
		int len;
		u8 *pos;

		if (urb->iso_frame_desc[i].status) {
			PDEBUG(DBG_V4L2, "Error in isochronous frame %d",
			       urb->iso_frame_desc[i].status);
			f->state = FRAME_ERROR;
			f->buf.bytesused = 0;
			continue;
		}

		len = urb->iso_frame_desc[i].actual_length;
		pos = urb->iso_frame_desc[i].offset + urb->transfer_buffer;

		/* If the packet is smaller than 12 bytes in size, skip it */
		if (len < 12)
			continue;

		/* Frame delimiter */
		if (pos[0] == 0xff && pos[1] == 0xbf &&
		    pos[4] == 0xff && pos[5] == 0xff) {
			PDEBUG(DBG_DATA, "Frame delimiter detected");

			/* Remove the extra fluff appended on each header */
			pos += 6;
			len -= 6;

			switch (f->state) {
			case FRAME_GRABBING:
				f->buf.bytesused =
					cam->sensor->current_fmt.sizeimage;
				do_gettimeofday(&f->buf.timestamp);
				f->state = FRAME_DONE;

				PDEBUG(DBG_V4L2, "Frame %d captured, contains "
						 "%d bytes", f->framenum,
						 f->buf.bytesused);

				spin_lock_irqsave(&cam->queue_lock, lock_flags);

				/* Add the current frame to the outqueue */
				list_move_tail(&f->frame, &cam->outqueue);

				if (list_empty(&cam->inqueue)) {
					cam->frame_current = NULL;
					spin_unlock_irqrestore(&cam->queue_lock,
						lock_flags);
					break;
				}

				/* Pull the next frame */
				f = list_entry(cam->inqueue.next,
					struct m5602_frame_t, frame);

				spin_unlock_irqrestore(&cam->queue_lock,
					lock_flags);

				wake_up_interruptible(&cam->wait_frame);

				/* Start with the new frame */
				PDEBUG(DBG_DATA, "Starting frame %d, "
					"copying %d bytes",
					f->framenum, len);

				memcpy(f->rawdata, pos, len);
				f->buf.sequence = ++cam->frame_count;
				f->framenum = f->buf.sequence;
				f->buf.bytesused = len;
				f->state = FRAME_GRABBING;

				cam->frame_current = f;
				break;

			case FRAME_QUEUED:
			case FRAME_ERROR:
				PDEBUG(DBG_DATA, "Starting frame %d, "
				       "copying %d bytes", f->framenum, len);

				memcpy(f->rawdata, pos, len);
				f->buf.sequence  = ++cam->frame_count;
				f->framenum = f->buf.sequence;
				f->buf.bytesused = len;
				f->state = FRAME_GRABBING;

				cam->frame_current = f;
				break;

			default:
				break;
			}
		} else {
			if (f->state == FRAME_GRABBING) {
				/* Remove urb header */
				len -= 4;
				pos += 4;

				PDEBUG(DBG_DATA, "Continuing frame %d, "
					"copying %d bytes", f->framenum, len);
				if ((f->buf.bytesused + len) <=
					(M5602_DEFAULT_FRAME_WIDTH *
					 M5602_DEFAULT_FRAME_HEIGHT)) {
					memcpy(f->rawdata + f->buf.bytesused,
					       pos, len);
					f->buf.bytesused += len;
				}
			}
		}
	}

resubmit:
	urb->dev = cam->udev;
	err = usb_submit_urb(urb, GFP_ATOMIC);
}

static int m5602_start_transfer(struct m5602_camera *cam)
{
	int i, j, retval = 0;
	struct urb *urb;
	struct usb_device *udev = cam->udev;

	PDEBUG(DBG_V4L2, "Initialize isochronous buffers and urbs, "
		"ISOC IN: 0x%x", cam->isoc_in_endpointAddr);
	PDEBUG(DBG_V4L2, "Number of urbs: %d, packet size: %zi, "
		"number of packets: %d", M5602_URBS,
		cam->isoc_in_size, M5602_ISOC_PACKETS);

	/* Allocate isoc buffers */
	for (i = 0; i < M5602_URBS; i++) {
		cam->isoc_in_buffer[i] =
			kzalloc(cam->isoc_in_size *
			M5602_ISOC_PACKETS, GFP_KERNEL);
		if (!cam->isoc_in_buffer[i]) {
			PDEBUG(DBG_DATA, "No enough memory for ISOC buffer %d",
			       i);
			retval = -ENOMEM;
			goto free_buffers;
		}
	}

	/* Allocate urbs */
	for (i = 0; i < M5602_URBS; i++) {
		urb = usb_alloc_urb(M5602_ISOC_PACKETS, GFP_KERNEL);

		if (!urb) {
			PDEBUG(DBG_DATA, "No enough memory for URB");
			retval = -ENOMEM;
			goto free_urbs;
		}

		cam->urb[i] = urb;
		urb->dev = udev;
		urb->context = cam;
		urb->pipe = usb_rcvisocpipe(udev, cam->isoc_in_endpointAddr);
		urb->transfer_flags = URB_ISO_ASAP;
		urb->number_of_packets = M5602_ISOC_PACKETS;
		urb->complete = m5602_urb_complete;
		urb->transfer_buffer = cam->isoc_in_buffer[i];
		urb->transfer_buffer_length =
			cam->isoc_in_size * M5602_ISOC_PACKETS;
		urb->interval = 1;

		for (j = 0; j < M5602_ISOC_PACKETS; j++) {
			urb->iso_frame_desc[j].offset = cam->isoc_in_size * j;
			urb->iso_frame_desc[j].length = cam->isoc_in_size;
		}
	}

	cam->frame_current = NULL;
	retval = usb_set_interface(cam->udev, 0, 2);
	if (retval)
		goto free_urbs;

	for (i = 0; i < M5602_URBS; i++) {
		if (cam->urb[i]) {
			retval = usb_submit_urb(cam->urb[i], GFP_KERNEL);
			if (retval) {
				PDEBUG(DBG_V4L2, "usb_submit_urb failed! %d",
				       retval);
				goto free_urbs;
			}
		}
	}

	/* Send start command to the camera */
	for (i = 0; i < M5602_URBS; i++) {
		u8 buffer[4] = {0x13, 0xf9, 0x0f, 0x01};
		memcpy(cam->control_buffer, buffer, sizeof(buffer));
		usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
			0x04, 0x40, 0x19, 0x0000, cam->control_buffer,
			4, M5602_URB_MSG_TIMEOUT);
	}

	PDEBUG(DBG_V4L2, "Transfer started");

	return retval;

free_urbs:
	for (i = 0; (i < M5602_URBS) && cam->urb[i]; i++)
		usb_free_urb(cam->urb[i]);

free_buffers:
	for (i = 0; (i < M5602_URBS) && cam->isoc_in_buffer[i]; i++) {
		kfree(cam->isoc_in_buffer[i]);
		cam->isoc_in_buffer[i] = NULL;
	}

	PDEBUG(DBG_V4L2, "Failed to start transfer");
	return retval;
}

static int m5602_stop_transfer(struct m5602_camera *cam)
{
	int i, retval = 0;

	PDEBUG(DBG_V4L2, "Freeing isochronous buffers and urbs");

	if (cam->state & DEV_DISCONNECTED)
		return -ENODEV;

	for (i = M5602_URBS - 1; i >= 0; i--) {
		PDEBUG(DBG_V4L2, "Freeing urb %d", i);
		if (cam->urb[i]) {
			usb_kill_urb(cam->urb[i]);
			usb_free_urb(cam->urb[i]);
			cam->urb[i] = NULL;
		}

		if (cam->isoc_in_buffer[i]) {
			kfree(cam->isoc_in_buffer[i]);
			cam->isoc_in_buffer[i] = NULL;
		}
	}

	PDEBUG(DBG_V4L2, "Isochronous buffers and urbs freed");

	usb_set_interface(cam->udev, 0, 0); /* 0 Mb/s */

	PDEBUG(DBG_V4L2, "Transfer stopped");
	return retval;
}

static int m5602_stream_interrupt(struct m5602_camera *cam)
{
	int timeout;

	cam->stream = STREAM_INTERRUPT;

	timeout = wait_event_timeout(cam->wait_stream,
		(cam->stream == STREAM_OFF) ||
		(cam->state & DEV_DISCONNECTED), M5602_URB_TIMEOUT);

	if (cam->state & DEV_DISCONNECTED)
		return -ENODEV;

	if (cam->stream != STREAM_OFF) {
		cam->state |= DEV_MISCONFIGURED;
		PDEBUG(DBG_V4L2, "The camera is misconfigured. To use it, "
				"close and open /dev/video%d again.",
				cam->vdev->minor);
		return -EIO;
	}
	return 0;
}

static void m5602_vm_open(struct vm_area_struct *vma)
{
	struct m5602_frame_t *f = vma->vm_private_data;
	f->vma_use_count++;
}

static void m5602_vm_close(struct vm_area_struct *vma)
{
	struct m5602_frame_t *f = vma->vm_private_data;
	f->vma_use_count--;
}

static struct vm_operations_struct m5602_vm_ops = {
	.open = m5602_vm_open,
	.close = m5602_vm_close
};

static int m5602_v4l_open(struct inode *inode, struct file *filp)
{
	struct m5602_camera *cam;
	int err = 0;

	if (!down_read_trylock(&m5602_disconnect))
		return -ERESTARTSYS;

	cam = video_get_drvdata(video_devdata(filp));

	info("%s on /dev/video%d opened", cam->name, cam->vdev->minor);

	if (mutex_lock_interruptible(&cam->dev_mutex)) {
		PDEBUG(DBG_INIT, "Could not grab cam mutex");
		up_read(&m5602_disconnect);
		return -ERESTARTSYS;
	}

	if (cam->state & DEV_DISCONNECTED) {
		PDEBUG(DBG_INIT, "Device disconnected");
		err = -ENODEV;
		goto out;
	}

	if (cam->users) {
		PDEBUG(DBG_V4L2, "%s on /dev/video%d has "
			"currently %d users", cam->name, cam->vdev->minor, cam->users);

		if ((filp->f_flags & O_NONBLOCK) ||
			(filp->f_flags & O_NDELAY)) {
			err = -EWOULDBLOCK;
			goto out;
		}

		PDEBUG(DBG_INIT, "A blocking open() has been requested. "
				"Waiting for the device to be released");

		mutex_unlock(&cam->dev_mutex);

		err = wait_event_interruptible_exclusive(cam->open,
			(cam->state & DEV_DISCONNECTED) || (cam->users == 0));
		if (err) {
			up_read(&m5602_disconnect);
			return err;
		}

		if (cam->state & DEV_DISCONNECTED) {
			up_read(&m5602_disconnect);
			return -ENODEV;
		}

		mutex_lock(&cam->dev_mutex);
	}

	if (cam->state & DEV_MISCONFIGURED) {
		PDEBUG(DBG_INIT, "Camera is misconfigured, reinitializing");
		err = m5602_initialize_camera(cam);
		if (err) {
			PDEBUG(DBG_INIT, "Initialization failed again");
			err = -EBUSY;
			goto out;
		}
		cam->state &= ~DEV_MISCONFIGURED;
	}

	filp->private_data = cam;
	cam->users++;
	cam->io = 0;
	cam->stream = STREAM_OFF;
	cam->nbuffers = 0;
	cam->frame_count = 0;

	m5602_empty_framequeues(cam);

	PDEBUG(DBG_V4L2, "Video device /dev/video%d is open", cam->vdev->minor);

out:
	mutex_unlock(&cam->dev_mutex);
	up_read(&m5602_disconnect);
	return err;
}

static int m5602_v4l_release(struct inode *inode, struct file *filp)
{
	struct m5602_camera *cam = video_get_drvdata(video_devdata(filp));

	mutex_lock(&cam->dev_mutex); /* prevent disconnect() to be called */

	if (cam->stream == STREAM_ON)
		m5602_stream_interrupt(cam);

	m5602_stop_transfer(cam);
	m5602_empty_framequeues(cam);
	m5602_release_buffers(cam);

	cam->users--;
	wake_up_interruptible_nr(&cam->open, 1);

	info("%s on /dev/video%d closed", cam->name, cam->vdev->minor);
	mutex_unlock(&cam->dev_mutex);

	return 0;
}

static int m5602_v4l_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct m5602_camera *cam = video_get_drvdata(video_devdata(filp));
	unsigned long size = vma->vm_end - vma->vm_start, start = vma->vm_start;
	void *pos;
	u32 i;

	PDEBUG(DBG_V4L2, "Mmap call, size: %ld, buffer length %ld",
	       size, PAGE_ALIGN(cam->frame[0].buf.length));

	if (mutex_lock_interruptible(&cam->fileop_mutex))
		return -ERESTARTSYS;

	if (cam->state & DEV_DISCONNECTED) {
		PDEBUG(DBG_V4L2, "Device not present");
		mutex_unlock(&cam->fileop_mutex);
		return -ENODEV;
	}

	if (cam->state & DEV_MISCONFIGURED) {
		PDEBUG(DBG_V4L2, "The camera is misconfigured");
		mutex_unlock(&cam->fileop_mutex);
		return -EIO;
	}

	if (!(vma->vm_flags & (VM_WRITE | VM_READ))) {
		PDEBUG(DBG_V4L2, "Invalid mmap flags");
		mutex_unlock(&cam->fileop_mutex);
		return -EACCES;
	}

	if (cam->io != V4L2_MEMORY_MMAP) {
		PDEBUG(DBG_V4L2, "Camera not configured for mmap");
		mutex_unlock(&cam->fileop_mutex);
		return -EINVAL;
	}

	if (size != PAGE_ALIGN(cam->frame[0].buf.length)) {
		PDEBUG(DBG_V4L2, "Size %ld differs from buffer length %ld",
		       size, PAGE_ALIGN(cam->frame[0].buf.length));
		mutex_unlock(&cam->fileop_mutex);
		return -EINVAL;
	}

	for (i = 0; i < cam->nbuffers; i++) {
		if ((cam->frame[i].buf.m.offset >> PAGE_SHIFT) == vma->vm_pgoff)
			break;
	}

	if (i == cam->nbuffers) {
		mutex_unlock(&cam->fileop_mutex);
		return -EINVAL;
	}

	/* VM_IO is eventually going to replace PageReserved altogether */
	vma->vm_flags |= VM_IO | VM_RESERVED; /* avoid to swap out this VMA */
	pos = cam->frame[i].rawdata;

	while (size > 0) { /* size is page-aligned */
		if (vm_insert_page(vma, start, vmalloc_to_page(pos))) {
			PDEBUG(DBG_V4L2, "Failed to insert a vm page");
			mutex_unlock(&cam->fileop_mutex);
			return -EAGAIN;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	vma->vm_ops = &m5602_vm_ops;
	vma->vm_private_data = &cam->frame[i];
	m5602_vm_open(vma);

	mutex_unlock(&cam->fileop_mutex);

	PDEBUG(DBG_V4L2, "Mmap call succeded");

	return 0;
}

static struct m5602_capture_format available_fmts[] =
{
	{V4L2_PIX_FMT_SBGGR8, "BAYER"}
};

/* Checks supplied v4l2_pix_formats returns 1 on ok, 0 on fail */
static int check_format(struct v4l2_format *fmt)
{
	if (!fmt)
		return 0;

	if (fmt->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return 0;

	PDEBUG(DBG_V4L2, "Trying to set pixel format 0x%08x [%c%c%c%c], "
			"width %d, height %d",
			fmt->fmt.pix.pixelformat,
			fmt->fmt.pix.pixelformat & 0xff,
			(fmt->fmt.pix.pixelformat >> 8) & 0xff,
			(fmt->fmt.pix.pixelformat >> 24) & 0xff,
			(fmt->fmt.pix.pixelformat >> 16) & 0xff,
			fmt->fmt.pix.width,
			fmt->fmt.pix.height);

	if ((fmt->fmt.pix.width != M5602_DEFAULT_FRAME_WIDTH) ||
	    (fmt->fmt.pix.height != M5602_DEFAULT_FRAME_HEIGHT)) {
		fmt->fmt.pix.width = M5602_DEFAULT_FRAME_WIDTH;
		fmt->fmt.pix.height = M5602_DEFAULT_FRAME_HEIGHT;
	}

	switch (fmt->fmt.pix.pixelformat) {
	case V4L2_PIX_FMT_SBGGR8:
		fmt->fmt.pix.priv = 1;
		break;

	default:
		PDEBUG(DBG_V4L2, "Unsupported pixel format");
		return 0;
	}

	fmt->fmt.pix.sizeimage = fmt->fmt.pix.width *
		fmt->fmt.pix.height *
		fmt->fmt.pix.priv;
	fmt->fmt.pix.field = V4L2_FIELD_NONE;
	fmt->fmt.pix.bytesperline =
		fmt->fmt.pix.sizeimage / fmt->fmt.pix.height;
	fmt->fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;

	PDEBUG(DBG_V4L2, "Pixel format ok");
	return 1;
}

static int m5602_v4l_ioctl(struct inode *inode,
			   struct file *file,
			   unsigned int cmd,
			   unsigned long arg)
{
	struct m5602_camera *cam = video_get_drvdata(video_devdata(file));
	int err;

	if (mutex_lock_interruptible(&cam->fileop_mutex))
		return -ERESTARTSYS;

	if (cam->state & DEV_DISCONNECTED) {
		PDEBUG(DBG_V4L2, "Device not present");
		mutex_unlock(&cam->fileop_mutex);
		return -ENODEV;
	}

	if (cam->state & DEV_MISCONFIGURED) {
		PDEBUG(DBG_V4L2, "The camera is misconfigured. "
				"Close and open it again");
		mutex_unlock(&cam->fileop_mutex);
		return -EIO;
	}

	err = m5602_v4l_do_ioctl(inode, file, cmd, (void __user *) arg);

	mutex_unlock(&cam->fileop_mutex);
	return err;
}

static int m5602_v4l_do_ioctl(struct inode *inode,
			      struct file *file,
			      unsigned int ioctlnr,
			      void *arg)
{
	struct m5602_camera *cam = (struct m5602_camera *) file->private_data;
	struct m5602_frame_t *f;
	struct m5602_sensor *s;
	int retval = 0, err;

	if (!cam) {
		PDEBUG(DBG_V4L2, "%s improperly initialized", cam->name);
		return -ENODEV;
	}

	switch (ioctlnr) {
	case VIDIOC_QUERYBUF: { /* only useable with V4L2_MEMORY_MMAP */
		struct v4l2_buffer buf;

		if (copy_from_user(&buf, arg, sizeof(buf)))
			return -EFAULT;

		PDEBUG(DBG_V4L2, "VIDIOC_QUERYBUF, index %d", buf.index);

		if ((buf.type != V4L2_BUF_TYPE_VIDEO_CAPTURE) ||
			(buf.index >= cam->nbuffers) ||
			(cam->io != V4L2_MEMORY_MMAP))
			return -EINVAL;

		memcpy(&buf, &cam->frame[buf.index].buf,
			sizeof(struct v4l2_buffer));

		if (cam->frame[buf.index].vma_use_count)
			buf.flags |= V4L2_BUF_FLAG_MAPPED;

		if (cam->frame[buf.index].state == FRAME_DONE)
			buf.flags |= V4L2_BUF_FLAG_DONE;
		else if (cam->frame[buf.index].state != FRAME_UNUSED)
			buf.flags |= V4L2_BUF_FLAG_QUEUED;

		if (copy_to_user(arg, &buf, sizeof(struct v4l2_buffer)))
			return -EFAULT;
		break;
	}

	case VIDIOC_QBUF: {
		struct v4l2_buffer buf;
		unsigned long lock_flags;

		if (copy_from_user(&buf, arg, sizeof(buf)))
			return -EFAULT;

		PDEBUG(DBG_V4L2, "VIDIOC_QBUF, buffer %d", buf.index);

		if ((buf.type != V4L2_BUF_TYPE_VIDEO_CAPTURE) ||
			(cam->io != V4L2_MEMORY_MMAP) ||
			(buf.index >= cam->nbuffers))
			return -EINVAL;

		if (cam->frame[buf.index].state != FRAME_UNUSED)
			return -EINVAL;

		cam->frame[buf.index].state = FRAME_QUEUED;
		cam->frame[buf.index].buf.flags |= V4L2_BUF_FLAG_QUEUED;
		cam->frame[buf.index].buf.flags &= ~V4L2_BUF_FLAG_DONE;

		spin_lock_irqsave(&cam->queue_lock, lock_flags);
		list_add_tail(&cam->frame[buf.index].frame, &cam->inqueue);
		spin_unlock_irqrestore(&cam->queue_lock, lock_flags);

		break;
	}

	case VIDIOC_DQBUF: {
		struct v4l2_buffer buf;
		unsigned long lock_flags;

		if (copy_from_user(&buf, arg, sizeof(buf)))
			return -EFAULT;

		if ((buf.type != V4L2_BUF_TYPE_VIDEO_CAPTURE) ||
			(cam->io != V4L2_MEMORY_MMAP))
			return -EINVAL;

		if (list_empty(&cam->outqueue)) {
			if (cam->stream == STREAM_OFF)
				return -EINVAL;

			if (file->f_flags & O_NONBLOCK)
				return -EAGAIN;

			PDEBUG(DBG_V4L2, "Waiting for a frame");
			wait_event_interruptible(cam->wait_frame,
					(!list_empty(&cam->outqueue)) ||
					(cam->state & DEV_DISCONNECTED) ||
					(cam->state & DEV_MISCONFIGURED));

			if (cam->state & DEV_DISCONNECTED)
				return -ENODEV;

			if (cam->state & DEV_MISCONFIGURED)
				return -EIO;

			if (list_empty(&cam->outqueue))
				return -EINVAL;
		}

		spin_lock_irqsave(&cam->queue_lock, lock_flags);
		f = list_entry(cam->outqueue.next, struct m5602_frame_t, frame);
		list_del(cam->outqueue.next);
		spin_unlock_irqrestore(&cam->queue_lock, lock_flags);

		PDEBUG(DBG_V4L2, "VIDIOC_DQBUF, frame %d", f->framenum);

		memcpy(&buf, &f->buf, sizeof(struct v4l2_buffer));

		f->state = FRAME_UNUSED;

		if (f->vma_use_count)
			buf.flags |= V4L2_BUF_FLAG_MAPPED;

		if (copy_to_user(arg, &buf, sizeof(buf)))
			return -EFAULT;
		break;
	}

	case VIDIOC_QUERYCAP: {
		struct v4l2_capability cap;

		PDEBUG(DBG_V4L2, "VIDIOC_QUERYCAP");
		memset(&cap, 0, sizeof(struct v4l2_capability));
		strlcpy(cap.driver, "m5602", sizeof(cap.driver));

		cap.capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING;
		strlcpy(cap.card, cam->vdev->name, sizeof(cap.card));

		if (usb_make_path(cam->udev, cap.bus_info,
		    sizeof(cap.bus_info)) < 0)
			strlcpy(cap.bus_info,
				cam->vdev->name, sizeof(cap.bus_info));

		if (copy_to_user(arg, &cap, sizeof(cap)))
			return -EFAULT;
		break;
	}

	/* CAPTURE IOCTL */
	case VIDIOC_ENUM_FMT: {
		struct v4l2_fmtdesc fmtd;
		PDEBUG(DBG_V4L2, "VIDIOC_ENUM_FMT");

		if (copy_from_user(&fmtd, arg, sizeof(fmtd)))
			return -EFAULT;

		if (fmtd.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
			return -EINVAL;

		if (fmtd.index >= (sizeof(available_fmts)
				  / sizeof(m5602_capture_format)))
			return -EINVAL;

		fmtd.pixelformat = available_fmts[fmtd.index].pixelformat;
		strlcpy(fmtd.description,
			available_fmts[fmtd.index].description,
			sizeof(fmtd.description));
		memset(fmtd.reserved, 0, sizeof(fmtd.reserved));
		fmtd.flags = 0;

		if (copy_to_user(arg, &fmtd, sizeof(fmtd)))
			return -EFAULT;
		break;
	}

	case VIDIOC_G_FMT: {
		struct v4l2_format fmt;

		PDEBUG(DBG_V4L2, "VIDIOC_G_FMT");
		if (copy_from_user(&fmt, arg, sizeof(fmt)))
			return -EFAULT;

		fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		memcpy(&(fmt.fmt.pix), &cam->sensor->current_fmt,
			 sizeof(struct v4l2_pix_format));

		if (copy_to_user(arg, &fmt, sizeof(fmt)))
			return -EFAULT;
		break;
	}

	case VIDIOC_TRY_FMT:
	case VIDIOC_S_FMT: {
		struct v4l2_format fmt;

		if (ioctlnr == VIDIOC_TRY_FMT)
			PDEBUG(DBG_V4L2, "VIDIOC_TRY_FMT");
		else
			PDEBUG(DBG_V4L2, "VIDIOC_S_FMT");

		if (copy_from_user(&fmt, arg, sizeof(fmt)))
			return -EFAULT;

		if (!check_format(&fmt))
			return -EINVAL;

		if (ioctlnr == VIDIOC_S_FMT) {
			if (cam->stream == STREAM_ON) {
				err = m5602_stream_interrupt(cam);
				if (err)
					return err;
			}

			memcpy(&cam->sensor->current_fmt, &fmt.fmt.pix,
				sizeof(struct v4l2_pix_format));
			m5602_empty_framequeues(cam);
		}

		if (copy_to_user(arg, &fmt, sizeof(fmt)))
			return -EFAULT;
		break;
	}

	/* INPUT SWITCHING */
	case VIDIOC_ENUMINPUT: {
		struct v4l2_input i;

		PDEBUG(DBG_V4L2, "VIDIOC_ENUMINPUT");
		if (copy_from_user(&i, arg, sizeof(i)))
			return -EFAULT;

		/* FIXME: Add support for a non-zero index */
		if (i.index)
			return -EINVAL;

		memset(&i, 0, sizeof(i));
		strlcpy(i.name, "USB", sizeof(i.name));
		i.type = V4L2_INPUT_TYPE_CAMERA;
		i.std = 0;

		if (copy_to_user(arg, &i, sizeof(i)))
			return -EFAULT;
		break;
	}

	case VIDIOC_G_INPUT: {
		int index = 0;

		PDEBUG(DBG_V4L2, "VIDIOC_G_INPUT");

		if (copy_to_user(arg, &index, sizeof(index)))
			return -EFAULT;
		break;
	}

	case VIDIOC_S_INPUT: {
		struct v4l2_input i;

		PDEBUG(DBG_V4L2, "VIDIOC_S_INPUT");
		if (copy_from_user(&i, arg, sizeof(i)))
			return -EFAULT;

		if (i.index != 0)
			return -EINVAL;

		break;
	}

	case VIDIOC_REQBUFS: {
		struct v4l2_requestbuffers rb;
		int i;

		PDEBUG(DBG_V4L2, "VIDIOC_REQBUFS");

		if (copy_from_user(&rb, arg, sizeof(rb)))
			return -EFAULT;

		if ((rb.type != V4L2_BUF_TYPE_VIDEO_CAPTURE) ||
			((rb.memory != V4L2_MEMORY_MMAP)))
			return -EINVAL;

		for (i = 0; i < cam->nbuffers; i++) {
			if (cam->frame[i].vma_use_count) {
				PDEBUG(DBG_V4L2, "VIDIOC_REQBUFS failed. "
					"Previous buffer are still mapped");
				return -EBUSY;
			}
		}

		if (cam->stream == STREAM_ON) {
			err = m5602_stream_interrupt(cam);
			if (err)
				return err;
		}

		m5602_empty_framequeues(cam);
		m5602_release_buffers(cam);

		if (rb.count)
			rb.count = m5602_request_buffers(cam, rb.count);

		if (rb.count == 0)
			return -EFAULT;

		if (copy_to_user(arg, &rb, sizeof(rb))) {
			m5602_release_buffers(cam);
			cam->io = 0;
			return -EFAULT;
		}

		cam->io = V4L2_MEMORY_MMAP;
		break;
	}

	case VIDIOC_STREAMON: {
		int type;

		PDEBUG(DBG_V4L2, "VIDIOC_STREAMON");

		if (copy_from_user(&type, arg, sizeof(type)))
			return -EFAULT;

		if ((type != V4L2_BUF_TYPE_VIDEO_CAPTURE) ||
			(cam->io != V4L2_MEMORY_MMAP))
			return -EINVAL;

		err = m5602_start_transfer(cam);
		if (err)
			break;

		cam->stream = STREAM_ON;
		break;
	}

	case VIDIOC_STREAMOFF: {
		int type;
		PDEBUG(DBG_V4L2, "VIDIOC_STREAMOFF");

		if (copy_from_user(&type, arg, sizeof(type)))
			return -EFAULT;

		if ((type != V4L2_BUF_TYPE_VIDEO_CAPTURE) ||
			(cam->io != V4L2_MEMORY_MMAP))
			return -EINVAL;

		if (cam->stream == STREAM_ON) {
			err = m5602_stream_interrupt(cam);
			if (err)
				return err;
		}

		m5602_empty_framequeues(cam);
		break;
	}

	/* STREAMING PARAMS */
	case VIDIOC_G_PARM: {
		struct v4l2_streamparm sp;

		if (copy_from_user(&sp, arg, sizeof(sp)))
			return -EFAULT;

		PDEBUG(DBG_V4L2, "VIDIOC_G_PARM %d", sp.type);

		if (sp.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
			return -EINVAL;

		sp.parm.capture.capability = 0;
		sp.parm.capture.capturemode = 0;
		sp.parm.capture.timeperframe.numerator = 0;
		sp.parm.capture.timeperframe.denominator = 0;
		sp.parm.capture.readbuffers = 1;
		sp.parm.capture.extendedmode = 0;
		memset(&sp.parm.capture.reserved, 0,
			sizeof(sp.parm.capture.reserved));

		if (copy_to_user(arg, &sp, sizeof(sp)))
			return -EFAULT;
		break;
	}

	/* CONTROLS */
	case VIDIOC_QUERYCTRL: {
		struct v4l2_queryctrl qc;
		int i, n;

		if (copy_from_user(&qc, arg, sizeof(qc)))
			return -EFAULT;

		s = cam->sensor;
		PDEBUG(DBG_V4L2, "VIDIOC_QUERYCTRL");

		n = sizeof(s->qctrl) / sizeof(s->qctrl[0]);

		retval = -EINVAL;
		for (i = 0; i < n; i++) {
			if (qc.id && qc.id == s->qctrl[i].id) {
				memcpy(&qc, &(s->qctrl[i]), sizeof(qc));
				if (copy_to_user(arg, &qc, sizeof(qc)))
					return -EFAULT;
				retval = 0;
				break;
			}
		}
		break;
	}

	case VIDIOC_G_CTRL: {
		struct v4l2_control ctrl;

		PDEBUG(DBG_V4L2, "VIDIOC_G_CTRL");
		s = cam->sensor;

		if (!s->get_ctrl)
			return -EINVAL;

		if (copy_from_user(&ctrl, arg, sizeof(ctrl)))
			return -EFAULT;

		err = s->get_ctrl(cam, &ctrl);
		if (err)
			return err;

		if (copy_to_user(arg, &ctrl, sizeof(ctrl)))
			return -EFAULT;
		break;
	}

	case VIDIOC_S_CTRL_OLD:
	case VIDIOC_S_CTRL: {
		struct v4l2_control ctrl;
		int i, n;

		PDEBUG(DBG_V4L2, "VIDIOC_S_CTRL");
		s = cam->sensor;

		if (!s->set_ctrl)
			return -EINVAL;

		if (copy_from_user(&ctrl, arg, sizeof(ctrl)))
			return -EFAULT;

		n = sizeof(s->qctrl) / sizeof(s->qctrl[0]);

		for (i = 0; i < n; i++) {
			if (ctrl.id == s->qctrl[i].id) {
				if (s->qctrl[i].flags & V4L2_CTRL_FLAG_DISABLED)
					return -EINVAL;

				if ((ctrl.value < s->qctrl[i].minimum) ||
					(ctrl.value > s->qctrl[i].maximum))
					return -ERANGE;

				ctrl.value -= ctrl.value % s->qctrl[i].step;
				break;
			}
		}

		err = s->set_ctrl(cam, &ctrl);
		if (err)
			return err;
		break;
	}

	case VIDIOC_CROPCAP: {
		struct v4l2_cropcap cc;

		PDEBUG(DBG_V4L2, "VIDIOC_CROPCAP");

		if (copy_from_user(&cc, arg, sizeof(cc)))
			return -EFAULT;

		if (cc.type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
			return -EINVAL;

		cc.pixelaspect.numerator = 1;
		cc.pixelaspect.denominator = 1;
		cc.bounds.top = 0;
		cc.bounds.left = 0;
		cc.bounds.width = M5602_DEFAULT_FRAME_WIDTH;
		cc.bounds.height = M5602_DEFAULT_FRAME_HEIGHT;
		cc.defrect.top = 0;
		cc.defrect.left = 0;
		cc.defrect.width = M5602_DEFAULT_FRAME_WIDTH;
		cc.defrect.height = M5602_DEFAULT_FRAME_HEIGHT;

		if (copy_to_user(arg, &cc, sizeof(cc)))
			return -EFAULT;
		break;
	}

	case VIDIOC_G_CROP:
	case VIDIOC_S_CROP:
		return -EINVAL;

	case VIDIOC_G_STD:
		PDEBUG(DBG_V4L2, "VIDIOC_G_STD");
		return -EINVAL;

	case VIDIOC_S_STD:
		PDEBUG(DBG_V4L2, "VIDIOC_S_STD");
		return -EINVAL;

	case VIDIOC_QUERYSTD:
		PDEBUG(DBG_V4L2, "VIDIOC_S_STD");
		return -EINVAL;

	case VIDIOC_ENUMSTD:
		PDEBUG(DBG_V4L2, "VIDIOC_ENUMSTD");
		return -EINVAL;

	case VIDIOC_QUERYMENU:
		PDEBUG(DBG_V4L2, "VIDIOC_QUERYMENU");
		return -EINVAL;

	case VIDIOC_ENUM_FRAMEINTERVALS:
		PDEBUG(DBG_V4L2, "VIDIOC_ENUM_FRAMEINTERVALS");
		return -EINVAL;

	default:
		PDEBUG(DBG_V4L2, "Unhandled IOCTL: 0x%x!", ioctlnr);
		retval = -EINVAL;
		break;
	}
	return retval;
}
