/**
 * @file microdia-decoder.c
 * @author Nicolas VIVIEN
 * @date 2008-02-01
 * @version v0.0.0
 *
 * @brief Driver for Microdia USB video camera
 *
 * @note Copyright (C) Nicolas VIVIEN
 *
 * @par Licences
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/kref.h>

#include <linux/usb.h>
#include <media/v4l2-common.h>

#include "microdia.h"
#include "sn9c20x.h"

#define MAX(a, b)	((a) > (b) ? (a) : (b))
#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define CLIP(a, low, high) MAX((low), MIN((high), (a)))

void raw6270_2i420(uint8_t *, uint8_t *, int,
			int, const int, const int);

void raw6270_2RGB24(uint8_t *raw, uint8_t *rgb, int width,
			int height, const int hflip,
			const int vflip);

void raw6270_2BGR24(uint8_t *raw, uint8_t *rgb, int width,
			int height, const int hflip,
			const int vflip);

void microdia_getraw(uint8_t *, uint8_t *, int);

void microdia_raw2i420(uint8_t *, uint8_t *, int,
			int, const int, const int);

void microdia_raw2bgr24(uint8_t *, uint8_t *, int,
			int, const int, const int);
/**
 * @brief Decompress a frame
 *
 * This function permits to decompress a frame from the video stream.
 *
 * @param dev Device structure
 * @param buffer Image data
 *
 * @returns 0 if all is OK
 */
int microdia_decompress(struct usb_microdia *dev, struct v4l2_buffer *buffer)
{
	int factor;
	int vflip;
	int hflip;
	int width, height;
	int ret = 0;

	void *data;
	void *image;

	if (dev == NULL) {
		ret = -EFAULT;
		goto done;
	}

	if (dev->flip_detect)
		dev->flip_detect(dev);

	if (dev->set_hvflip) {
		hflip = 0;
		vflip = 0;
	} else {
		hflip = dev->vsettings.hflip;
		vflip = dev->vsettings.vflip;
	}

	image  = dev->queue.scratch;

	data = dev->queue.mem + buffer->m.offset;
	memcpy(image, data, MICRODIA_FRAME_SIZE);

	factor = 1;

	width = dev->vsettings.format.width;
	height = dev->vsettings.format.height;

	switch (dev->vsettings.format.pixelformat) {
	case V4L2_PIX_FMT_RGB24:
		if (dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_6270_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_627B_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_6288_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_62B3_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_62BB_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_145F_VID, USB_013D_PID)) {
			raw6270_2RGB24(image, data, width,
				       height, hflip, vflip);
		}
		break;
	case V4L2_PIX_FMT_BGR24:
		if (dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_6270_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_627B_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_6288_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_62B3_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_62BB_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_145F_VID, USB_013D_PID)) {
			raw6270_2BGR24(image, data, width,
				       height, hflip, vflip);
		} else {
			microdia_raw2bgr24(image, data, width,
					   height, hflip, vflip);
		}
		break;
	case V4L2_PIX_FMT_YUV420:
		if (dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_6270_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_627B_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_6288_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_62B3_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_0C45_VID, USB_62BB_PID) ||
		    dev->webcam_model ==
		    CAMERA_MODEL(USB_145F_VID, USB_013D_PID)) {
			raw6270_2i420(image, data, width,
				      height, hflip, vflip);
		} else {
			microdia_raw2i420(image, data, width,
					  height, hflip, vflip);
		}
		break;
	case V4L2_PIX_FMT_YUYV:
	default:
		break;
	}

	buffer->bytesused = dev->vsettings.format.sizeimage;
done:
	return ret;
}

/**
 * @brief This function permits to get the raw data. (without treatments)
 *
 * @param bayer Buffer with the bayer data
 * @param size Length of bayer buffer
 *
 * @retval raw Buffer with the data from video sensor
 */
void microdia_getraw(uint8_t *bayer, uint8_t *raw,
		int size) {
	memcpy(raw, bayer, size);
}

/* Table to translate Y offset to UV offset */
static int UVTranslate[32] =	{
				0, 1, 2, 3,
				8, 9, 10, 11,
				16, 17, 18, 19,
				24, 25, 26, 27,
				4, 5, 6, 7,
				12, 13, 14, 15,
				20, 21, 22, 23,
				28, 29, 30, 31
				};

static int Y_coords_624x[128][2] = {
{ 0,  0}, { 1,  0}, { 2,  0}, { 3,  0}, { 4,  0}, { 5,  0}, { 6,  0}, { 7,  0},
{ 0,  1}, { 1,  1}, { 2,  1}, { 3,  1}, { 4,  1}, { 5,  1}, { 6,  1}, { 7,  1},
{ 0,  2}, { 1,  2}, { 2,  2}, { 3,  2}, { 4,  2}, { 5,  2}, { 6,  2}, { 7,  2},
{ 0,  3}, { 1,  3}, { 2,  3}, { 3,  3}, { 4,  3}, { 5,  3}, { 6,  3}, { 7,  3},

{ 0,  4}, { 1,  4}, { 2,  4}, { 3,  4}, { 4,  4}, { 5,  4}, { 6,  4}, { 7,  4},
{ 0,  5}, { 1,  5}, { 2,  5}, { 3,  5}, { 4,  5}, { 5,  5}, { 6,  5}, { 7,  5},
{ 0,  6}, { 1,  6}, { 2,  6}, { 3,  6}, { 4,  6}, { 5,  6}, { 6,  6}, { 7,  6},
{ 0,  7}, { 1,  7}, { 2,  7}, { 3,  7}, { 4,  7}, { 5,  7}, { 6,  7}, { 7,  7},

{ 8,  0}, { 9,  0}, {10,  0}, {11,  0}, {12,  0}, {13,  0}, {14,  0}, {15,  0},
{ 8,  1}, { 9,  1}, {10,  1}, {11,  1}, {12,  1}, {13,  1}, {14,  1}, {15,  1},
{ 8,  2}, { 9,  2}, {10,  2}, {11,  2}, {12,  2}, {13,  2}, {14,  2}, {15,  2},
{ 8,  3}, { 9,  3}, {10,  3}, {11,  3}, {12,  3}, {13,  3}, {14,  3}, {15,  3},

{ 8,  4}, { 9,  4}, {10,  4}, {11,  4}, {12,  4}, {13,  4}, {14,  4}, {15,  4},
{ 8,  5}, { 9,  5}, {10,  5}, {11,  5}, {12,  5}, {13,  5}, {14,  5}, {15,  5},
{ 8,  6}, { 9,  6}, {10,  6}, {11,  6}, {12,  6}, {13,  6}, {14,  6}, {15,  6},
{ 8,  7}, { 9,  7}, {10,  7}, {11,  7}, {12,  7}, {13,  7}, {14,  7}, {15,  7}
};

void microdia_raw2bgr24(uint8_t *raw, uint8_t *rgb,
			int width, int height, const int hflip,
			const int vflip)
{
	int i = 0, x = 0, y = 0;
	unsigned char *buf = raw;
	unsigned char *buf3 = rgb;

	while (i < (width * height + (width * height) / 2)) {
		int tile = 0;
		for (tile = 0; tile < 4; tile++) {
			int subX = 0;
			int subY = 0;
			for (subY = 0; subY < 4; subY++) {
				for (subX = 0; subX < 8; subX++) {
					int subI = i + tile * 32 + 8 * subY + subX;
					int subU = i + 128 + UVTranslate[tile * 8 + 4 * (subY >> 1) + (subX >> 1)];
					int subV = subU + 32;

					int relX = x + (((tile == 0) || (tile == 1)) ? 0 : 8) + subX; /* tile 0, 1 to into left column*/
					int relY = y + (((tile == 0) || (tile == 2)) ? 0 : 4) + subY; /* tile 0, 2 go into top row */

					if (hflip)
						relX = width - relX;
					if (vflip)
						relY = height - relY;

					if ((relX < width) && (relY < height)) {
						unsigned char *ptr;
						int c, d, e;
						c = buf[subI] - 16;
						d = buf[subU] - 128;
						e = buf[subV] - 128;

						ptr = buf3 + relY * width * 3 + relX * 3;
						*ptr = CLIP((298 * c + 516 * d + 128) >> 8, 0, 255);
						ptr++;
						*ptr = CLIP((298 * c - 100 * d - 208 * e + 128) >> 8, 0, 255);
						ptr++;
						*ptr = CLIP((298 * c + 409 * e + 128) >> 8, 0, 255);
					}
				}
			}
		}

		i += 192;
		x += 16;
		if (x >= width) {
			x = 0;
			y += 8;
		}
	}
}

void microdia_raw2i420(uint8_t *raw, uint8_t *i420,
			int width, int height, const int hflip,
			const int vflip)
{
	int i = 0, x = 0, y = 0, j, relX, relY, x_div2, y_div2;
	unsigned char *buf = raw, *ptr;
	unsigned char *buf3 = i420;
	int view_size = width * height;
	int view_size_div4 = view_size >> 2;
	int image_x_div2 = width >> 1;
	int image_y_div2 = height >> 1;
	int view_x_div2 = width >> 1;

	while (i < (width * height + (width * height) / 2)) {
		for (j = 0; j < 128; j++) {
			relX = x + Y_coords_624x[j][0];
			relY = y + Y_coords_624x[j][1];
			if (hflip)
				relX = width - relX - 1;
			if (vflip)
				relY = height - relY - 1;

			if ((relX < width) && (relY < height)) {
				ptr = buf3 + relY * width + relX;
				*ptr = buf[i + j];
			}

		}
		x_div2 = x >> 1;
		y_div2 = y >> 1;
		for (j = 0; j < 32; j++) {
			relX = (x_div2) + (j & 0x07);
			relY = (y_div2) + (j >> 3);
			if (hflip)
				relX = image_x_div2 - relX - 1;
			if (vflip)
				relY = image_y_div2 - relY - 1;

			if ((relX < width) && (relY < height)) {
				ptr = buf3 + view_size +
					relY * (view_x_div2) + relX;
				*ptr = buf[i + 128 + j];
				ptr += view_size_div4;
				*ptr = buf[i + 160 + j];
			}

		}

		i += 192;
		x += 16;
		if (x >= width) {
			x = 0;
			y += 8;
		}
	}
}

/**
 *
 * @brief This function permits to convert an image from 6270 raw format to i420
 * @param raw Buffer with the bayer data
 * @param image Size of image
 * @param view Size of view
 * @param hflip Horizontal flip - not implemented
 * @param vflip Vertical flip - not implemented
 *
 * @retval i420 Buffer with the i420 data
 * Format of stream is uvyy,uvyy - repeated 320 times for 640x480
	y - repeated 640 times for 640x480
	First 1280 bytes is maybe dummy ***FIXME***
 */

void raw6270_2i420(uint8_t *raw, uint8_t *i420, int width,
			int height, const int hflip,
			const int vflip)
{
	int i, j, YIndex = 0, UVIndex = 0;
	unsigned char *y, *u, *v;
	uint8_t *buf;
	/* For fast calculations */
	int x_div_2, y_div_2;
	/* Skip first 1280 bytes strange dummy bytes */
	raw += 1280;

	x_div_2 = width / 2; /* for fast calculation if x=640 x_div_2=320 */
	y_div_2 = height / 2; /* for fast calculation if y=480 y_div_4=240 */
	buf = raw;
	y = i420;
	u = i420 + width * height;
	v = i420 + width * height + (width >> 1) * (height >> 1);

	YIndex = 0;
	UVIndex = 0;
	/* we skipped 1280 bytes, it's almost two lines */
	for (i = 0; i < y_div_2 - 1; i++) {
		for (j = 0; j < x_div_2; j++) {
			u[UVIndex] = *buf;
			buf++;
			v[UVIndex] = *buf;
			buf++;
			y[YIndex] = *buf;
			buf++;
			YIndex += 1;
			y[YIndex] = *buf;
			buf++;
			YIndex += 1;
			UVIndex += 1;
		}
		YIndex -= width;
		YIndex += width;

		for (j = 0; j < width; j++) {
			y[YIndex] = *buf;
			buf++;
			YIndex += 1;
		}
		YIndex -= width;
		YIndex += width;

		UVIndex -= x_div_2;
		UVIndex += (width >> 1);
	}
}

/**
 *
 * @brief This function permits to convert an image from 6270 rawformat to BGR24
 * @param raw Buffer with the bayer data
 * @param image Size of image
 * @param view Size of view
 * @param hflip Horizontal flip - not implemented yet
 * @param vflip Vertical flip - not implemented yet
 *
 * @retval rgb Buffer with the rgb data
	Format of stream is uvyy,uvyy - repeated 320 times for 640x480
	y - repeated 640 times for 640x480
	First 1280 bytes is maybe dumy ***FIXME***
 */

void raw6270_2BGR24(uint8_t *raw, uint8_t *rgb, int width,
			int height, const int hflip,
			const int vflip)
{
	int i, j, incX;
/* Maybe is not necesery to use long variables but I not sure in this moment */
	long y11, y12, y21, y22, u, v, y, C, D, E;
	long ScaleIncX, ScaleIncY;
	long pointerIncX, pointerIncY;
	uint8_t *bufUVYY;
	uint8_t *bufY;
	uint8_t *out_row1;
	uint8_t *out_row2;

	/* For fast calculations*/
	int x_div_2, y_div_2;
	/* Skip first 1280 bytes strange dummy bytes */
	raw += 1280;

	x_div_2 = width / 2;	/* for fast calculation if x=640 x_div_2=320 */
	y_div_2 = height / 2;	/* for fast calculation if y=480 y_div_4=240 */
	incX = 3 * width;	/* Incrementation koeficient for next row*/
	bufUVYY = raw;
	bufY = raw + 2 * width;

	/*
	// Because I can't use float ratio is multiply by 1000
	// then 1000 is equal of increment image (X or Y) with 1
	*/
	ScaleIncX = (1000 * width) / width;
	ScaleIncY = (1000 * height) / height;

	out_row1 = rgb;
	out_row2 = out_row1 + incX;
	/* we skipped 1280 bytes, it's almost two lines*/
	pointerIncY = 1000;
	for (i = 0; i < y_div_2-1; i++) {
		pointerIncX = 1000;
		for (j = 0; j < x_div_2; j++) {
			pointerIncX += ScaleIncX;

			u = (unsigned char)(*bufUVYY);
			bufUVYY++;

			v = (unsigned char)(*bufUVYY);
			bufUVYY++;

			y11 = (unsigned char)(*bufUVYY);
			bufUVYY++;

			y12 = (unsigned char)(*bufUVYY);
			bufUVYY++;

			y21 = (unsigned char)(*bufY);
			bufY++;

			y22 = (unsigned char)(*bufY);
			bufY++;

			if ((pointerIncX > 999) && (pointerIncY > 499)) {
				pointerIncX -= 1001;
				y = y11;
				C = y - 16;
				D = u - 128;
				E = v - 128;

				*out_row1 = CLIP((298 * C + 516 * D + 128) >> 8, 0, 255);
				out_row1++;
				*out_row1 = CLIP((298 * C - 100 * D - 208 * E + 128) >> 8, 0, 255);
				out_row1++;
				*out_row1 = CLIP((298 * C + 409 * E + 128) >> 8, 0, 255);
				out_row1++;

				y = y12;

				C = y - 16;
				D = u - 128;
				E = v - 128;


				*out_row1 = CLIP((298 * C + 516 * D + 128) >> 8, 0, 255);
				out_row1++;
				*out_row1 = CLIP((298 * C - 100 * D - 208 * E + 128) >> 8, 0, 255);
				out_row1++;
				*out_row1 = CLIP((298 * C + 409 * E + 128) >> 8, 0, 255);
				out_row1++;

				/*
				// Second row of stream is displayed only
				// if image is greath than half
				*/
				if (ScaleIncY > 501) {
					y = y21;

					C = y - 16;
					D = u - 128;
					E = v - 128;

					*out_row2 = CLIP((298 * C + 516 * D + 128) >> 8, 0, 255);
					out_row2++;
					*out_row2 = CLIP((298 * C - 100 * D - 208 * E + 128) >> 8, 0, 255);
					out_row2++;
					*out_row2 = CLIP((298 * C + 409 * E + 128) >> 8, 0, 255);
					out_row2++;

					y = y22;

					C = y - 16;
					D = u - 128;
					E = v - 128;

					*out_row2 = CLIP((298 * C + 516 * D + 128) >> 8, 0, 255);
					out_row2++;
					*out_row2 = CLIP((298 * C - 100 * D - 208 * E + 128) >> 8, 0, 255);
					out_row2++;
					*out_row2 = CLIP((298 * C + 409 * E + 128) >> 8, 0, 255);
					out_row2++;
				}
			}
		}
		/*
		// this comparation can be optimized and moved 	in
		// if((pointerIncX>999)&&(pointerIncY>499))
		*/
		if (pointerIncY > 499)
			pointerIncY -= 500;
		/*
		// Use 2 rows only if
		// vertical ratio > 0.5 (increment is ratio*1000)
		*/
		if (ScaleIncY > 501) {
			out_row1 = out_row1 + incX;
			out_row2 = out_row2 + incX;
		}

		pointerIncY += ScaleIncY;
		bufUVYY = bufUVYY + width;
		bufY = bufY + 2 * width;
	}
}

/**
 *
 * @brief This function permits to convert an image from 6270 rawformat to RGB24
 * @brief The same function as convert to BGR24 but only B and R is change order
 * @param raw Buffer with the bayer data
 * @param image Size of image
 * @param view Size of view
 * @param hflip Horizontal flip
 * @param vflip Vertical flip
 *
 * @retval rgb Buffer with the rgb data
	Format of stream is uvyy,uvyy - repeated 320 times for 640x480
	y - repeated 640 times for 640x480
	First 1280 bytes is maybe dumy ***FIXME***
 */

void raw6270_2RGB24(uint8_t *raw, uint8_t *rgb, int width,
			int height, const int hflip,
			const int vflip)
{
	int i, j, incX;
/* Maybe is not necesery to use long variables but I not sure in this moment */
	long y11, y12, y21, y22, u, v, y, C, D, E;
	long ScaleIncX, ScaleIncY;
	long pointerIncX, pointerIncY;
	uint8_t *bufUVYY;
	uint8_t *bufY;
	uint8_t *out_row1;
	uint8_t *out_row2;

	/* For fast calculations */
	int x_div_2, y_div_2;
	/* Skip first 1280 bytes strange dummy bytes */
	raw += 1280;

	x_div_2 = width / 2;	/* for fast calculation if x=640 x_div_2=320 */
	y_div_2 = height / 2;	/* for fast calculation if y=480 y_div_4=240 */
	incX = 3 * width;
	bufUVYY = raw;
	bufY = raw + 2 * width;

	/*
	// Because I can't use float ratio is multiply by 1000 then
	// 1000 is equal of increment image (X or Y) with 1
	*/
	ScaleIncX = (1000 * width) / width;
	ScaleIncY = (1000 * height) / height;

	out_row1 = rgb;
	out_row2 = out_row1 + incX;
	/* we skipped 1280 bytes, it's almost two lines */
	pointerIncY = 1000;
	for (i = 0; i < y_div_2-1; i++) {
		pointerIncX = 1000;
		for (j = 0; j < x_div_2; j++) {
			pointerIncX += ScaleIncX;

			u = (unsigned char)(*bufUVYY);
			bufUVYY++;

			v = (unsigned char)(*bufUVYY);
			bufUVYY++;

			y11 = (unsigned char)(*bufUVYY);
			bufUVYY++;

			y12 = (unsigned char)(*bufUVYY);
			bufUVYY++;

			y21 = (unsigned char)(*bufY);
			bufY++;

			y22 = (unsigned char)(*bufY);
			bufY++;

			if ((pointerIncX > 999) && (pointerIncY > 499)) {
				pointerIncX -= 1001;
				y = y11;
				C = y - 16;
				D = u - 128;
				E = v - 128;

				*out_row1 = CLIP((298 * C + 409 * E + 128) >> 8, 0, 255);
				out_row1++;
				*out_row1 = CLIP((298 * C - 100 * D - 208 * E + 128) >> 8, 0, 255);
				out_row1++;
				*out_row1 = CLIP((298 * C + 516 * D + 128) >> 8, 0, 255);
				out_row1++;

				y = y12;

				C = y - 16;
				D = u - 128;
				E = v - 128;

				*out_row1 = CLIP((298 * C + 409 * E + 128) >> 8, 0, 255);
				out_row1++;
				*out_row1 = CLIP((298 * C - 100 * D - 208 * E + 128) >> 8, 0, 255);
				out_row1++;
				*out_row1 = CLIP((298 * C + 516 * D + 128) >> 8, 0, 255);
				out_row1++;

				/*
				// Second row of stream is displayed only
				// if image is greath than half
				*/
				if (ScaleIncY > 501) {
					y = y21;

					C = y - 16;
					D = u - 128;
					E = v - 128;

					*out_row2 = CLIP((298 * C + 409 * E + 128) >> 8, 0, 255);
					out_row2++;
					*out_row2 = CLIP((298 * C - 100 * D - 208 * E + 128) >> 8, 0, 255);
					out_row2++;
					*out_row2 = CLIP((298 * C + 516 * D + 128) >> 8, 0, 255);
					out_row2++;

					y = y22;

					C = y - 16;
					D = u - 128;
					E = v - 128;

					*out_row2 = CLIP((298 * C + 409 * E + 128) >> 8, 0, 255);
					out_row2++;
					*out_row2 = CLIP((298 * C - 100 * D - 208 * E + 128) >> 8, 0, 255);
					out_row2++;
					*out_row2 = CLIP((298 * C + 516 * D + 128) >> 8, 0, 255);
					out_row2++;
				}
			}
		}
		/*
		// this comparation can be optimized and moved 	in
		// if((pointerIncX>999)&&(pointerIncY>499))
		*/
		if (pointerIncY > 499)
			pointerIncY -= 500;
/* Use 2 rows only if vertical ratio > 0.5 (increment is ratio*1000) */
		if (ScaleIncY > 501) {
			out_row1 = out_row1 + incX;
			out_row2 = out_row2 + incX;
		}

		pointerIncY += ScaleIncY;
		bufUVYY = bufUVYY + width;
		bufY = bufY + 2 * width;
	}
}

