/*
    Copyright 2006 Cedric Schieli

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/vmalloc.h>

#include "usbvideo.h"
#include "bayer.h"

#define FUNCAM_VENDOR_ID	0x0471
#define FUNCAM_PRODUCT_ID	0x0321

#define MAX_SN9C2028		4	/* How many devices we allow to connect */

/* Header signatures */

#define HDRSIG_VGA		0x00
#define HDRSIG_SIF		0x02
#define HDRSIG_QSIF		0x04

#define SN9C2028_CTRL_TIMEOUT	300
#define SN9C2028_CTRL_MAXPOLL	1000

#define SN9C2028_BUFFER_SIZE	(16+638*10)*480/8+12
#define	SN9C2028_FUNCAM	1	/* Philips Funcam DMVC300K/00 */

/* Video sizes supported */
#define VIDEOSIZE_QSIF		VIDEOSIZE(160,120)
#define VIDEOSIZE_SIF		VIDEOSIZE(320,240)
#define	VIDEOSIZE_VGA		VIDEOSIZE(640,480)

/* Video sizes supported */
enum {
	SIZE_QSIF = 0,
	SIZE_SIF,
	SIZE_VGA,
	/* Add/remove/rearrange items before this line */
	SIZE_LastItem
};

enum BufferState {
	Buffer_Empty = 0,
	Buffer_Partial,
	Buffer_Complete,
};

/*
 * This structure lives in uvd->user field.
 */
typedef struct {
	int initialized;	/* Had we already sent init sequence? */
	int camera_model;	/* What type of IBM camera we got? */
	int has_hdr;
	unsigned char control_buffer[6];
	int autoExposure;
	int underExposured;
	int overExposured;
	enum BufferState state;
	unsigned char *buffer;
	int pos;
} sn9c2028_t;
#define	SN9C2028_T(uvd)	((sn9c2028_t *)((uvd)->user_data))

static struct usbvideo *cams;

static int debug;
static int flags; /* = FLAGS_DISPLAY_HINTS | FLAGS_OVERLAY_STATS; */

static const int min_canvasWidth  = 4;
static const int min_canvasHeight = 4;

static int size = SIZE_QSIF;
static int speed = 0;

module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Debug level: 0-9 (default=0)");
module_param(flags, int, 0);
MODULE_PARM_DESC(flags, "Bitfield: 0=VIDIOCSYNC, 1=B/W, 2=show hints, 3=show stats, 4=test pattern, 5=separate frames, 6=clean frames");
module_param(size, int, 0);
MODULE_PARM_DESC(size, "Image size: 0=160x120 1=320x240 2=640x480  (default=0)");
module_param(speed, int, 0);
MODULE_PARM_DESC(speed, "Speed: 0-4 (default=0)");

MODULE_AUTHOR ("Cedric Schieli");
MODULE_DESCRIPTION ("SN9C2028 based USB Camera Driver for Linux (c) 2006");
MODULE_LICENSE("GPL");

/*
 * sn9c2028_size_to_videosize()
 *
 * This procedure converts module option 'size' into the actual
 * videosize_t that defines the image size in pixels. We need
 * simplified 'size' because user wants a simple enumerated list
 * of choices, not an infinite set of possibilities.
 */
static videosize_t sn9c2028_size_to_videosize(int size)
{
	videosize_t vs = VIDEOSIZE_QSIF;
	RESTRICT_TO_RANGE(size, 0, (SIZE_LastItem-1));
	switch (size) {
	case SIZE_QSIF:
		vs = VIDEOSIZE_QSIF;
		break;
	case SIZE_SIF:
		vs = VIDEOSIZE_SIF;
		break;
	case SIZE_VGA:
		vs = VIDEOSIZE_VGA;
		break;
	default:
		err("size=%d. is not valid", size);
		break;
	}
	return vs;
}

/* 
 * The decoding algorithm originates with Bertrik Sikken. This version adapted
 * from the one written for the webcam-osx project, for the sn9c2028.
 */

/* Four defines for bitstream operations, for the decode function */

#define PEEK_BITS(num,to) {\
        if (bitnum<num){\
                do {\
                        bitbuf=(bitbuf<<8)|(cam->buffer[cam->pos++]);\
       	                bitnum+=8;\
                }\
                while(bitnum<24);\
        }\
        to=bitbuf>>(bitnum-num);\
}

/*
 * PEEK_BITS puts the next <num> bits into the low bits of <to>. 
 * when the buffer is empty, it is completely refilled. 
 * This strategy tries to reduce memory access. Note that the high bits 
 * are NOT set to zero!
 */

#define EAT_BITS(num) { bitnum-=num; }

/*
 * EAT_BITS consumes <num> bits (PEEK_BITS does not consume anything, 
 * it just peeks)
 */

#define PARSE_PIXEL(val) {\
    PEEK_BITS(10,bits);\
    if ((bits&0x00000200)==0) { \
        EAT_BITS(1); \
    } \
    else if ((bits&0x00000380)==0x00000280) { \
        EAT_BITS(3); \
        val+=3; \
        if (val>255) val=255; \
    }\
    else if ((bits&0x00000380)==0x00000300) { \
        EAT_BITS(3); \
        val-=3; \
        if (val<0) val=0; \
    }\
    else if ((bits&0x000003c0)==0x00000200) { \
        EAT_BITS(4); \
        val+=8; \
        if (val>255) val=255;\
    }\
    else if ((bits&0x000003c0)==0x00000240) { \
        EAT_BITS(4); \
        val-=8; \
        if (val<0) val=0;\
    }\
    else if ((bits&0x000003c0)==0x000003c0) { \
        EAT_BITS(4); \
        val-=20; \
        if (val<0) val=0;\
    }\
    else if ((bits&0x000003e0)==0x00000380) { \
        EAT_BITS(5); \
        val+=20; \
        if (val>255) val=255;\
    }\
    else { \
        EAT_BITS(10); \
        val=8*(bits&0x0000001f)+0; \
    }\
}

#define PUT_PIXEL_PAIR(f,odd) {\
	if (odd) {\
		*f++ = 0;\
		*f++ = c1;\
		*f++ = 0;\
		*f++ = 0;\
		*f++ = 0;\
		*f++ = c2;\
	} else {\
		*f++ = c1;\
		*f++ = 0;\
		*f++ = 0;\
		*f++ = 0;\
		*f++ = c2;\
		*f++ = 0;\
	}\
}

int sn9c2028_ctrl_status(struct uvd* uvd) {
	int retval;
	unsigned char status;

	retval = usb_control_msg(uvd->dev, usb_rcvctrlpipe(uvd->dev, 0),
				 0, 0xc1, 1, 0, &status, 1,
				 SN9C2028_CTRL_TIMEOUT);
	if (retval != 1) {
		err("Failed to read status (%d)", retval);
		return retval;
	}
	
	return status;
}

int sn9c2028_ctrl_wait(struct uvd* uvd, int expected) {
	int i, retval;

	for (i = 0; i < SN9C2028_CTRL_MAXPOLL; i++) {
		retval = sn9c2028_ctrl_status(uvd);
		if (retval < 0 || retval == expected || retval > 2)
			break;
		msleep(5);
	}
	return retval;
}

unsigned char *sn9c2028_inout(struct uvd* uvd, unsigned char c0,
					       unsigned char c1,
					       unsigned char c2,
					       unsigned char c3,
					       unsigned char c4,
					       unsigned char c5)
{
	int retval;
	unsigned char *buf = SN9C2028_T(uvd)->control_buffer;

	//info("inout(%2x %2x %2x %2x %2x %2x)", c0, c1, c2, c3, c4, c5);

	buf[0] = c0;
	buf[1] = c1;
	buf[2] = c2;
	buf[3] = c3;
	buf[4] = c4;
	buf[5] = c5;

	retval = usb_control_msg(uvd->dev, usb_sndctrlpipe(uvd->dev, 0),
				 8, 0x41, 2, 0, buf, 6,
				 SN9C2028_CTRL_TIMEOUT);
	if (retval != 6) {
		err("Failed to send cmd (%d)", retval);
		return NULL;
	}

	retval = sn9c2028_ctrl_wait(uvd, 2);
	if (retval != 2) {
		err("Bad status (%d)", retval);
		return NULL;
	}
	
	retval = usb_control_msg(uvd->dev, usb_rcvctrlpipe(uvd->dev, 0),
				 0, 0xc1, 4, 0, buf, 4,
				 SN9C2028_CTRL_TIMEOUT);
	if (retval != 4) {
		err("Failed to read result cmd (%d)", retval);
		return NULL;
	}

	return buf;
}

unsigned char *sn9c2028_out(struct uvd* uvd, unsigned char c0,
					     unsigned char c1,
					     unsigned char c2,
					     unsigned char c3,
					     unsigned char c4,
					     unsigned char c5)
{
	int retval;
	unsigned char *buf = SN9C2028_T(uvd)->control_buffer;

	//info("out(%2x %2x %2x %2x %2x %2x)", c0, c1, c2, c3, c4, c5);

	buf[0] = c0;
	buf[1] = c1;
	buf[2] = c2;
	buf[3] = c3;
	buf[4] = c4;
	buf[5] = c5;

	retval = usb_control_msg(uvd->dev, usb_sndctrlpipe(uvd->dev, 0),
				 8, 0x41, 2, 0, buf, 6,
				 SN9C2028_CTRL_TIMEOUT);
	if (retval != 6) {
		err("Failed to send cmd (%d)", retval);
		return NULL;
	}

	retval = sn9c2028_ctrl_wait(uvd, 0);
	if (retval != 0) {
		err("Bad status (%d)", retval);
		return NULL;
	}

	return buf;
}

int sn9c2028_find_header(struct uvd *uvd, int length) {
	int pos;
	int l = length - 5;
	for (pos = 0; pos < l; pos++) {
		if ((RING_QUEUE_PEEK(&uvd->dp, pos) == 0xFF) &&
		    (RING_QUEUE_PEEK(&uvd->dp, pos + 1) == 0xFF) &&
		    (RING_QUEUE_PEEK(&uvd->dp, pos + 2) == 0x00) &&
		    (RING_QUEUE_PEEK(&uvd->dp, pos + 3) == 0xC4) &&
		    (RING_QUEUE_PEEK(&uvd->dp, pos + 4) == 0xC4) &&
		    (RING_QUEUE_PEEK(&uvd->dp, pos + 5) == 0x96)) {
			return pos;
		}
	}
	return -1;
}

/*
 * sn9c2028_ProcessIsocData()
 *
 * Generic routine to parse the ring queue data. It employs either
 * sn9c2028_find_header() or sn9c2028_parse_lines() to do most
 * of work.
 *
 */
static void sn9c2028_ProcessIsocData(struct uvd *uvd, struct usbvideo_frame *frame)
{
	sn9c2028_t *cam = SN9C2028_T(uvd);
	int pos, len, lightness;
	
	assert(cam != NULL);
	
	if (frame == NULL) {
		err("Invalid frame");
		return;
	}

redo:
	switch (cam->state) {
	case Buffer_Empty:
		len = RingQueue_GetLength(&uvd->dp);
		pos = sn9c2028_find_header(uvd, len);
		if (pos < 0) {
			if (len > 5)
				RING_QUEUE_DEQUEUE_BYTES(&uvd->dp, len - 5);
			return;
		}
		RING_QUEUE_DEQUEUE_BYTES(&uvd->dp, pos + 6);
		uvd->stats.header_count++;
		cam->pos = 0;
		cam->state = Buffer_Partial;
	case Buffer_Partial:
		len = RingQueue_GetLength(&uvd->dp);
		pos = sn9c2028_find_header(uvd, len);
		if (pos < 0) {
			if (len > 5) {
				RingQueue_Dequeue(&uvd->dp, cam->buffer + cam->pos, len - 5);
				cam->pos += (len - 5);
			}
			return;
		}
		RingQueue_Dequeue(&uvd->dp, cam->buffer + cam->pos, pos);
		cam->state = Buffer_Complete;
	case Buffer_Complete:
		if (cam->buffer[0] != 0x88) {
			if (debug > 0)
				info("Ignoring frame (%02x)", cam->buffer[0]);
			cam->state = Buffer_Empty;
			goto redo;
		}
		lightness = (cam->buffer[5] << 8) + cam->buffer[4];
		if (debug > 0)
			info("lightness %d", lightness);
		if (lightness < 8000)
			cam->underExposured++;
		else
			cam->underExposured = 0;
		if (lightness > 10000)
			cam->overExposured++;
		else
			cam->overExposured = 0;
		{
			unsigned int oldAutoExposure = cam->autoExposure;
			unsigned int correction = (200 - oldAutoExposure) / 100;
			if (cam->underExposured > 3) {
				cam->autoExposure -= correction;
				if (cam->autoExposure < 0) cam->autoExposure = 0;
			} else if (cam->overExposured > 3) {
				cam->autoExposure += correction;
				if (cam->autoExposure > 100) cam->autoExposure = 100;
			}
			if (cam->autoExposure != oldAutoExposure) {
				unsigned char *buf;
				unsigned int aExp = ((100 - cam->autoExposure) * 1000) / 100;
				cam->overExposured = 0;
				cam->underExposured = 0;
				if (debug > 0)
					info("autoExposure: %d, shutter: %04x", cam->autoExposure, aExp);
				sn9c2028_out(uvd, 0x1c, 0x09, (aExp>>8)&0xff, aExp&0xff, 0x00, 0x00);
				buf = sn9c2028_inout(uvd, 0x12, 0x27, 0x01, 0x00, 0x00, 0x00);
				if (debug > 0)
					info("%02x", buf[1]);
			}
		}
		cam->pos = 6;
		{
			unsigned char *f;
			int odd_line, trail;
			short c1, c2;
			unsigned int width, height, x, y, linesize;
			unsigned int abswidth, absheight;
			unsigned long bitbuf = 0;
			unsigned short bitnum = 0, bits;
	
			abswidth = VIDEOSIZE_X(frame->request);
			absheight = VIDEOSIZE_Y(frame->request);
			linesize = abswidth * V4L_BYTES_PER_PIXEL;

			assert(frame->data != NULL);
			f = frame->data;

			switch (size) {
			case SIZE_QSIF:
				width = 160;
				height = 120;
				break;
			case SIZE_SIF:
				width = 320;
				height = 240;
				break;
			case SIZE_VGA:
				width = 640;
				height = 480;
				break;
			default:
				err("Unsupported size (%d)", size);
				frame->frameState = FrameState_Error;
				return;
			}
			
			for (y = 0; y < height; y++) {

				if (y == absheight)
					break;
				
				odd_line = y & 0x1;

		                PEEK_BITS(8, bits);
		                EAT_BITS(8);
		                c1 = bits & 0x000000ff;

	        	        PEEK_BITS(8, bits);
	                	EAT_BITS(8);
		                c2 = bits & 0x000000ff;

				PUT_PIXEL_PAIR(f, odd_line);

				for (x = 2; x < width; x += 2)
				{
					PARSE_PIXEL(c1);
					PARSE_PIXEL(c2);

					if (x < VIDEOSIZE_X(frame->request))
						PUT_PIXEL_PAIR(f, odd_line);
				}

				trail = abswidth - x;
				if (trail > 0) {
					memset(f, 0, trail);
					f += trail;
					memset(f, 0, trail);
					f += trail;
					memset(f, 0, trail);
					f += trail;
				}
			}

			trail = absheight - y;
			if (trail > 0) {
				memset(f, 0, trail * linesize);
			}
			
			frame->seqRead_Length += ((unsigned long)f - (unsigned long)frame->data);
			frame->deinterlace = Deinterlace_None;
			frame->frameState = FrameState_Done;
			uvd->curframe = -1;
			uvd->stats.frame_num++;
			cam->state = Buffer_Empty;
			return;
		}
	}
}

static int sn9c2028_funcam_setup(struct uvd *uvd)
{
	return (CAMERA_IS_OPERATIONAL(uvd) ? 0 : -EFAULT);
}

/*
 * sn9c2028_video_stop()
 *
 * This code tells camera to stop streaming. The interface remains
 * configured and bandwidth - claimed.
 */
static void sn9c2028_video_stop(struct uvd *uvd)
{
	switch (SN9C2028_T(uvd)->camera_model) {
	case SN9C2028_FUNCAM:
		sn9c2028_out(uvd, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00);
		break;
	} /* switch */
}

/*
 * sn9c2028_reinit_iso()
 *
 * This procedure sends couple of commands to the camera and then
 * resets the video pipe. This sequence was observed to reinit the
 * camera or, at least, to initiate ISO data stream.
 *
 */
static void sn9c2028_reinit_iso(struct uvd *uvd, int do_stop)
{
	sn9c2028_t *cam;
	unsigned char *buf;
	unsigned char subsample, asic128;

	cam = SN9C2028_T(uvd);

	cam->autoExposure = 50;
	cam->underExposured = 0;
	cam->overExposured = 0;
	
	switch (size) {
	case SIZE_QSIF:
		subsample = 0x04;
		asic128 = 0x26;
		break;
	case SIZE_SIF:
		subsample = 0x02;
		asic128 = 0x16;
		break;
	case SIZE_VGA:
		subsample = 0x01;
		asic128 = 0x06;
		break;
	default:
		err("Unsupported size (%d)", size);
		return;
	}
		
	switch (cam->camera_model) {
	case SN9C2028_FUNCAM:
		if (do_stop)
			sn9c2028_video_stop(uvd);
		// setModeToPCCAM
		sn9c2028_inout(uvd, 0x0c, 0x01, 0x00, 0x00, 0x00, 0x00);
		// setSubsampling
		sn9c2028_inout(uvd, 0x16, subsample, 0x00, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x20, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x21, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x22, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x23, 0x01, 0x02, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x24, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x25, 0x01, 0x28, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x26, 0x01, 0x1e, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x28, 0x01, 0x06, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x29, 0x01, 0x20, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x2a, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x2b, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x2c, 0x01, 0x02, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x2d, 0x01, 0x03, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x2e, 0x01, 0x0f, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x2f, 0x01, 0x0c, 0x00, 0x00);
		buf = sn9c2028_inout(uvd, 0x12, 0x34, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x34, 0x01, 0xa1, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x35, 0x01, 0x00, 0x00, 0x00);
		// first row
		sn9c2028_out(uvd, 0x1c, 0x01, 0x00, 0x0a, 0x00, 0x00);
		// first column
		sn9c2028_out(uvd, 0x1c, 0x02, 0x00, 0x14, 0x00, 0x00);
		// window height
		sn9c2028_out(uvd, 0x1c, 0x03, 0x01, 0xe0, 0x00, 0x00);
		// window width
		sn9c2028_out(uvd, 0x1c, 0x04, 0x02, 0x80, 0x00, 0x00);
		// horizontal blanking
		sn9c2028_out(uvd, 0x1c, 0x05, 0x00, 0x3d, 0x00, 0x00);
		// vertical blanking
		sn9c2028_out(uvd, 0x1c, 0x06, 0x00, 0x07, 0x00, 0x00);
		// pixel data rate
		sn9c2028_out(uvd, 0x1c, 0x0a, 0x00, 0x00, 0x00, 0x00);
		// zoom
		sn9c2028_out(uvd, 0x1c, 0x1e, 0x00, 0x00, 0x00, 0x00);
		// read mode
		sn9c2028_out(uvd, 0x1c, 0x20, 0x10, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x25, 0x01, 0x28, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x26, 0x01, 0x1e, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x22, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x23, 0x01, 0x02, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x24, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x28, 0x01, asic128, 0x00, 0x00);
		sn9c2028_inout(uvd, 0x13, 0x27, 0x01, 0x20 + (speed ? (1<<(speed-1)) : 0), 0x00, 0x00);
		buf = sn9c2028_inout(uvd, 0x12, 0x27, 0x01, 0x00, 0x00, 0x00);
		// (shutter width) integration rows number
		//sn9c2028_out(uvd, 0x1c, 0x09, 0x01, 0xe0, 0x00, 0x00);
		sn9c2028_out(uvd, 0x1c, 0x09, 0x01, 0x44, 0x00, 0x00);
		buf = sn9c2028_inout(uvd, 0x12, 0x27, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_out(uvd, 0x20, 0x29, 0x30, 0x00, 0x00, 0x00);
		buf = sn9c2028_inout(uvd, 0x12, 0x27, 0x01, 0x00, 0x00, 0x00);
		sn9c2028_out(uvd, 0x21, 0x20, 0x88, 0x08, 0x00, 0x00);
//		sn9c2028_out(uvd, 0x21, 0x20, 0xff, 0x0f, 0x00, 0x00);
		sn9c2028_out(uvd, 0x20, 0x34, 0xa1, 0x00, 0x00, 0x00);
		usb_clear_halt(uvd->dev, usb_rcvisocpipe(uvd->dev, uvd->video_endp));
		break;
	}
}

static void sn9c2028_video_start(struct uvd *uvd)
{
	sn9c2028_reinit_iso(uvd, 0);
}

/*
 * Return negative code on failure, 0 on success.
 */
static int sn9c2028_setup_on_open(struct uvd *uvd)
{
	int setup_ok = 0; /* Success by default */
	/* Send init sequence only once, it's large! */
	if (!SN9C2028_T(uvd)->initialized) { /* FIXME rename */
		switch (SN9C2028_T(uvd)->camera_model) {
		case SN9C2028_FUNCAM:
			setup_ok = sn9c2028_funcam_setup(uvd);
			break;
		default:
			/* We do all setup when Isoc stream is requested */
			break;
		}
		SN9C2028_T(uvd)->buffer = vmalloc(SN9C2028_BUFFER_SIZE);
		if (SN9C2028_T(uvd)->buffer == NULL)
			setup_ok = 0;
		SN9C2028_T(uvd)->initialized = (setup_ok != 0);
	}
	SN9C2028_T(uvd)->state = Buffer_Empty;
	return setup_ok;
}

static void sn9c2028_configure_video(struct uvd *uvd)
{
	if (uvd == NULL)
		return;

	memset(&uvd->vpic, 0, sizeof(uvd->vpic));
	memset(&uvd->vpic_old, 0x55, sizeof(uvd->vpic_old));

/*
	uvd->vpic.colour = init_color << 8;
	uvd->vpic.hue = init_hue << 8;
	uvd->vpic.brightness = init_brightness << 8;
	uvd->vpic.contrast = init_contrast << 8;
	uvd->vpic.whiteness = 105 << 8;
*/
	uvd->vpic.depth = 24;
	uvd->vpic.palette = VIDEO_PALETTE_RGB24;

	memset(&uvd->vcap, 0, sizeof(uvd->vcap));
	strcpy(uvd->vcap.name, "SN9C2028 USB Camera");
	uvd->vcap.type = VID_TYPE_CAPTURE;
	uvd->vcap.channels = 1;
	uvd->vcap.audios = 0;
	uvd->vcap.maxwidth = VIDEOSIZE_X(uvd->canvas);
	uvd->vcap.maxheight = VIDEOSIZE_Y(uvd->canvas);
	uvd->vcap.minwidth = min_canvasWidth;
	uvd->vcap.minheight = min_canvasHeight;

	memset(&uvd->vchan, 0, sizeof(uvd->vchan));
	uvd->vchan.flags = 0;
	uvd->vchan.tuners = 0;
	uvd->vchan.channel = 0;
	uvd->vchan.type = VIDEO_TYPE_CAMERA;
	strcpy(uvd->vchan.name, "Camera");
}

/*
 * sn9c2028_probe()
 *
 * This procedure queries device descriptor and accepts the interface
 * if it looks like SN9C2028 based camera.
 *
 */
static int sn9c2028_probe(struct usb_interface *intf, const struct usb_device_id *devid)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	struct uvd *uvd = NULL;
	int ix, i, nas, model=0, canvasX=0, canvasY=0;
	int actInterface=-1, inactInterface=-1, maxPS=0;
	__u8 ifnum = intf->altsetting->desc.bInterfaceNumber;
	unsigned char video_ep = 0;
	char *brand = NULL;

	if (debug >= 1)
		info("sn9c2028_probe(%p,%u.)", intf, ifnum);

	/* We don't handle multi-config cameras */
	if (dev->descriptor.bNumConfigurations != 1)
		return -ENODEV;

	/* Check the version/revision */
	switch (le16_to_cpu(dev->descriptor.idVendor)) {
	case 0x0471:
		switch (le16_to_cpu(dev->descriptor.idProduct)) {
		case 0x0321:
			if (ifnum != 0)
				return -ENODEV;
			model = SN9C2028_FUNCAM;
			brand = "Philips Funcam";
			break;
		}
		break;
	default:
		return -ENODEV;
	}

	/* Print detailed info on what we found so far */
	info("%s USB camera found", brand);

	/* Validate found interface: must have five endpoints */
	nas = intf->num_altsetting;
	if (debug > 0)
		info("Number of alternate settings=%d.", nas);
	if (nas < 2) {
		err("Too few alternate settings for this camera!");
		return -ENODEV;
	}
	/* Validate all alternate settings */
	for (ix=0; ix < nas; ix++) {
		const struct usb_host_interface *interface;
		const struct usb_endpoint_descriptor *endpoint;

		interface = &intf->altsetting[ix];
		i = interface->desc.bAlternateSetting;
		if (interface->desc.bNumEndpoints != 5) {
			err("Interface %d. has %u. endpoints!",
			    ifnum, (unsigned)(interface->desc.bNumEndpoints));
			return -ENODEV;
		}
		endpoint = &interface->endpoint[0].desc;
		if (video_ep == 0)
			video_ep = endpoint->bEndpointAddress;
		else if (video_ep != endpoint->bEndpointAddress) {
			err("Alternate settings have different endpoint addresses!");
			return -ENODEV;
		}
		if ((endpoint->bmAttributes & 0x03) != 0x01) {
			err("Interface %d. has non-ISO endpoint!", ifnum);
			return -ENODEV;
		}
		if ((endpoint->bEndpointAddress & 0x80) == 0) {
			err("Interface %d. has ISO OUT endpoint!", ifnum);
			return -ENODEV;
		}
		if (le16_to_cpu(endpoint->wMaxPacketSize) == 0) {
			if (inactInterface < 0)
				inactInterface = i;
			else {
				err("More than one inactive alt. setting!");
				return -ENODEV;
			}
		} else {
			actInterface = i;
			maxPS = le16_to_cpu(endpoint->wMaxPacketSize);
			if (debug > 0)
				info("Active setting=%d. maxPS=%d.", i, maxPS);
		}
	}
	if ((maxPS <= 0) || (actInterface < 0) || (inactInterface < 0)) {
		err("Failed to recognize the camera!");
		return -ENODEV;
	}

	/* Validate options */
	switch (model) {
	case SN9C2028_FUNCAM:
		RESTRICT_TO_RANGE(size, SIZE_QSIF, SIZE_VGA);
		RESTRICT_TO_RANGE(speed, 0, 4);
		canvasX = 640;
		canvasY = 480;
		break;
	default:
		err("SN9C2028 camera: Model %d. not supported!", model);
		return -ENODEV;
	}

	uvd = usbvideo_AllocateDevice(cams);
	if (uvd != NULL) {
		/* Here uvd is a fully allocated uvd object */
		uvd->flags = flags;
		uvd->debug = debug;
		uvd->dev = dev;
		uvd->iface = ifnum;
		uvd->ifaceAltInactive = inactInterface;
		uvd->ifaceAltActive = actInterface;
		uvd->video_endp = video_ep;
		uvd->iso_packet_len = maxPS;
		uvd->paletteBits = 1L << VIDEO_PALETTE_RGB24;
		uvd->defaultPalette = VIDEO_PALETTE_RGB24;
		uvd->canvas = VIDEOSIZE(canvasX, canvasY);
		uvd->videosize = sn9c2028_size_to_videosize(size);

		/* Initialize sn9c2028-specific data */
		assert(SN9C2028_T(uvd) != NULL);
		SN9C2028_T(uvd)->camera_model = model;
		SN9C2028_T(uvd)->initialized = 0;

		sn9c2028_configure_video(uvd);

		i = usbvideo_RegisterVideoDevice(uvd);
		if (i != 0) {
			err("usbvideo_RegisterVideoDevice() failed.");
			uvd = NULL;
		}
	}
	usb_set_intfdata (intf, uvd);
	return 0;
}

void sn9c2028_bayer_interpolate(struct uvd *uvd, struct usbvideo_frame *frame)
{
	if ((uvd == NULL) || (frame == NULL))
		return;
	
	bayer_interpolate(frame->data, VIDEOSIZE_X(frame->request), VIDEOSIZE_Y(frame->request), BAYER_TILE_BGGR);
}



static struct usb_device_id id_table[] = {
	{ USB_DEVICE(FUNCAM_VENDOR_ID, FUNCAM_PRODUCT_ID) },				/* Philips Funcam aka DMVC300K/00 */
	{ }  /* Terminating entry */
};

static void sn9c2028_cam_cleanup(struct uvd *uvd) {
	vfree(SN9C2028_T(uvd)->buffer);
}

/*
 * sn9c2028_init()
 *
 * This code is run to initialize the driver.
 *
 */
static int __init sn9c2028_init(void)
{
	struct usbvideo_cb cbTbl;
	memset(&cbTbl, 0, sizeof(cbTbl));
	cbTbl.probe = sn9c2028_probe;
	cbTbl.setupOnOpen = sn9c2028_setup_on_open;
	cbTbl.videoStart = sn9c2028_video_start;
	cbTbl.videoStop = sn9c2028_video_stop;
	cbTbl.processData = sn9c2028_ProcessIsocData;
	cbTbl.postProcess = sn9c2028_bayer_interpolate;
	cbTbl.userFree = sn9c2028_cam_cleanup;
	return usbvideo_register(
		&cams,
		MAX_SN9C2028,
		sizeof(sn9c2028_t),
		"sn9c2028",
		&cbTbl,
		THIS_MODULE,
		id_table);
}

static void __exit sn9c2028_cleanup(void)
{
	usbvideo_Deregister(&cams);
}

MODULE_DEVICE_TABLE(usb, id_table);

module_init(sn9c2028_init);
module_exit(sn9c2028_cleanup);
