/*
 * Driver for SQ930B/C based webcams
 * Copyright (C) 2007 Sam Revitch <samr7@cs.washington.edu>
 *
 * This driver is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This driver is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this driver; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <linux/kernel.h>
#include <linux/usb.h>
#include <linux/i2c.h>
#include "usbcam.h"

#define USBCAM_DBG_SQ_INIT	(USBCAM_DBGBIT_MD_START + 0)
#define USBCAM_DBG_SQ_CMD	(USBCAM_DBGBIT_MD_START + 1)
#define USBCAM_DBG_SQ_MDINTF	(USBCAM_DBGBIT_MD_START + 2)
#define USBCAM_DBG_SQ_CTRL	(USBCAM_DBGBIT_MD_START + 3)


#define SQ930_VERSION KERNEL_VERSION(0,1,0)
#define SQ930_VERSION_EXTRA "-dev"

/*
 * Private control IDs, because V4L2 doesn't provide everything we need
 * for this driver.
 */
enum {
	V4L2_CID_SQ930_POWERLINE = (V4L2_CID_PRIVATE_BASE + 0),
	V4L2_CID_SQ930_AUTO_EXPOSURE,
	V4L2_CID_SQ930_PAN_ABSOLUTE,
	V4L2_CID_SQ930_TILT_ABSOLUTE,
	V4L2_CID_SQ930_PANTILT_SPEED,
	V4L2_CID_SQ930_PAN_RELATIVE = (V4L2_CID_PRIVATE_BASE + 7),
	V4L2_CID_SQ930_TILT_RELATIVE,
	V4L2_CID_SQ930_PANTILT_RESET,
};


/*
 * Internal data structures of the SQ930 driver
 */
#if defined(__KERNEL__)

#define sq_info(RV, FMT...) usbcam_info((RV)->sqd_parent, FMT)
#define sq_err(RV, FMT...) usbcam_err((RV)->sqd_parent, FMT)
#define sq_dbg(RV, SUBSYS, FMT...) usbcam_dbg((RV)->sqd_parent, SUBSYS, FMT)
#define assert(X) usbcam_assert(X)

struct sq930_dev;

struct sq930_ucbuswrite_cmd {
	u16		bw_addr;
	u8		bw_data;
};

struct sq930_i2cwrite_cmd {
	u8		sqi_byte0;
	u8		sqi_byte1;
};

struct sq930_resolution {
	unsigned int	sqr_width;
	unsigned int	sqr_height;
	unsigned int	sqr_sizeid;
	unsigned int	sqr_reqsize;
	unsigned int	sqr_maxpkt;
	unsigned int	sqr_minframeival;
	unsigned int	sqr_minppl;
};

struct sq930_fmt {
	struct usbcam_pix_fmt base;
	const struct sq930_resolution *sqf_res_tbl;
	u8 sqf_bpp;
};

struct sq930_sensor {
	const char	*sqs_name;
	unsigned int	sqs_cmos : 1;
	int		(*sqs_probe)(struct sq930_dev *);
	int		(*sqs_init)(struct sq930_dev *);
	int		(*sqs_cap_start)(struct sq930_dev *);
	void		(*sqs_cap_stop)(struct sq930_dev *);
};

struct sq930_pt_axis {
	int		sqa_min;
	int		sqa_max;
	int		sqa_abspos;
	int		sqa_reqpos;
	unsigned int	sqa_reqreset : 1,
			sqa_invert : 1;
	u8		sqa_devaxis;
};

struct sq930_model {
	const char			*sqm_name;
	const struct sq930_sensor	*sqm_sensor;
	unsigned int			sqm_wclm_f5f0 : 1;
	u8				sqm_is_slaveid;
	char				sqm_gpio_led;
	char				sqm_gpio_ispower;
	char				sqm_gpio_isreset;
	char				sqm_gpio_isresetbar;
	char				sqm_gpio_i2c_scl;
	char				sqm_gpio_i2c_sda;
	u32				sqm_gpio_extra;
	struct sq930_pt_axis		sqm_pt_axes[2];
};

/*
 * struct sq930_dev is the per-device private structure for this
 * minidriver.  We store all the details of the current state of the
 * camera in here, other than those explicitly shared with usbcam.
 */

struct sq930_dev {
	struct usbcam_dev *sqd_parent;
	int sqd_timeout;

	u8 sqd_dev_info[8];

	unsigned int sqd_info_cmos : 1,
		sqd_info_highspeed : 1;
	u16 sqd_info_chip_model;
	u8 sqd_info_fw_ver;

	u8 sqd_frame_meta[8];

	const struct sq930_model *sqd_model;
	const struct sq930_sensor *sqd_sensor;

	u32 sqd_gpio;
	u8 sqd_gpio_init;

	/* Pan/Tilt related members */
	u8 sqd_pt_speed;
	u8 sqd_pt_lastaxis;
	u8 sqd_pt_reqspeed;
	u8 sqd_pt_reqspeed_reset;
	u8 sqd_pt_actaxis;
	int sqd_pt_actdelta;
	unsigned int sqd_pt_enable : 1;
	struct sq930_pt_axis sqd_pt[2];
	struct usbcam_delayedwork sqd_pt_work;

	const struct sq930_fmt *sqd_fmts;
	int sqd_nfmts;

	const struct sq930_fmt *sqd_fmt;
	const struct sq930_resolution *sqd_res;

	int sqd_frame_meta_enable;	/* Currently always on */
	int sqd_frame_off;
	int sqd_useless_pkts;
	int sqd_seq_good_frames;
	int sqd_nab;
	int sqd_quant;
	struct usbcam_urbstream sqd_bulk;

	int sqd_powerline;
	int sqd_exposure;
	int sqd_aec;
	int sqd_ppl;
};

#define udp_sq930(UDP) ((struct sq930_dev *)((UDP)->ud_minidrv_data))

/* Raw controller commands (DMA = buf is kmalloc()'d) */
extern int sq930_control_dma(struct sq930_dev *sdp, int write,
			     u16 val, u16 idx, u8 *buf, int len);
extern int sq930_control(struct sq930_dev *sdp, int write, u16 val, u16 idx,
			 u8 *buf, int len);

/* Microcontroller local bus access */
extern int sq930_ucbus_read(struct sq930_dev *sdp, u16 addr,
			    u8 *data, int dlen);
extern int sq930_ucbus_write(struct sq930_dev *sdp,
			     const struct sq930_ucbuswrite_cmd *cmd, int ncmds,
			     int batchsize, int delay_ms);

/* I2C bus access */
extern int sq930_i2c_read(struct sq930_dev *sdp, u8 slaveid,
			  u8 offset, u8 *data, int dlen);
extern int sq930_i2c_write(struct sq930_dev *sdp, u8 slaveid,
			   const struct sq930_i2cwrite_cmd *cmds, int ncmds,
			   int batchsize, int delay_ms);
extern int sq930_i2c_poll_id(struct sq930_dev *sdp, u8 slaveid);
extern int sq930_i2c_reset(struct sq930_dev *sdp);

/* GPIO access */
extern int sq930_gpio_get(struct sq930_dev *sdp, u32 mask, u32 *result);
extern int sq930_gpio_set(struct sq930_dev *sdp, u32 val, u32 mask,
			  int delay_ms);

static inline int sq930_gpio_get_line(struct sq930_dev *sdp, int line) {
	u32 val;
	int res = sq930_gpio_get(sdp, (1 << line), &val);
	return res ? res : (val ? 1 : 0);
}
#define sq930_gpio_set_line(SDP, LINE, ON, DELAY) \
	sq930_gpio_set((SDP), (ON) ? (1 << (LINE)) : 0, (1 << (LINE)), (DELAY))

/* Misc. controller commands */
extern int sq930_cap_start_set(struct sq930_dev *sdp, int sizeid,
			       int framemeta, int qtable,
			       const u8 *cmdbytes, const u8 *jpegbytes);
extern int sq930_exposure_set(struct sq930_dev *sdp, u16 intstartclk,
			      u16 frameclks, u8 gain, u8 i2cslaveid,
			      const struct sq930_i2cwrite_cmd *i2c, int ni2c);

#endif  /* defined(__KERNEL__) */
