/*
 * Driver for SQ930B/C based webcams
 * Copyright (C) 2007 Sam Revitch <samr7@cs.washington.edu>
 *
 * This driver is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This driver is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this driver; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
/*
 * The Creative Labs Webcam Live! Motion has a Sharp LZ24BP CCD image
 * sensor.  The device can provide raw BA81 pixel data or JPEG data.
 * A high speed link is required to acquire raw data in larger sizes or
 * at higher frame rates.
 */

#define CONFIG_USB_USBCAM_DEBUG
#define USBCAM_DEBUG_DEFAULT 0xff7a

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/usb.h>
#include "sq930.h"


enum {
	SQ930_GPIO_DFL_I2C_SDA = 8,
	SQ930_GPIO_DFL_I2C_SCL = 9,
	SQ930_GPIO_DFL_PAN_SENSE = 10,
	SQ930_GPIO_DFL_TILT_SENSE = 24,
	SQ930_GPIO_DFL_LED = 28,
};



static int ledmode = 2;
module_param(ledmode, int, S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(ledmode, "LED mode (0=off,1=on,2=capture)");

static int ptreset = 1;
module_param(ptreset, int, S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(ledmode, "Execute Pan/Tilt Reset on Detection (0=no 1=yes)");


#define sq930_dev(PP) ((PP)->sqd_parent->ud_dev)


static const struct sq930_sensor *sq930_sensors[];


/*
 * Basic supported commands, sent in the low byte of the value parameter
 * of a device vendor control transfer.  Some (pan/tilt) specify an upper
 * byte for convenience and specificity.
 */

enum {
	SQ930_CTRL_UCBUS_IO		= 0x0001,
	SQ930_CTRL_I2C_IO		= 0x0002,
	SQ930_CTRL_GPIO			= 0x0005,
	SQ930_CTRL_CAP_START		= 0x0010,
	SQ930_CTRL_CAP_STOP		= 0x0011,

	SQ930_CTRL_PANTILT_SPEED	= 0x001b,
	SQ930_CTRL_PANTILT_START	= 0x011b,
	SQ930_CTRL_PANTILT_STOP		= 0x021b,
	SQ930_CTRL_PANTILT_STATE	= 0x031b,

	SQ930_CTRL_I2C_POLL_ID		= 0x001c,
	SQ930_CTRL_SET_EXPOSURE		= 0x001d,
	SQ930_CTRL_RESET		= 0x001e,
	SQ930_CTRL_GET_DEV_INFO		= 0x001f,

	SQ930_AXIS_PAN			= 0,
	SQ930_AXIS_TILT			= 1,
	SQ930_AXIS_NONE			= 2,

	//SQ930_RESETTING			= 0x866,
	SQ930_RESETTING			= 0x4f8,
};

int sq930_control_dma(struct sq930_dev *sdp, int write, u16 val, u16 idx,
		      u8 *buf, int len)
{
	int res;

	sq_dbg(sdp, SQ_CMD, "%s: %s %04x/%04x %d",
	       __FUNCTION__, write ? "write" : "read",
	       val, idx, len);

	res = usb_control_msg(sq930_dev(sdp),
			      write ? usb_sndctrlpipe(sq930_dev(sdp), 0)
			      : usb_rcvctrlpipe(sq930_dev(sdp), 0),
			      0x0c,
			      (write ? USB_DIR_OUT : USB_DIR_IN) |
			      USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			      val, idx, (char*) buf, len,
			      sdp->sqd_timeout);
	if (res < 0) {
		sq_dbg(sdp, SQ_CMD, "%s: %s %04x/%04x %d failed: %d",
		       __FUNCTION__, write ? "write" : "read",
		       val, idx, len, res);
		return res;
	}
	if (res != len) {
		sq_err(sdp, "%s: %04x/%04x: short %s (exp:%d got:%d)",
		       __FUNCTION__, val, idx, write ? "write" : "read",
		       len, res);
		return -EIO;
	}
	return 0;
}

int sq930_control(struct sq930_dev *sdp, int write, u16 val, u16 idx,
		  u8 *buf, int len)
{
	u8 *bbuf = NULL;
	int res;

	if (len) {
		bbuf = (u8 *) kmalloc(len, GFP_KERNEL);
		if (!bbuf)
			return -ENOMEM;
		if (write)
			memcpy(bbuf, buf, len);
	}
	res = sq930_control_dma(sdp, write, val, idx, bbuf, len);
	if (bbuf) {
		if (!write)
			memcpy(buf, bbuf, len);
		kfree(bbuf);
	}
	return res;
}


int sq930_ucbus_read(struct sq930_dev *sdp, u16 addr, u8 *data, int dlen)
{
	return sq930_control(sdp, 0, (addr & 0xff) << 8 | SQ930_CTRL_UCBUS_IO,
			     (addr & 0xff00) >> 8, data, dlen);
}

int sq930_ucbus_write(struct sq930_dev *sdp,
		      const struct sq930_ucbuswrite_cmd *cmd, int ncmds,
		      int batchsize, int delay_ms)
{
	u8 *buf = NULL;
	u16 val, idx;
	int len, i, res = 0;

	if (ncmds <= 0) {
		sq_err(sdp, "%s: %d cmds?", __FUNCTION__, ncmds);
		return -EINVAL;
	}

	if (!batchsize)
		batchsize = 11;

	while (ncmds) {
		len = ncmds;
		if (len > batchsize)
			len = batchsize;

		val = (cmd->bw_addr & 0xff) << 8 | SQ930_CTRL_UCBUS_IO;
		idx = (cmd->bw_addr & 0xff00) >> 8 | (cmd->bw_data << 8);
		cmd++;
		ncmds--;

		len = (len - 1) * 3;
		if (len) {
			if (!buf)
				buf = (u8 *) kmalloc(len, GFP_KERNEL);
			if (!buf)
				return -ENOMEM;
			for (i = 0; i < len; cmd++, ncmds--) {
				buf[i++] = cmd->bw_addr & 0xff;
				buf[i++] = (cmd->bw_addr & 0xff00) >> 8;
				buf[i++] = cmd->bw_data;
			}
		}

		res = sq930_control_dma(sdp, 1, val, idx, buf, len);
		if (res)
			break;
		//usbcam_hexdump(sdp->sqd_parent, buf, len);
		if (delay_ms)
			msleep(delay_ms);
	}

	if (buf)
		kfree(buf);
	return res;
}


int sq930_gpio_get(struct sq930_dev *sdp, u32 mask, u32 *result)
{
	int res, regno;
	u8 val;
	u32 vread = 0;

	for (regno = 0; mask != 0; regno++, mask >>= 8) {
		if (!(mask & 0xff))
			continue;

		/* Messing with these GPIO lines may hang the device */
		if ((regno == 0) || (regno == 2)) {
			sq_err(sdp, "You don't care about GPIO register %d",
			       regno);
			continue;
		}

		res = sq930_control(sdp, 0, (regno << 8) | SQ930_CTRL_GPIO, 0,
				    &val, 1);
		if (res)
			return res;

		vread |= ((~val & (mask & 0xff)) << (regno * 8));
	}

	*result = vread;
	return 0;
}

int sq930_gpio_set(struct sq930_dev *sdp, u32 val, u32 mask, int delay_ms)
{
	int res, regno;
	u8 vnew;
	u32 vexist;

	/* GPIO bits appear to be inverted */
	val = ~val;

	for (regno = 0, vexist = sdp->sqd_gpio;
	     mask != 0;
	     regno++, val >>= 8, mask >>= 8, vexist >>= 8) {
		if (!(mask & 0xff))
			continue;
		vnew = val & (mask & 0xff);
		vnew |= (vexist & 0xff) & ~(mask & 0xff);
		if ((vnew == (vexist & 0xff)) &&
		    (sdp->sqd_gpio_init & (1 << regno)))
			continue;

		/* Messing with these GPIO lines may hang the device */
		if ((regno == 0) || (regno == 2)) {
			sq_err(sdp, "You don't want to set GPIO register %d",
			       regno);
			continue;
		}

		res = sq930_control(sdp, 1, (regno << 8) | SQ930_CTRL_GPIO,
				    vnew << 8, NULL, 0);
		if (res)
			return res;

		sdp->sqd_gpio = ((sdp->sqd_gpio & ~(0xff << (regno * 8))) |
				 (vnew << (regno * 8)));
		sdp->sqd_gpio_init |= (1 << regno);
		msleep(delay_ms);
	}

	return 0;
}


/*
 * This command will issue a series of i2c read requests.
 * It appears to auto-increment the offset after receiving a single byte.
 */
int sq930_i2c_read(struct sq930_dev *sdp, u8 slaveid,
		   u8 offset, u8 *data, int dlen)
{
	return sq930_control(sdp, 0, (slaveid << 8) | SQ930_CTRL_I2C_IO,
			     offset, data, dlen);
}

int sq930_i2c_write(struct sq930_dev *sdp, u8 slaveid,
		    const struct sq930_i2cwrite_cmd *cmd, int ncmds,
		    int batchsize, int delay_ms)
{
	u8 *buf = NULL;
	u16 idx;
	int len, i, res = 0;

	if (ncmds <= 0) {
		sq_err(sdp, "%s: %d cmds?", __FUNCTION__, ncmds);
		return -EINVAL;
	}

	if (!batchsize)
		batchsize = 17;

	while (ncmds) {
		len = ncmds;
		if (len > batchsize)
			len = batchsize;

		idx = cmd->sqi_byte0 | (cmd->sqi_byte1 << 8);
		cmd++;
		ncmds--;

		len = (len - 1) * 2;
		if (len) {
			if (!buf)
				buf = (u8 *) kmalloc(len, GFP_KERNEL);
			if (!buf)
				return -ENOMEM;
			for (i = 0; i < len; cmd++, ncmds--) {
				buf[i++] = cmd->sqi_byte0;
				buf[i++] = cmd->sqi_byte1;
			}
		}

		res = sq930_control_dma(sdp, 1,
					(slaveid << 8) | SQ930_CTRL_I2C_IO,
					idx, buf, len);
		if (res)
			break;
		if (delay_ms)
			msleep(delay_ms);
	}

	if (buf)
		kfree(buf);
	return res;
}

int sq930_i2c_poll_id(struct sq930_dev *sdp, u8 slaveid)
{
	u8 stat = 0;
	int res;

	res = sq930_control(sdp, 0, (slaveid << 8) | SQ930_CTRL_I2C_POLL_ID,
			    0, &stat, sizeof(stat));
	if (res)
		return res;
	if ((stat != 0) && (stat != 1)) {
		sq_err(sdp, "Unexpected I2C poll result %d", stat);
		return 0;
	}
	return stat;
}

int sq930_i2c_reset(struct sq930_dev *sdp)
{
	u32 both_down;
	int scl, sda, res;

	scl = SQ930_GPIO_DFL_I2C_SCL;
	sda = SQ930_GPIO_DFL_I2C_SDA;
	if (sdp->sqd_model->sqm_gpio_i2c_scl)
		scl = sdp->sqd_model->sqm_gpio_i2c_scl;
	if (sdp->sqd_model->sqm_gpio_i2c_sda)
		sda = sdp->sqd_model->sqm_gpio_i2c_sda;

	both_down = (1 << scl) | (1 << sda);
	if ((res = sq930_gpio_set(sdp, both_down, both_down, 5)) ||
	    (res = sq930_gpio_set_line(sdp, scl, 0, 5)) ||
	    (res = sq930_gpio_set_line(sdp, sda, 0, 5)))
		return res;
	return 0;
}


int sq930_cap_start_set(struct sq930_dev *sdp, int sizeid, int framemeta,
			int qtable, const u8 *cmdbytes, const u8 *jpegbytes)
{
	u16 val, idx;
	int res;
	u8 *buf = NULL;

	buf = kmalloc(32, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	memcpy(buf, cmdbytes, 16);
	if (jpegbytes) {
		memcpy(buf + 16, jpegbytes, 16);
	} else {
		qtable = 0;
		memset(buf + 16, 0, 16);
	}

	val = SQ930_CTRL_CAP_START | (jpegbytes ? 0x200 : 0x100) |
		(framemeta ? 0x800 : 0) | (qtable << 12);
	idx = sizeid | (jpegbytes ? 0x500 : 0xa00);
	res = sq930_control_dma(sdp, 1, val, idx, buf, 32);
	if (!res) {
		sdp->sqd_quant = qtable;
		sdp->sqd_frame_meta_enable = framemeta ? 1 : 0;
		msleep(10);
	}
	kfree(buf);
	return res;
}


int sq930_exposure_set(struct sq930_dev *sdp, u16 intstartclk,
		       u16 frameclks, u8 gain, u8 i2cslaveid,
		       const struct sq930_i2cwrite_cmd *i2c, int ni2c)
{
	u8 *buf;
	int res, pos = 0;

	assert(intstartclk <= frameclks);

	buf = (u8 *) kmalloc(5 + (ni2c ? (2 * (ni2c + 1)) : 0),
			     GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	if (ni2c) {
		buf[0] = i2cslaveid;
		buf[1] = 2 * ni2c;
		pos = 2;
		while (ni2c--) {
			buf[pos++] = i2c->sqi_byte0;
			buf[pos++] = i2c->sqi_byte1;
		}
	}

	*(__be16*) &buf[pos] = cpu_to_be16(intstartclk);
	pos += 2;
	*(__be16*) &buf[pos] = cpu_to_be16(frameclks);
	pos += 2;
	buf[pos++] = gain;

	res = sq930_control_dma(sdp, 1, SQ930_CTRL_SET_EXPOSURE |
				(ni2c ? 0x100 : 0), 0,
				buf, pos);
	kfree(buf);
	return res;
}


enum {
	VSEN_RST_BEGIN = 1,
	VSEN_RST_END = 0,

	SQ930_CLKFREQ_60MHZ		= 0,
	SQ930_CLKFREQ_OTHER		= 0x80,
};

static int sq930_ucreg_clkfreq_set(struct sq930_dev *sdp, int val)
{
	struct sq930_ucbuswrite_cmd clkfreq_cmd = {
		.bw_addr = 0xf031,
		.bw_data = val
	};
	return sq930_ucbus_write(sdp, &clkfreq_cmd, 1, 0, 10);
}

static int sq930_led_set(struct sq930_dev *sdp, int led_on)
{
	u32 ledval;

	switch (ledmode) {
	case 0:
		led_on = 0;
		break;
	case 1:
		led_on = 1;
		break;
	}

	ledval = SQ930_GPIO_DFL_LED;
	if (sdp->sqd_model->sqm_gpio_led)
		ledval = sdp->sqd_model->sqm_gpio_led;
	return sq930_gpio_set_line(sdp, ledval, led_on, 0);
}

/*
 * Pan/tilt commands, for motorized SQ930B devices
 *
 * The sqd_pt array elements are indexed by logical axis.  The
 * sqa_devaxis member is used to map each to a device motor control.
 */

static void sq930_pantilt_stopped(struct sq930_dev *sdp,
				  struct sq930_pt_axis *sap, int rem)
{
	sq_dbg(sdp, SQ_CTRL, "ptstopped: active %d %d rem %d",
	       sdp->sqd_pt_actaxis, sdp->sqd_pt_actdelta, rem);

	if (sdp->sqd_pt_actdelta != SQ930_RESETTING) {
		if (sdp->sqd_pt_actdelta < 0)
			rem = -rem;
		sap->sqa_abspos += sdp->sqd_pt_actdelta - rem;
	} else if (!rem) {
		sap->sqa_reqreset = 0;
		sap->sqa_abspos = sap->sqa_min;
	}
	sdp->sqd_pt_lastaxis = sdp->sqd_pt_actaxis;
	sdp->sqd_pt_actaxis = SQ930_AXIS_NONE;
	sdp->sqd_pt_actdelta = 0;
}

static int sq930_pantilt_state(struct sq930_dev *sdp)
{
	u8 buf[1];
	int res;

	res = sq930_control(sdp, 0, SQ930_CTRL_PANTILT_STATE, 0, buf, 1);
	if (res)
		return res;

	if (sdp->sqd_pt_actaxis == SQ930_AXIS_NONE) {
		if (buf[0])
			sq_err(sdp, "actaxis=NONE but state=%02x", buf[0]);
		return buf[0] ? -EBUSY : 0;
	}

	sq_dbg(sdp, SQ_CTRL, "pantilt_state: %02x", buf[0]);

	switch (sdp->sqd_pt[sdp->sqd_pt_actaxis].sqa_devaxis) {
	case 1:
		if (buf[0] & ~0x10)
			sq_err(sdp, "devaxis=1 but state=%02x", buf[0]);
		if (buf[0] & 0x10)
			return -EBUSY;
		break;
	case 2:
		if (buf[0] & ~0x1)
			sq_err(sdp, "devaxis=2 but state=%02x", buf[0]);
		if (buf[0] & 0x1)
			return -EBUSY;
		break;
	default:
		assert(0);
	}

	sq930_pantilt_stopped(sdp, &sdp->sqd_pt[sdp->sqd_pt_actaxis], 0);
	return 0;
}

static int sq930_pantilt_stop(struct sq930_dev *sdp)
{
	u8 buf[4];
	int rem, res;

	if (sdp->sqd_pt_actdelta == SQ930_RESETTING) {
		res = sq930_pantilt_state(sdp);
		if (!res)
			return 0;
	}

	res = sq930_control(sdp, 0, SQ930_CTRL_PANTILT_STOP, 0,
			    buf, sizeof(buf));
	if (res)
		return res;

	if (sdp->sqd_pt_actaxis == SQ930_AXIS_NONE)
		return 0;

	switch (sdp->sqd_pt[sdp->sqd_pt_actaxis].sqa_devaxis) {
	case 1:
		rem = le16_to_cpup((__le16*) &buf[0]);
		break;
	case 2:
		rem = le16_to_cpup((__le16*) &buf[2]);
		break;
	default:
		assert(0);
		return -EINVAL;
	}

	sq930_pantilt_stopped(sdp, &sdp->sqd_pt[sdp->sqd_pt_actaxis], rem);
	return 0;
}

static int sq930_pantilt_speed(struct sq930_dev *sdp, int speed)
{
	u8 buf[11];
	int value, res;

	memset(buf, 0, sizeof(buf));

	if (speed <= 1) {
		value = 63035;
	} else if (speed >= 26) {
		value = 1000;
	} else
		value = 65535 - (2500 * speed);

	buf[0] = 0x10;
	*(__le16*) &buf[2] = cpu_to_le16(value);
	buf[3] = 1;
	buf[4] = 3;
	buf[5] = 2;

	res = sq930_control(sdp, 1, SQ930_CTRL_PANTILT_SPEED, 0,
			     buf, sizeof(buf));
	if (res)
		return res;

	sq_dbg(sdp, SQ_CTRL, "ptspeed: delay %d", value);
	sdp->sqd_pt_speed = speed;
	return 0;
}

static int sq930_pantilt_start(struct sq930_dev *sdp, int axis,
			       int reset, int delta)
{
	u8 buf[7];
	u8 devaxis;
	int dir, res;

	if (!delta && !reset)
		return 0;

	devaxis = sdp->sqd_pt[axis].sqa_devaxis;
	if (!devaxis)
		return -EINVAL;

	if (sdp->sqd_pt_actdelta) {
		res = sq930_pantilt_stop(sdp);
		if (res)
			return res;
	}

	if (reset) {
		if (sdp->sqd_pt_speed != sdp->sqd_pt_reqspeed_reset) {
			res = sq930_pantilt_speed(sdp,
						  sdp->sqd_pt_reqspeed_reset);
			if (res)
				return res;
		}

		dir = 1;
		delta = SQ930_RESETTING;
	}

	else {
		if (sdp->sqd_pt_speed != sdp->sqd_pt_reqspeed) {
			res = sq930_pantilt_speed(sdp, sdp->sqd_pt_reqspeed);
			if (res)
				return res;
		}
		if (delta < 0) {
			dir = 1;
			delta = -delta;
		} else {
			dir = 0;
		}
	}

	memset(buf, 0, sizeof(buf));
	buf[0] = reset ? (devaxis << 4) : devaxis;
	buf[1] = dir ? 1 : 0;
	*(__le16*) &buf[2] = cpu_to_le16(delta);

	res = sq930_control(sdp, 1, SQ930_CTRL_PANTILT_START, 0,
			    buf, sizeof(buf));

	if (res)
		return res;

	sdp->sqd_pt_actaxis = axis;
	sdp->sqd_pt_actdelta = (reset || !dir) ? delta : -delta;
	return 0;
}

static int sq930_pantilt_continue(struct sq930_dev *sdp)
{
	struct sq930_pt_axis *sap;
	int delta, res, i;

	if (sdp->sqd_pt_actdelta) {
		res = 0;
		goto done;
	}

	for (i = 0; i < ARRAY_SIZE(sdp->sqd_pt); i++) {
		sap = &sdp->sqd_pt[i];
		if (sap->sqa_reqreset) {
			res = sq930_pantilt_start(sdp, i, 1, 0);
			goto done;
		}
	}

	if (sdp->sqd_pt_lastaxis != SQ930_AXIS_NONE) {
		i = sdp->sqd_pt_lastaxis;
		sap = &sdp->sqd_pt[i];
		if (sap->sqa_abspos != sap->sqa_reqpos)
			goto do_reqpos;
	}

	for (i = 0; i < ARRAY_SIZE(sdp->sqd_pt); i++) {
		sap = &sdp->sqd_pt[i];
		if (sap->sqa_abspos != sap->sqa_reqpos) {
		do_reqpos:
			delta = sap->sqa_reqpos - sap->sqa_abspos;
			res = sq930_pantilt_start(sdp, i, 0, delta);
			goto done;
		}
	}

	/* No pan/tilt actions left to do, don't reschedule */
	return 0;

done:
	if (!res) {
		usbcam_delayedwork_cancel(&sdp->sqd_pt_work);
		usbcam_delayedwork_queue(&sdp->sqd_pt_work, 100);
	}
	return res;
}

static void sq930_pantilt_poll(struct usbcam_workitem *wip)
{
	struct sq930_dev *sdp = container_of(wip, struct sq930_dev,
					     sqd_pt_work.dw_work);
	int res;

	res = sq930_pantilt_state(sdp);
	if (res == -EBUSY) {
		usbcam_delayedwork_queue(&sdp->sqd_pt_work, 100);
		return;
	}
	if (res) {
		sq_err(sdp, "%s: pantilt state query failed: %d",
		       __FUNCTION__, res);
		res = 0;
	}

	sq930_pantilt_continue(sdp);
}

static int sq930_pantilt_ctrl_get(struct usbcam_dev *udp,
				  const struct usbcam_ctrl *ctrlp,
				  struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	struct sq930_pt_axis *sap;
	int axis, res;

	switch (ctrlp->uc_v4l.id) {
	case V4L2_CID_SQ930_PAN_ABSOLUTE:
		axis = SQ930_AXIS_PAN;
		break;
	case V4L2_CID_SQ930_TILT_ABSOLUTE:
		axis = SQ930_AXIS_TILT;
		break;
	default:
		return -EINVAL;
	}

	sap = &sdp->sqd_pt[axis];

	if (!sap->sqa_devaxis)
		return -EINVAL;

	if (sdp->sqd_pt_actaxis == axis) {
		res = sq930_pantilt_stop(sdp);
		if (res)
			return res;
	}

	cp->value = sap->sqa_invert ? -sap->sqa_abspos : sap->sqa_abspos;
	return sq930_pantilt_continue(sdp);
}

static int sq930_pantilt_ctrl_set(struct usbcam_dev *udp,
				  const struct usbcam_ctrl *ctrlp,
				  const struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	struct sq930_pt_axis *sap;
	int axis, relative, res, delta, value;

	if (sdp->sqd_pt_actaxis) {
		res = sq930_pantilt_stop(sdp);
		if (res)
			return res;
	}

	relative = 0;
	switch (ctrlp->uc_v4l.id) {
	case V4L2_CID_SQ930_PAN_RELATIVE:
		relative = 1;
	case V4L2_CID_SQ930_PAN_ABSOLUTE:
		axis = SQ930_AXIS_PAN;
		break;
	case V4L2_CID_SQ930_TILT_RELATIVE:
		relative = 1;
	case V4L2_CID_SQ930_TILT_ABSOLUTE:
		axis = SQ930_AXIS_TILT;
		break;
	default:
		return -EINVAL;
	}

	sap = &sdp->sqd_pt[axis];
	if (!sap->sqa_devaxis)
		return -EINVAL;

	value = cp->value;
	if (sap->sqa_invert)
		value = -value;
	if (!relative)
		delta = value - sap->sqa_abspos;
	else {
		delta = value;
		if ((sap->sqa_abspos + delta) < sap->sqa_min)
			delta = sap->sqa_min - sap->sqa_abspos;
		else if ((sap->sqa_abspos + delta) > sap->sqa_max)
			delta = sap->sqa_max - sap->sqa_abspos;
	}

	sq_dbg(sdp, SQ_CTRL, "set %s: %d old %d delta %d",
	       (axis == SQ930_AXIS_PAN) ? "pan" : "tilt",
	       value, sap->sqa_abspos, delta);
	sap->sqa_reqpos = sap->sqa_abspos + delta;
	return sq930_pantilt_continue(sdp);
}

static int sq930_ptspeed_ctrl_get(struct usbcam_dev *udp,
				  const struct usbcam_ctrl *ctrlp,
				  struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	if (!sdp->sqd_pt_enable)
		return -EINVAL;
	cp->value = sdp->sqd_pt_reqspeed;
	return 0;
}

static int sq930_ptspeed_ctrl_set(struct usbcam_dev *udp,
				  const struct usbcam_ctrl *ctrlp,
				  const struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	int res;

	if (!sdp->sqd_pt_enable)
		return -EINVAL;

	if (!sdp->sqd_pt_actdelta) {
		res = sq930_pantilt_stop(sdp);
		if (res)
			return res;
	}
	sdp->sqd_pt_reqspeed = cp->value;
		return 0;
	return sq930_pantilt_speed(sdp, cp->value);
}

static int sq930_ptreset_ctrl_set(struct usbcam_dev *udp,
				  const struct usbcam_ctrl *ctrlp,
				  const struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	int res;

	if (!sdp->sqd_pt_enable)
		return -EINVAL;

	if (sdp->sqd_pt_actdelta) {
		res = sq930_pantilt_stop(sdp);
		if (res)
			return res;
	}
	sdp->sqd_pt[SQ930_AXIS_PAN].sqa_reqreset = 1;
	sdp->sqd_pt[SQ930_AXIS_TILT].sqa_reqreset = 1;
	return sq930_pantilt_continue(sdp);
}

static const struct usbcam_ctrl sq930_pt_controls[] = {
	{
		.uc_v4l = { .id = V4L2_CID_SQ930_PAN_ABSOLUTE,
			    .type = V4L2_CTRL_TYPE_INTEGER,
			    .name = "Pan (Absolute)",
			    .step = 1,
			    .flags = 0 },
		.get_fn = sq930_pantilt_ctrl_get,
		.set_fn = sq930_pantilt_ctrl_set,
	},
	{
		.uc_v4l = { .id = V4L2_CID_SQ930_TILT_ABSOLUTE,
			    .type = V4L2_CTRL_TYPE_INTEGER,
			    .name = "Tilt (Absolute)",
			    .step = 1,
			    .flags = 0 },
		.get_fn = sq930_pantilt_ctrl_get,
		.set_fn = sq930_pantilt_ctrl_set,
	},
	{
		.uc_v4l = { .id = V4L2_CID_SQ930_PAN_RELATIVE,
			    .type = V4L2_CTRL_TYPE_INTEGER,
			    .name = "Pan (Relative)",
			    .step = 1,
			    .flags = 0 },
		.set_fn = sq930_pantilt_ctrl_set,
	},
	{
		.uc_v4l = { .id = V4L2_CID_SQ930_TILT_RELATIVE,
			    .type = V4L2_CTRL_TYPE_INTEGER,
			    .name = "Tilt (Relative)",
			    .step = 1,
			    .flags = 0 },
		.set_fn = sq930_pantilt_ctrl_set,
	},
	{
		.uc_v4l = { .id = V4L2_CID_SQ930_PANTILT_SPEED,
			    .type = V4L2_CTRL_TYPE_INTEGER,
			    .name = "Pan/Tilt Speed",
			    .step = 1,
			    .minimum = 1,
			    .maximum = 25,
			    .default_value = 10,
			    .flags = 0 },
		.get_fn = sq930_ptspeed_ctrl_get,
		.set_fn = sq930_ptspeed_ctrl_set,
	},
	{
		.uc_v4l = { .id = V4L2_CID_SQ930_PANTILT_RESET,
			    .type = V4L2_CTRL_TYPE_BUTTON,
			    .name = "Pan/Tilt Reset",
			    .flags = 0 },
		.set_fn = sq930_ptreset_ctrl_set,
	},
};

static int sq930_pt_init(struct sq930_dev *sdp)
{
	int i;
	struct usbcam_ctrl *ctrlp;
	struct sq930_pt_axis *sap;
	int axis, relative;
	int naxes = 0;

	/* Copy the details out of the model structure */
	for (i = 0; i < ARRAY_SIZE(sdp->sqd_pt); i++) {
		if (!sdp->sqd_model->sqm_pt_axes[i].sqa_devaxis)
			continue;
		sdp->sqd_pt[i] = sdp->sqd_model->sqm_pt_axes[i];
		naxes++;
	}

	if (!naxes)
		return 0;

	sdp->sqd_pt_enable = 1;

	/* Add the controls */
	for (i = 0; i < ARRAY_SIZE(sq930_pt_controls); i++) {
		ctrlp = usbcam_ctrl_add_tmpl(sdp->sqd_parent,
					     &sq930_pt_controls[i],
					     sizeof(sq930_pt_controls[0]));
		if (!ctrlp)
			return -ENOMEM;

		relative = 0;
		switch (ctrlp->uc_v4l.id) {
		case V4L2_CID_SQ930_PAN_RELATIVE:
			relative = 1;
		case V4L2_CID_SQ930_PAN_ABSOLUTE:
			axis = SQ930_AXIS_PAN;
			break;
		case V4L2_CID_SQ930_TILT_RELATIVE:
			relative = 1;
		case V4L2_CID_SQ930_TILT_ABSOLUTE:
			axis = SQ930_AXIS_TILT;
			break;
		default:
			continue;
		}

		sap = &sdp->sqd_pt[axis];
		if (!sap->sqa_devaxis)
			continue;
		if (relative) {
			ctrlp->uc_v4l.maximum = sap->sqa_max - sap->sqa_min;
			ctrlp->uc_v4l.minimum =  -ctrlp->uc_v4l.maximum;
		} else {
			ctrlp->uc_v4l.maximum = sap->sqa_max;
			ctrlp->uc_v4l.minimum = sap->sqa_min;
		}
	}

	return 0;
}


static const struct sq930_resolution sq930_ba81_res[] = {
	{ 160, 120, 0,  19208,  19200, 33333, 0x310 },
	{ 176, 144, 1,  25352,      0, 33333, 0x310 },
	{ 320, 240, 2,  76808,      0, 33333, 0x310 },
	{ 352, 288, 3, 101384,      0, 33333, 0x310 },
	{ 640, 480, 4, 307208,      0, 33333, 0x310 },
	{ }
};

static const struct sq930_resolution sq930_jpeg_res[] = {
	{ 160, 120, 0,   5128,      0, 33333, 0x310 },
	{ 320, 240, 2,   5128,      0, 33333, 0x310 },
	{ 640, 480, 4,   5128,      0, 50000, 0x564 },
	{ }
};

static const struct sq930_resolution sq930_decimated_res[] = {
	{ 160, 120, 2,  76808,      0, 33333, 0x310 },
	{ 320, 240, 4, 307208,      0, 33333, 0x310 },
	{ }
};

static const struct sq930_fmt sq930_fmts[] = {
	{
		.base = { .description = "BA81 Raw",
			  .pixelformat = V4L2_PIX_FMT_SBGGR8,
			  .flags = 0 },
		.sqf_bpp = 1,
		.sqf_res_tbl = sq930_ba81_res,
	},
	{
		.base = { .description = "JPEG",
			  .pixelformat = V4L2_PIX_FMT_JPEG,
			  .flags = V4L2_FMT_FLAG_COMPRESSED },
		.sqf_bpp = 0,
		.sqf_res_tbl = sq930_jpeg_res,
	},
	/* Technically this is incorrect, this driver produces standard JPEG */
	{
		.base = { .description = "JPEG",
			  .pixelformat = V4L2_PIX_FMT_MJPEG,
			  .flags = V4L2_FMT_FLAG_COMPRESSED },
		.sqf_bpp = 0,
		.sqf_res_tbl = sq930_jpeg_res,
	},
	{
		.base = { .description = "RGB24 Decimated",
			  .pixelformat = V4L2_PIX_FMT_RGB24,
			  .flags = 0 },
		.sqf_bpp = 3,
		.sqf_res_tbl = sq930_decimated_res,
	},
	{
		.base = { .description = "BGR24 Decimated",
			  .pixelformat = V4L2_PIX_FMT_BGR24,
			  .flags = 0 },
		.sqf_bpp = 3,
		.sqf_res_tbl = sq930_decimated_res,
	},
};

static void sq930_do_stop(struct sq930_dev *sdp)
{
	if (sdp->sqd_parent->ud_capturing) {
		sdp->sqd_parent->ud_capturing = 0;
		usbcam_curframe_abortall(sdp->sqd_parent);
		usbcam_urbstream_stop(&sdp->sqd_bulk,
				      !sdp->sqd_parent->ud_disconnected);
		usbcam_urbstream_cleanup(&sdp->sqd_bulk);

		if (!sdp->sqd_parent->ud_disconnected) {
			(void) sq930_control(sdp, 1, SQ930_CTRL_CAP_STOP, 0,
					     NULL, 0);
			(void) sq930_led_set(sdp, 0);
		}
	}
}

static int sq930_check_frame_meta(struct sq930_dev *sdp,
				  const u8 *framemeta)
{
	assert(sdp->sqd_frame_meta_enable);
	//usbcam_hexdump(sdp->sqd_parent, framemeta, 8);

	if (framemeta[1]) {
		/* Frame is bad, try increasing PPL */

	}

	return 0;
}

/*
 * Transform the BA81 data to an RGB format, in an output image of
 * half horizontal and vertical resolution.  Discard one of the green
 * components.
 *
 * This is a geometry transformation, not a color space conversion or
 * a true demosaic operation.
 */
void sq930_rgb_xform(struct sq930_dev *sdp,
		     struct usbcam_curframe *cfp,
		     const u8 *pktdata, int pktlen)
{
	int ba81_rowlen;
	int ba81_rowpos;
	int ba81_rowhalf;
	int output_row;
	u8 red_off, green_off, blue_off, bpp;
	u8 *rowstart, *pixel;

	switch (sdp->sqd_parent->ud_format.pixelformat) {
	case V4L2_PIX_FMT_RGB24:
		bpp = 3;
		red_off = 0;
		green_off = 1;
		blue_off = 2;
		break;
	case V4L2_PIX_FMT_BGR24:
		bpp = 3;
		red_off = 2;
		green_off = 1;
		blue_off = 0;
		break;
	default:
		assert(0);
		return;
	}

	ba81_rowlen = (sdp->sqd_res->sqr_width * 2);
	output_row = sdp->sqd_frame_off / (ba81_rowlen * 2);
	ba81_rowpos = sdp->sqd_frame_off % (ba81_rowlen * 2);
	ba81_rowhalf = 0;
	if (ba81_rowpos > ba81_rowlen) {
		ba81_rowpos -= ba81_rowlen;
		ba81_rowhalf = 1;
	}

	rowstart = cfp->uf_base + (output_row *
				   sdp->sqd_parent->ud_format.bytesperline);
	pixel = rowstart + ((ba81_rowpos / 2) * bpp);

	if (ba81_rowhalf)
		goto lower_half;

upper_half:
	while (pktlen--) {
		if (ba81_rowpos & 1) {
			pixel[green_off] = *pktdata++;
			pixel += bpp;
			if (unlikely(++ba81_rowpos == ba81_rowlen)) {
				ba81_rowpos = 0;
				pixel = rowstart;
				goto lower_half;
			}
		} else {
			pixel[blue_off] = *pktdata++;
			ba81_rowpos++;
		}
	}
	return;

lower_half:
	while (pktlen--) {
		if (ba81_rowpos & 1) {
			pixel[red_off] = *pktdata++;
			pixel += bpp;
			if (unlikely(++ba81_rowpos == ba81_rowlen)) {
				ba81_rowpos = 0;
				rowstart += sdp->sqd_parent->
					ud_format.bytesperline;
				pixel = rowstart;
				if (++output_row == sdp->sqd_parent->
				    ud_format.height)
					break;
				goto upper_half;
			}
		} else {
			pktdata++;
			ba81_rowpos++;
		}
	}
	return;
}

static void sq930_bulk_raw_done(struct usbcam_dev *udp,
				struct usbcam_urbstream *usp,
				const u8 *pktdata,
				int pktlen, int pktstatus)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	struct usbcam_curframe cf;
	int mdlen = sdp->sqd_frame_meta_enable
		? sizeof(sdp->sqd_frame_meta) : 0;
	int sstart, dstart, copylen;

	if (pktstatus) {
		sq930_do_stop(sdp);
		return;
	}

	/*
	 * The vmap'd frame buffer is unfortunately not a valid DMA
	 * target, and it's sad that we have to memcpy it here and
	 * can't just capture into it directly.
	 */
	if (usbcam_curframe_get(sdp->sqd_parent, &cf)) {
		sq_dbg(sdp, SQ_MDINTF, "No frame ready, stopping");
		sq930_do_stop(sdp);
		return;
	}

	if ((sdp->sqd_frame_off + pktlen) > sdp->sqd_res->sqr_reqsize) {
		sq_err(sdp, "too much data, frame_off=%d pktlen=%d reqsize=%d",
		       sdp->sqd_frame_off, pktlen, sdp->sqd_res->sqr_reqsize);
		pktlen = sdp->sqd_res->sqr_reqsize - sdp->sqd_frame_off;
	}

	copylen = pktlen;
	if ((copylen + sdp->sqd_frame_off) >
	    (sdp->sqd_res->sqr_reqsize - mdlen))
		copylen = (sdp->sqd_res->sqr_reqsize - mdlen) -
			sdp->sqd_frame_off;

	if (copylen > 0) {
		if (sdp->sqd_parent->ud_format.pixelformat ==
		    V4L2_PIX_FMT_SBGGR8)
			memcpy(cf.uf_base + sdp->sqd_frame_off,
			       pktdata,
			       copylen);
		else
			sq930_rgb_xform(sdp, &cf, pktdata, copylen);
	}

	sdp->sqd_frame_off += pktlen;
	if (mdlen &&
	    (sdp->sqd_frame_off > (sdp->sqd_res->sqr_reqsize - mdlen))) {
		/* Handle the frame metadata */
		dstart = 0;
		sstart = ((sdp->sqd_res->sqr_reqsize - mdlen) -
			  (sdp->sqd_frame_off - pktlen));
		copylen = (sdp->sqd_frame_off -
			   (sdp->sqd_res->sqr_reqsize - mdlen));

		if (sstart < 0) {
			dstart = -sstart;
			copylen -= dstart;
			sstart = 0;
		}

		if (copylen > 0)
			memcpy(&sdp->sqd_frame_meta[dstart],
			       &pktdata[sstart],
			       copylen);
	}

	if (sdp->sqd_frame_off >= sdp->sqd_res->sqr_reqsize) {
		if (sdp->sqd_frame_meta_enable) {
			switch (sq930_check_frame_meta(sdp,
						       sdp->sqd_frame_meta)) {
			case -EIO:
				sq930_do_stop(sdp);
				return;
			case -EPIPE:
				sdp->sqd_seq_good_frames = 0;
				sdp->sqd_useless_pkts++;
				usbcam_curframe_testpattern(udp);
				break;
			}
		}

		sdp->sqd_frame_off = 0;
		sdp->sqd_seq_good_frames++;
		usbcam_curframe_complete(udp);
	}
}


/*
 * JPEG format helpers
 *
 * The camera controller returns raw entropy-coded JPEG MCUs.  Apps want
 * data that is more cooked, including markers, tables, and other info.
 * We define a set of routines to append markers to the frame buffer:
 *
 * sq930_jpeg_soi
 * sq930_jpeg_dqt
 * sq930_jpeg_dht
 * sq930_jpeg_sof
 * sq930_jpeg_sos
 * sq930_jpeg_eoi
 * sq930_jpeg_ecdcopy
 */

static int sq930_jpeg_soi(struct sq930_dev *sdp, struct usbcam_curframe *cfp)
{
	u8 *buf = cfp->uf_base + sdp->sqd_frame_off;
	if ((sdp->sqd_frame_off + 2) > cfp->uf_size)
		return -ENOSPC;
	buf[0] = 0xff;
	buf[1] = 0xd8;
	sdp->sqd_frame_off += 2;
	return 0;
}

static int sq930_jpeg_eoi(struct sq930_dev *sdp, struct usbcam_curframe *cfp)
{
	u8 *buf = cfp->uf_base + sdp->sqd_frame_off;
	if ((sdp->sqd_frame_off + 2) > cfp->uf_size)
		return -ENOSPC;
	buf[0] = 0xff;
	buf[1] = 0xd9;
	sdp->sqd_frame_off += 2;
	return 0;
}

/*
 * Device-specific quantization tables for luminance and chromiance planes
 * The set of tables used by the device are selected by the set resolution
 * command.
 */
static const u8 sq930_quant_lum[4][64] = {
	{ 2, 1, 1, 1, 1, 1, 2, 1, 1, 1, 2, 2, 2, 2, 3, 5, 3, 3,
	  2, 2, 3, 6, 4, 4, 3, 5, 7, 6, 7, 7, 7, 6, 7, 6, 8, 9,
	  11, 9, 8, 8, 10, 8, 6, 7, 10, 13, 10, 10, 11, 12, 12,
	  13, 12, 7, 9, 14, 15, 14, 12, 15, 11, 12, 12, 12 },
	{ 3, 2, 2, 2, 2, 1, 3, 2, 2, 2, 3, 3, 3, 3, 4, 7, 4, 4, 4,
	  4, 4, 9, 6, 6, 5, 7, 10, 9, 11, 11, 10, 9, 10, 10, 12,
	  13, 17, 14, 12, 12, 16, 12, 10, 10, 15, 20, 15, 16, 17,
	  18, 19, 19, 19, 11, 14, 21, 22, 21, 18, 22, 17, 18, 19, 18 },
	{ 6, 4, 4, 5, 4, 3, 6, 5, 4, 5, 6, 6, 6, 7, 9, 15, 9, 9, 8,
	  8, 9, 18, 13, 13, 10, 15, 21, 19, 22, 22, 21, 19, 21, 20, 24,
	  27, 34, 29, 24, 25, 32, 25, 20, 21, 30, 40, 30, 32, 35, 36,
	  38, 39, 38, 23, 28, 42, 45, 42, 37, 45, 34, 37, 38, 37 },
	{ 10, 6, 7, 8, 7, 6, 10, 8, 8, 8, 11, 10, 10, 11, 15, 25, 16, 15,
	  13, 13, 15, 30, 21, 23, 18, 25, 36, 31, 38, 37, 35, 31, 35, 34,
	  40, 45, 57, 48, 40, 42, 54, 43, 34, 35, 50, 68, 50, 54, 59, 61,
	  64, 65, 64, 38, 48, 70, 75, 70, 62, 75, 57, 63, 64, 61 }
};

static const u8 sq930_quant_chr[4][64] = {
	{ 3, 3, 3, 4, 3, 4, 8, 4, 4, 8, 18, 12, 10, 12, 18, 18, 18, 18,
	  18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18,
	  18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18,
	  18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18, 18 },
	{ 4, 5, 5, 6, 5, 6, 13, 7, 7, 13, 27, 18, 15, 18, 27, 27, 27, 27,
	  27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27,
	  27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27,
	  27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27 },
	{ 9, 10, 10, 13, 11, 13, 26, 14, 14, 26, 55, 37, 31, 37, 55, 55,
	  55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55,
	  55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55,
	  55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55, 55 },
	{ 15, 16, 16, 22, 19, 22, 44, 24, 24, 44, 92, 61, 52, 61, 92, 92,
	  92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	  92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92,
	  92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92 }
};

static int sq930_jpeg_dqt(struct sq930_dev *sdp, struct usbcam_curframe *cfp)
{
	u8 *buf = cfp->uf_base + sdp->sqd_frame_off;
	const int qtlen = 1 + 64 + 1 + 64;

	if ((sdp->sqd_frame_off + 2 + 2 + qtlen) > cfp->uf_size)
		return -ENOSPC;

	buf[0] = 0xff;
	buf[1] = 0xdb;
	*(__be16 *) &buf[2] = cpu_to_be16(2 + qtlen);
	buf += 4;

	buf[0] = 0;
	memcpy(&buf[1], sq930_quant_lum[sdp->sqd_quant], 64);
	buf += 65;

	buf[0] = 1;
	memcpy(&buf[1], sq930_quant_chr[sdp->sqd_quant], 64);
	buf += 65;

	sdp->sqd_frame_off += 2 + 2 + qtlen;
	return 0;
}

/* Standard JPEG Huffman tables recommended by ITU T.81 */
static const u8 sq930_huff_tbl[] = {
	0x00,
	0x00, 0x01, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09,
	0x0a, 0x0b,
	0x01,
	0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09,
	0x0a, 0x0b,
	0x10,
	0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03, 0x05, 0x05,
	0x04, 0x04, 0x00, 0x00, 0x01, 0x7d,
	0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12, 0x21, 0x31,
	0x41, 0x06, 0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32,
	0x81, 0x91, 0xa1, 0x08, 0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52,
	0xd1, 0xf0, 0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
	0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a,
	0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45,
	0x46, 0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
	0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x83,
	0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93, 0x94,
	0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
	0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
	0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
	0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8,
	0xd9, 0xda, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8,
	0xe9, 0xea, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
	0xf9, 0xfa,
	0x11,
	0x00, 0x02, 0x01, 0x02, 0x04, 0x04, 0x03, 0x04, 0x07, 0x05,
	0x04, 0x04, 0x00, 0x01, 0x02, 0x77,
	0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21, 0x31, 0x06,
	0x12, 0x41, 0x51, 0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81,
	0x08, 0x14, 0x42, 0x91, 0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33,
	0x52, 0xf0, 0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
	0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26, 0x27, 0x28,
	0x29, 0x2a, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44,
	0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55, 0x56,
	0x57, 0x58, 0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
	0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a,
	0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92,
	0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3,
	0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
	0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
	0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6,
	0xd7, 0xd8, 0xd9, 0xda, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
	0xe8, 0xe9, 0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
	0xf9, 0xfa
};

static int sq930_jpeg_dht(struct sq930_dev *sdp, struct usbcam_curframe *cfp)
{
	u8 *buf = cfp->uf_base + sdp->sqd_frame_off;
	const int htlen = 2 + sizeof(sq930_huff_tbl);

	if ((sdp->sqd_frame_off + 2 + htlen) > cfp->uf_size)
		return -ENOSPC;

	buf[0] = 0xff;
	buf[1] = 0xc4;
	*(__be16 *) &buf[2] = cpu_to_be16(htlen);
	memcpy(&buf[4], sq930_huff_tbl, sizeof(sq930_huff_tbl));

	sdp->sqd_frame_off += 2 + htlen;
	return 0;
}

static int sq930_jpeg_sof(struct sq930_dev *sdp, struct usbcam_curframe *cfp)
{
	u8 *buf = cfp->uf_base + sdp->sqd_frame_off;
	const u8 stdsof[] = {
		0x03, 0x01, 0x21, 0x00, 0x02, 0x11, 0x01, 0x03, 0x11, 0x01,
	};
	const int soflen = 2 + 1 + 2 + 2 + sizeof(stdsof);

	if ((sdp->sqd_frame_off + 2 + soflen) > cfp->uf_size)
		return -ENOSPC;

	buf[0] = 0xff;
	buf[1] = 0xc0;
	*(__be16 *) &buf[2] = cpu_to_be16(soflen);
	buf[4] = 8;
	*(__be16 *) &buf[5] = cpu_to_be16(sdp->sqd_res->sqr_height);
	*(__be16 *) &buf[7] = cpu_to_be16(sdp->sqd_res->sqr_width);
	memcpy(&buf[9], stdsof, sizeof(stdsof));

	sdp->sqd_frame_off += 2 + soflen;
	return 0;

}

static int sq930_jpeg_sos(struct sq930_dev *sdp, struct usbcam_curframe *cfp)
{
	u8 *buf = cfp->uf_base + sdp->sqd_frame_off;
	const u8 stdsos[] = {
		0x03, 0x01, 0x00, 0x02, 0x11, 0x03, 0x11, 0x00, 0x3f, 0x00
	};
	const int soslen = 2 + sizeof(stdsos);

	if ((sdp->sqd_frame_off + 2 + soslen) > cfp->uf_size)
		return -ENOSPC;

	buf[0] = 0xff;
	buf[1] = 0xda;
	*(__be16 *) &buf[2] = cpu_to_be16(soslen);
	memcpy(&buf[4], stdsos, sizeof(stdsos));

	sdp->sqd_frame_off += 2 + soslen;
	return 0;
}

static int sq930_jpeg_ecdcopy(struct sq930_dev *sdp,
			      struct usbcam_curframe *cfp,
			      const u8 *pktdata, int pktlen)
{
	u8 *buf = cfp->uf_base + sdp->sqd_frame_off;
	int rembytes = cfp->uf_size - (sdp->sqd_frame_off + pktlen);

	/*
	 * Sequences of 0xff 0xXX bytes in entropy-coded data run the
	 * risk of being mistaken for markers.  We stuff each 0xff
	 * with a trailing zero byte as per ITU T.81 F.1.2.3.
	 */
	if (rembytes < 0)
		return -ENOSPC;
	while (pktlen-- > 0) {
		if ((*buf++ = *pktdata++) == 0xff) {
			if (--rembytes < 0)
				return -ENOSPC;
			*buf++ = 0;
		}
	}
	sdp->sqd_frame_off = buf - cfp->uf_base;
	return 0;
}

static void sq930_bulk_jpeg_done(struct usbcam_dev *udp,
				 struct usbcam_urbstream *usp,
				 const u8 *pktdata,
				 int pktlen, int pktstatus)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	struct usbcam_curframe cf;
	const int mdlen = sizeof(sdp->sqd_frame_meta);
	int pos, off, nab, datalen, copylen, framecomplete;

	if (pktstatus || (pktlen != sdp->sqd_res->sqr_reqsize)) {
		sq930_do_stop(sdp);
		return;
	}

	datalen = pktlen;
	if (sdp->sqd_frame_meta_enable) {
		datalen -= mdlen;

		/* Check the frame metadata */
		switch (sq930_check_frame_meta(sdp, &pktdata[datalen])) {
		case -EIO:
			sq930_do_stop(sdp);
			return;
		case -EPIPE:
			if (sdp->sqd_frame_off >= 0) {
				usbcam_curframe_testpattern(udp);
				sdp->sqd_frame_off = -1;
				sdp->sqd_nab = 0;
			}
			return;
		}
	}

	/*
	 * Scan the compressed data buffer for the marker sequence and
	 * copy sections into the frame buffer
	 */
	nab = sdp->sqd_nab;
	off = 0;
	pos = 0;
	while (1) {
		framecomplete = 0;
		for (; pos < datalen; pos++) {
			if (pktdata[pos] != 0xab) {
				nab = 0;
				continue;
			}

			if (++nab == 8) {
				nab = 0;
				framecomplete = 1;
				pos++;
				sq_dbg(sdp, SQ_MDINTF,
				       "Found frame marker @ %d", pos);
				break;
			}
		}

		if (sdp->sqd_frame_off < 0) {
			off = pos;
			if (framecomplete) {
				sdp->sqd_useless_pkts = 0;
				goto frame_abort_done;
			}

			sdp->sqd_useless_pkts++;
			if (sdp->sqd_useless_pkts > 500) {
				sq_err(sdp, "Too many bad packets, stopping");
				sq930_do_stop(sdp);
			}
			break;
		}

		copylen = pos - off - (framecomplete ? 8 : 0);
		if (copylen < 0)
			copylen = 0;

		if (usbcam_curframe_get(sdp->sqd_parent, &cf)) {
			sq_dbg(sdp, SQ_MDINTF, "No frame ready, stopping");
			sq930_do_stop(sdp);
			return;
		}

		if (!sdp->sqd_frame_off) {
			/* Insert a JPEG header */
			if (sq930_jpeg_soi(sdp, &cf) ||
			    sq930_jpeg_dqt(sdp, &cf) ||
			    sq930_jpeg_dht(sdp, &cf) ||
			    sq930_jpeg_sof(sdp, &cf) ||
			    sq930_jpeg_sos(sdp, &cf)) {
				sq_err(sdp, "No room for JPEG header?!");
				sq930_do_stop(sdp);
				break;
			}
		}

		/* Copy the entropy coded data into the output buffer */
		if (sq930_jpeg_ecdcopy(sdp, &cf, &pktdata[off], copylen)) {
			sq_err(sdp, "JPEG ECD overflow");
			sq930_do_stop(sdp);
			break;
		}

		if (!framecomplete)
			break;

		/* Append an end-of-image marker */
		if (sq930_jpeg_eoi(sdp, &cf)) {
			sq_err(sdp, "No room for JPEG EOI");
			sq930_do_stop(sdp);
			break;
		}

		off = pos;
		cf.uf_size = sdp->sqd_frame_off;
		sdp->sqd_seq_good_frames++;
		usbcam_curframe_complete_detail(udp, &cf);

	frame_abort_done:
		sdp->sqd_frame_off = 0;

		if (pos == datalen)
			break;
	}

	sdp->sqd_nab = nab;
	return;
}

static void sq930_bulk_submit_error(struct usbcam_dev *udp,
				    struct usbcam_urbstream *usp, int status)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	sq930_do_stop(sdp);
}

static struct usbcam_urbstream_ops sq930_bulk_raw_ops = {
	.packet_done = sq930_bulk_raw_done,
	.submit_error = sq930_bulk_submit_error,
};

static struct usbcam_urbstream_ops sq930_bulk_jpeg_ops = {
	.packet_done = sq930_bulk_jpeg_done,
	.submit_error = sq930_bulk_submit_error,
};



enum {
	SQ930_DI_GENERIC = 0,
	SQ930_DI_CREATIVE_WCLUN,
	SQ930_DI_CREATIVE_WCLU,
	SQ930_DI_CREATIVE_WCLM,
};

static const struct sq930_model *sq930_find_model(int driverinfo);
static int sq930_usbcam_init(struct usbcam_dev *udp,
			      const struct usb_device_id *devid)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	int i, res;

	memset(sdp, 0, sizeof(*sdp));
	usbcam_urbstream_init(&sdp->sqd_bulk, udp, 1);
	usbcam_delayedwork_init(udp, &sdp->sqd_pt_work, sq930_pantilt_poll);

	sdp->sqd_parent = udp;
	sdp->sqd_pt_reqspeed = 11;
	sdp->sqd_pt_reqspeed_reset = 25;
	sdp->sqd_aec = 1;
	sdp->sqd_pt_lastaxis = SQ930_AXIS_NONE;
	sdp->sqd_gpio = 0xffffffff;
	sdp->sqd_timeout = 1000;
	sdp->sqd_frame_meta_enable = 1;

	sdp->sqd_model = sq930_find_model(devid->driver_info);

	if (!sdp->sqd_model) {
		sq_err(sdp, "Could not find appropriate model for %04x:%04x",
		       le16_to_cpu(sq930_dev(sdp)->descriptor.idVendor),
		       le16_to_cpu(sq930_dev(sdp)->descriptor.idProduct));
		return -ENODEV;
	}

#if 0
	res = sq930_control(sdp, 1, SQ930_CTRL_RESET, 0, NULL, 0);
	if (res) {
		sq_err(sdp, "Could not issue reset command: %d", res);
		goto failed;
	}
#endif

	res = sq930_control(sdp, 0, SQ930_CTRL_GET_DEV_INFO, 0,
			    sdp->sqd_dev_info, sizeof(sdp->sqd_dev_info));
	if (res) {
		sq_err(sdp, "Could not query device info: %d", res);
		goto failed;
	}

	/* Parse out the known interesting details */
	sdp->sqd_info_highspeed = (sdp->sqd_dev_info[0] == 3) ? 1 : 0;
	sdp->sqd_info_fw_ver = sdp->sqd_dev_info[2];
	sdp->sqd_info_chip_model =
		be16_to_cpup((__be16*)&sdp->sqd_dev_info[3]);
	sdp->sqd_info_cmos = (sdp->sqd_dev_info[5] & 8) ? 1 : 0;

	if (sq930_dev(sdp)->product)
		sq_info(sdp, "Product: %s", sq930_dev(sdp)->product);

	if (!sq930_dev(sdp)->product ||
	    strcmp(sq930_dev(sdp)->product, sdp->sqd_model->sqm_name))
		sq_info(sdp, "Device model: %s", sdp->sqd_model->sqm_name);

	if ((sdp->sqd_info_chip_model != 0x930b) &&
	    (sdp->sqd_info_chip_model != 0x930c)) {
		sq_err(sdp, "Device reports type %04x, huh?",
		       sdp->sqd_info_chip_model);
		res = -ENODEV;
		goto failed;
	}

	if (!sdp->sqd_info_highspeed &&
	    (udp->ud_dev->speed != USB_SPEED_FULL)) {
		sq_err(sdp, "Device reports full speed, bus reports %d",
		       udp->ud_dev->speed);
	} else if (sdp->sqd_info_highspeed &&
		   (udp->ud_dev->speed != USB_SPEED_HIGH)) {
		sq_err(sdp, "Device reports high speed, bus reports %d",
		       udp->ud_dev->speed);
	}

	sq_info(sdp, "Chip: %04x FW: %02x",
		sdp->sqd_info_chip_model, sdp->sqd_info_fw_ver);

	res = sq930_ucreg_clkfreq_set(sdp, SQ930_CLKFREQ_60MHZ);
	if (res) {
		sq_err(sdp, "Error setting clock speed: %d", res);
		goto failed;
	}

	sdp->sqd_sensor = sdp->sqd_model->sqm_sensor;
	if (!sdp->sqd_sensor) {
		/* Probe for the damn thing */
		for (i = 0; sq930_sensors[i] != NULL; i++) {
			if (sq930_sensors[i]->sqs_probe &&
			    !sq930_sensors[i]->sqs_probe(sdp))
				break;
		}
		sdp->sqd_sensor = sq930_sensors[i];

		if (!sdp->sqd_sensor) {
			sq_err(sdp, "Could not determine image sensor");
			res = -ENODEV;
			goto failed;
		}
	}

	sq_info(sdp, "Sensor: %s", sdp->sqd_sensor->sqs_name);

	sdp->sqd_fmts = sq930_fmts;
	sdp->sqd_nfmts = ARRAY_SIZE(sq930_fmts);

	if (sdp->sqd_model->sqm_gpio_extra) {
		res = sq930_gpio_set(sdp,
				     sdp->sqd_model->sqm_gpio_extra,
				     sdp->sqd_model->sqm_gpio_extra,
				     0);
		if (res) {
			sq_err(sdp, "Error setting extra GPIOs: %d", res);
			goto failed;
		}
	}

	res = sq930_pt_init(sdp);
	if (res) {
		sq_err(sdp, "Error initializing pan/tilt: %d", res);
		goto failed;
	}

	if (sdp->sqd_sensor->sqs_init) {
		res = sdp->sqd_sensor->sqs_init(sdp);
		if (res) {
			sq_err(sdp, "Image sensor init failed: %d", res);
			goto failed;
		}
	}

	(void) sq930_led_set(sdp, 0);

	sdp->sqd_fmt = &sdp->sqd_fmts[0];
	sdp->sqd_res = &sdp->sqd_fmt->sqf_res_tbl[0];

	udp->ud_fmt_array = &sdp->sqd_fmts[0].base;
	udp->ud_fmt_array_len = sdp->sqd_nfmts;
	udp->ud_fmt_array_elem_size = sizeof(sdp->sqd_fmts[0]);

	udp->ud_format.width = sdp->sqd_res->sqr_width;
	udp->ud_format.height = sdp->sqd_res->sqr_height;
	udp->ud_format.pixelformat = sdp->sqd_fmt->base.pixelformat;
	udp->ud_format.field = V4L2_FIELD_INTERLACED;

	if (!sdp->sqd_fmt->sqf_bpp) {
		udp->ud_format.bytesperline = 0;
		udp->ud_format.sizeimage = 307200;
	} else {
		udp->ud_format.bytesperline = (udp->ud_format.width *
					       sdp->sqd_fmt->sqf_bpp);
		udp->ud_format.sizeimage = (udp->ud_format.bytesperline *
					    udp->ud_format.height);
	}
	udp->ud_format.colorspace = V4L2_COLORSPACE_SMPTE170M;

	sdp->sqd_exposure = 200;

	if (sdp->sqd_pt_enable) {
		res = usbcam_work_ref(sdp->sqd_parent);
		if (res)
			goto failed;
		if (ptreset) {
			/* Start the pan/tilt reset process */
			sdp->sqd_pt[SQ930_AXIS_PAN].sqa_reqreset = 1;
			sdp->sqd_pt[SQ930_AXIS_TILT].sqa_reqreset = 1;
			res = sq930_pantilt_continue(sdp);
			if (res)
				goto failed_work;
		}
	}

	return 0;

failed_work:
	if (sdp->sqd_pt_enable) {
		usbcam_delayedwork_cancel(&sdp->sqd_pt_work);
		usbcam_work_unref(sdp->sqd_parent);
	}
failed:
	return res;
}

static void sq930_usbcam_disconnect(struct usbcam_dev *udp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	sq930_do_stop(sdp);

	if (sdp->sqd_pt_enable) {
		usbcam_delayedwork_cancel(&sdp->sqd_pt_work);
		usbcam_work_unref(sdp->sqd_parent);
	}
}

static int sq930_try_format(struct sq930_dev *sdp,
			    struct v4l2_pix_format *f,
			    const struct sq930_fmt **fmt_out,
			    const struct sq930_resolution **res_out)
{
	const struct sq930_fmt *fmt;
	const struct sq930_resolution *res, *restbl;
	int i;

	fmt = NULL;
	for (i = 0; i < sdp->sqd_nfmts; i++) {
		if (sdp->sqd_fmts[i].base.pixelformat == f->pixelformat) {
			fmt = &sdp->sqd_fmts[i];
			break;
		}
	}

	if (!fmt)
		return -EINVAL;

	restbl = fmt->sqf_res_tbl;
	if (!restbl || !restbl[0].sqr_width) {
		sq_err(sdp, "invalid resolution table");
		return -EINVAL;
	}

	/* Find the most acceptable resolution */
	res = NULL;
	for (i = 0; restbl[i].sqr_width > 0; i++) {
		if (!res) {
			res = &restbl[i];
		}
		else if (res->sqr_width > f->width) {
			if (restbl[i].sqr_width < res->sqr_width)
				res = &restbl[i];
		}
		else if (res->sqr_height > f->height) {
			if ((restbl[i].sqr_width <= f->width) &&
			    (restbl[i].sqr_height < res->sqr_height))
				res = &restbl[i];
		}
		else if ((restbl[i].sqr_width <= f->width) &&
			 (restbl[i].sqr_height <= f->height) &&
			 ((restbl[i].sqr_width > res->sqr_width) ||
			  ((restbl[i].sqr_width == res->sqr_width) &&
			   (restbl[i].sqr_height > res->sqr_height)))) {
			res = &restbl[i];
		}
	}

	if ((f->width > 1) && (f->height > 1) &&
	    (f->width != 48) && (f->height != 32) &&
	    ((res->sqr_width > f->width) ||
	     (res->sqr_height > f->height)))
		return -EINVAL;

	memset(f, 0, sizeof(*f));

	f->width = res->sqr_width;
	f->height = res->sqr_height;
	f->pixelformat = fmt->base.pixelformat;
	f->field = V4L2_FIELD_INTERLACED;
	f->colorspace = V4L2_COLORSPACE_SMPTE170M;

	if (!fmt->sqf_bpp) {
		f->bytesperline = 0;
		f->sizeimage = 307200;
	} else {
		f->bytesperline = f->width * fmt->sqf_bpp;
		f->sizeimage = f->bytesperline * f->height;
	}

	if (fmt_out)
		(*fmt_out) = fmt;
	if (res_out)
		(*res_out) = res;

	return 0;
}

static int sq930_usbcam_try_format(struct usbcam_dev *udp,
				   struct v4l2_pix_format *f)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	return sq930_try_format(sdp, f, NULL, NULL);
}

static int sq930_usbcam_set_format(struct usbcam_dev *udp,
				   struct v4l2_pix_format *f)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	int res;
	res = sq930_try_format(sdp, f, &sdp->sqd_fmt, &sdp->sqd_res);
	if (!res)
		sdp->sqd_parent->ud_format = *f;
	return res;
}

static int sq930_usbcam_cap_start(struct usbcam_dev *udp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	int res;

	(void) sq930_led_set(sdp, 1);

	sdp->sqd_quant = (sdp->sqd_res->sqr_width >= 640) ? 3 : 1;
	sdp->sqd_ppl = sdp->sqd_res->sqr_minppl;

	res = sdp->sqd_sensor->sqs_cap_start(sdp);
	if (res) {
		sq_err(sdp, "Image sensor cap_start failed: %d", res);
		goto out_led;
	}

	res = usbcam_urbstream_config_bulk(&sdp->sqd_bulk,
					   (sdp->sqd_fmt->base.flags &
					    V4L2_FMT_FLAG_COMPRESSED)
					   ? &sq930_bulk_jpeg_ops
					   : &sq930_bulk_raw_ops,
					   4, sdp->sqd_res->sqr_reqsize,
					   sdp->sqd_res->sqr_maxpkt,
					   1000);
	if (res) {
		sq_err(sdp, "Error configuring bulk transport: %d", res);
		goto out_led;
	}

	sdp->sqd_frame_off = 0;
	sdp->sqd_useless_pkts = 0;
	sdp->sqd_nab = 0;

	udp->ud_capturing = 1;

	res = usbcam_urbstream_start(&sdp->sqd_bulk);
	if (res)
		sq930_do_stop(sdp);

	return res;

out_led:
	sq930_led_set(sdp, 0);
	return res;
}

static void sq930_usbcam_cap_stop(struct usbcam_dev *udp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	sq930_do_stop(sdp);
}


static struct usbcam_dev_ops sq930_usbcam_dev_ops = {
	.init		= sq930_usbcam_init,
	.disconnect	= sq930_usbcam_disconnect,
	.try_format	= sq930_usbcam_try_format,
	.set_format	= sq930_usbcam_set_format,
	.cap_start	= sq930_usbcam_cap_start,
	.cap_stop	= sq930_usbcam_cap_stop,
};

/*
 * Available image sensors, defined in external modules
 */
extern const struct sq930_sensor sq930_sensor_lz24bp;

static const struct sq930_sensor *sq930_sensors[] = {
	&sq930_sensor_lz24bp,
	NULL,
};

/*
 * Known camera models and characteristics
 */
static const struct sq930_model sq930_models[] = {
	[SQ930_DI_GENERIC] = {
		.sqm_name = "Generic SQ930B/C Webcam",
	},
	[SQ930_DI_CREATIVE_WCLU] = {
		.sqm_name = "Creative WebCam Live! Ultra",
		.sqm_sensor = &sq930_sensor_lz24bp,
	},
	[SQ930_DI_CREATIVE_WCLUN] = {
		.sqm_name = "Creative WebCam Live! Ultra for Notebooks",
		.sqm_sensor = &sq930_sensor_lz24bp,
		.sqm_gpio_extra = 0x02004000,
	},
	[SQ930_DI_CREATIVE_WCLM] = {
		.sqm_name = "Creative WebCam Live! Motion",
		.sqm_sensor = &sq930_sensor_lz24bp,
		.sqm_gpio_isresetbar = -1,	/* GPIO 10 used by P/T */
		.sqm_pt_axes = {
			/*
			 * The WCLM has a few reversals
			 * The angle unit appears to be 1/2 degrees
			 */
			[SQ930_AXIS_PAN] = {
				.sqa_devaxis = 2,
				.sqa_min = -140,
				.sqa_max = 140,
			},
			[SQ930_AXIS_TILT] = {
				.sqa_devaxis = 1,
				.sqa_min = -60,
				.sqa_max = 60,
				.sqa_invert = 1,
			}
		},
	},
};

static const struct sq930_model *sq930_find_model(int driverinfo)
{
	if ((driverinfo < 0) || (driverinfo > ARRAY_SIZE(sq930_models)))
		return NULL;
	return &sq930_models[driverinfo];
}

static const struct usb_device_id id_table[] = {
	{ USB_DEVICE(0x041E, 0x403C), .driver_info = SQ930_DI_CREATIVE_WCLU },
	{ USB_DEVICE(0x041E, 0x403D), .driver_info = SQ930_DI_CREATIVE_WCLUN },
	{ USB_DEVICE(0x041E, 0x4041), .driver_info = SQ930_DI_CREATIVE_WCLM },
	{ USB_DEVICE(0x2770, 0x930B), .driver_info = SQ930_DI_GENERIC },
	{ USB_DEVICE(0x2770, 0x930C), .driver_info = SQ930_DI_GENERIC },
	{ }
};

DEFINE_USBCAM_MINIDRV_MODULE(SQ930_VERSION,
			     SQ930_VERSION_EXTRA,
			     &sq930_usbcam_dev_ops,
			     sizeof(struct sq930_dev),
			     id_table)

//MODULE_DEVICE_TABLE(usb, id_table);
MODULE_DESCRIPTION("Driver for SQ930B/C based Webcams");
MODULE_AUTHOR("Sam Revitch <samr7@cs.washington.edu>");
MODULE_LICENSE("GPL");
