/*
 * Driver for SQ930B/C based webcams
 * Copyright (C) 2007 Sam Revitch <samr7@cs.washington.edu>
 *
 * This driver is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This driver is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this driver; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

/*
 * This file supports SQ930B-based devices with a Sharp LZ24BP CCD
 * image sensor.
 */

#include "sq930.h"

/*
 * We store shutter timing and gain values in three tables.  In two of
 * the three tables, the timing values also encode flicker suppression
 * for indoor lighting.
 */
struct lz24bp_exposure_vals {
	uint16_t shutter;
	uint8_t gain;
};
static const struct lz24bp_exposure_vals lz24bp_std_tbl[] = {
	{ 0x1,   0x40 }, { 0x1,   0x42 }, { 0x1,   0x44 }, { 0x1,   0x46 },
	{ 0x1,   0x48 }, { 0x1,   0x4B }, { 0x1,   0x4D }, { 0x1,   0x50 },
	{ 0x1,   0x52 }, { 0x1,   0x55 }, { 0x1,   0x58 }, { 0x1,   0x5B },
	{ 0x1,   0x5E }, { 0x1,   0x61 }, { 0x1,   0x64 }, { 0x1,   0x67 },
	{ 0x1,   0x6B }, { 0x1,   0x6E }, { 0x1,   0x72 }, { 0x1,   0x75 },
	{ 0x1,   0x79 }, { 0x1,   0x7D }, { 0x2,   0x41 }, { 0x2,   0x43 },
	{ 0x2,   0x45 }, { 0x2,   0x47 }, { 0x2,   0x4A }, { 0x2,   0x4C },
	{ 0x2,   0x4F }, { 0x2,   0x51 }, { 0x2,   0x54 }, { 0x2,   0x57 },
	{ 0x2,   0x5A }, { 0x2,   0x5C }, { 0x3,   0x40 }, { 0x3,   0x42 },
	{ 0x3,   0x44 }, { 0x3,   0x46 }, { 0x3,   0x48 }, { 0x3,   0x4B },
	{ 0x3,   0x4D }, { 0x3,   0x50 }, { 0x3,   0x53 }, { 0x4,   0x40 },
	{ 0x4,   0x42 }, { 0x4,   0x44 }, { 0x4,   0x47 }, { 0x4,   0x49 },
	{ 0x4,   0x4B }, { 0x4,   0x4E }, { 0x5,   0x40 }, { 0x5,   0x42 },
	{ 0x5,   0x45 }, { 0x5,   0x47 }, { 0x5,   0x49 }, { 0x5,   0x4C },
	{ 0x6,   0x41 }, { 0x6,   0x43 }, { 0x6,   0x46 }, { 0x6,   0x48 },
	{ 0x7,   0x40 }, { 0x7,   0x42 }, { 0x7,   0x44 }, { 0x7,   0x46 },
	{ 0x7,   0x49 }, { 0x8,   0x42 }, { 0x8,   0x44 }, { 0x8,   0x46 },
	{ 0x9,   0x40 }, { 0x9,   0x42 }, { 0x9,   0x45 }, { 0xA,   0x40 },
	{ 0xA,   0x42 }, { 0xA,   0x44 }, { 0xB,   0x40 }, { 0xB,   0x42 },
	{ 0xB,   0x44 }, { 0xC,   0x41 }, { 0xC,   0x43 }, { 0xD,   0x40 },
	{ 0xD,   0x42 }, { 0xD,   0x44 }, { 0xE,   0x41 }, { 0xE,   0x43 },
	{ 0xF,   0x41 }, { 0xF,   0x43 }, { 0x10,  0x41 }, { 0x10,  0x43 },
	{ 0x11,  0x41 }, { 0x12,  0x40 }, { 0x12,  0x42 }, { 0x13,  0x41 },
	{ 0x13,  0x43 }, { 0x14,  0x41 }, { 0x15,  0x40 }, { 0x16,  0x40 },
	{ 0x16,  0x42 }, { 0x17,  0x41 }, { 0x18,  0x40 }, { 0x19,  0x40 },
	{ 0x19,  0x42 }, { 0x1A,  0x41 }, { 0x1B,  0x41 }, { 0x1C,  0x41 },
	{ 0x1D,  0x41 }, { 0x1E,  0x41 }, { 0x1F,  0x41 }, { 0x20,  0x41 },
	{ 0x21,  0x41 }, { 0x22,  0x41 }, { 0x23,  0x41 }, { 0x25,  0x40 },
	{ 0x26,  0x40 }, { 0x27,  0x41 }, { 0x29,  0x40 }, { 0x2A,  0x40 },
	{ 0x2B,  0x41 }, { 0x2D,  0x40 }, { 0x2E,  0x41 }, { 0x30,  0x40 },
	{ 0x31,  0x41 }, { 0x33,  0x40 }, { 0x35,  0x40 }, { 0x37,  0x40 },
	{ 0x38,  0x41 }, { 0x3A,  0x40 }, { 0x3C,  0x40 }, { 0x3E,  0x40 },
	{ 0x40,  0x40 }, { 0x43,  0x40 }, { 0x45,  0x40 }, { 0x47,  0x40 },
	{ 0x49,  0x40 }, { 0x4C,  0x40 }, { 0x4E,  0x40 }, { 0x51,  0x40 },
	{ 0x54,  0x40 }, { 0x57,  0x40 }, { 0x59,  0x40 }, { 0x5C,  0x40 },
	{ 0x60,  0x40 }, { 0x63,  0x40 }, { 0x66,  0x40 }, { 0x6A,  0x40 },
	{ 0x6D,  0x40 }, { 0x71,  0x40 }, { 0x74,  0x40 }, { 0x78,  0x40 },
	{ 0x7C,  0x40 }, { 0x81,  0x40 }, { 0x85,  0x40 }, { 0x89,  0x40 },
	{ 0x8E,  0x40 }, { 0x93,  0x40 }, { 0x97,  0x40 }, { 0x9D,  0x40 },
	{ 0xA2,  0x40 }, { 0xA7,  0x40 }, { 0xAD,  0x40 }, { 0xB3,  0x40 },
	{ 0xB8,  0x40 }, { 0xBF,  0x40 }, { 0xC5,  0x40 }, { 0x0CC, 0x40 },
	{ 0x0D2, 0x40 }, { 0x0D9, 0x40 }, { 0x0E1, 0x40 }, { 0x0E8, 0x40 },
	{ 0x0F0, 0x40 }, { 0x0F8, 0x40 }, { 0x100, 0x40 }, { 0x109, 0x40 },
	{ 0x112, 0x40 }, { 0x11B, 0x40 }, { 0x124, 0x40 }, { 0x12E, 0x40 },
	{ 0x138, 0x40 }, { 0x142, 0x40 }, { 0x14D, 0x40 }, { 0x158, 0x40 },
	{ 0x164, 0x40 }, { 0x170, 0x40 }, { 0x17C, 0x40 }, { 0x188, 0x40 },
	{ 0x196, 0x40 }, { 0x1A3, 0x40 }, { 0x1B1, 0x40 }, { 0x1BF, 0x40 },
	{ 0x1CE, 0x40 }, { 0x1DE, 0x40 }, { 0x1EE, 0x40 }, { 0x1FE, 0x40 },
	{ 0x20F, 0x40 }, { 0x221, 0x40 }, { 0x233, 0x40 }, { 0x246, 0x40 },
	{ 0x259, 0x40 }, { 0x26D, 0x40 }, { 0x26D, 0x42 }, { 0x26D, 0x44 },
	{ 0x26D, 0x46 }, { 0x26D, 0x49 }, { 0x26D, 0x4B }, { 0x26D, 0x4D },
	{ 0x26D, 0x50 }, { 0x26D, 0x53 }, { 0x26D, 0x56 }, { 0x26D, 0x58 },
	{ 0x26D, 0x5B }, { 0x26D, 0x5E }, { 0x26D, 0x62 }, { 0x26D, 0x65 },
	{ 0x26D, 0x68 }, { 0x26D, 0x6C }, { 0x26D, 0x6F }, { 0x26D, 0x73 },
	{ 0x26D, 0x77 }, { 0x26D, 0x7B }, { 0x26D, 0x7F }, { 0x27E, 0x80 },
	{ 0x294, 0x80 }, { 0x2AA, 0x80 }, { 0x2C0, 0x80 }, { 0x2D8, 0x80 },
	{ 0x2F0, 0x80 }, { 0x309, 0x80 }, { 0x323, 0x80 }, { 0x33E, 0x80 },
	{ 0x359, 0x80 }, { 0x376, 0x80 }, { 0x394, 0x80 }, { 0x3B2, 0x80 },
	{ 0x3D2, 0x80 }, { 0x3F2, 0x80 }, { 0x414, 0x80 }, { 0x437, 0x80 },
	{ 0x45B, 0x80 }, { 0x480, 0x80 }, { 0x4A6, 0x80 }, { 0x4CE, 0x80 },
	{ 0x4F7, 0x80 }, { 0x521, 0x80 }, { 0x54D, 0x80 }, { 0x57A, 0x80 },
	{ 0x5A9, 0x80 }, { 0x5D9, 0x80 }, { 0x60B, 0x80 }, { 0x63E, 0x80 },
	{ 0x673, 0x80 }, { 0x6AA, 0x80 }, { 0x6E3, 0x80 }, { 0x71E, 0x80 },
	{ 0x75B, 0x80 }, { 0x799, 0x80 }, { 0x7DA, 0x80 }, { 0x81D, 0x80 },
	{ 0x862, 0x80 }, { 0x8AA, 0x80 }, { 0x8F4, 0x80 }, { 0x940, 0x80 },
	{ 0x98F, 0x80 }, { 0x9E0, 0x80 }, { 0xA35, 0x80 }, { 0xA8C, 0x80 },
	{ 0xAE6, 0x80 }, { 0xB43, 0x80 }, { 0xBA3, 0x80 }, { 0xC06, 0x80 },
	{ 0xC6C, 0x80 }, { 0xCD6, 0x80 }, { 0xD44, 0x80 }, { 0xDB5, 0x80 },
	{ 0xE2A, 0x80 }, { 0xEA2, 0x80 }, { 0xF1F, 0x80 }, { 0xFA0, 0x80 },
	{ 0xFFF, 0x84 }, { 0xFFF, 0x88 }, { 0xFFF, 0x8D }, { 0xFFF, 0x91 },
	{ 0xFFF, 0x96 }, { 0xFFF, 0x9B }, { 0xFFF, 0xA1 }, { 0xFFF, 0xA6 },
	{ 0xFFF, 0xAB }, { 0xFFF, 0xB1 }, { 0xFFF, 0xB7 }, { 0xFFF, 0xBD },
	{ 0xFFF, 0xC3 }, { 0xFFF, 0xCA }, { 0xFFF, 0xD1 }, { 0xFFF, 0xD8 },
	{ 0xFFF, 0xDF }, { 0xFFF, 0xE6 }, { 0xFFF, 0xEE }, { 0xFFF, 0xF6 },
	{ 0xFFF, 0xFE },
};
static const struct lz24bp_exposure_vals lz24bp_50hz_tbl[] = {
	{ 0xBF,  0x40 }, { 0xBF,  0x42 }, { 0xBF,  0x44 }, { 0xBF,  0x46 },
	{ 0xBF,  0x49 }, { 0xBF,  0x4B }, { 0xBF,  0x4E }, { 0xBF,  0x50 },
	{ 0xBF,  0x53 }, { 0xBF,  0x56 }, { 0xBF,  0x58 }, { 0xBF,  0x5B },
	{ 0xBF,  0x5E }, { 0xBF,  0x62 }, { 0xBF,  0x65 }, { 0xBF,  0x68 },
	{ 0xBF,  0x6C }, { 0xBF,  0x6F }, { 0xBF,  0x73 }, { 0xBF,  0x77 },
	{ 0xBF,  0x7B }, { 0xBF,  0x7F }, { 0x17E, 0x41 }, { 0x17E, 0x44 },
	{ 0x17E, 0x46 }, { 0x17E, 0x48 }, { 0x17E, 0x4B }, { 0x17E, 0x4D },
	{ 0x17E, 0x50 }, { 0x17E, 0x52 }, { 0x17E, 0x55 }, { 0x17E, 0x58 },
	{ 0x17E, 0x5B }, { 0x17E, 0x5E }, { 0x23D, 0x41 }, { 0x23D, 0x43 },
	{ 0x23D, 0x45 }, { 0x23D, 0x47 }, { 0x23D, 0x4A }, { 0x23D, 0x4C },
	{ 0x23D, 0x4F }, { 0x23D, 0x51 }, { 0x23D, 0x54 }, { 0x23D, 0x57 },
	{ 0x23D, 0x5A }, { 0x23D, 0x5D }, { 0x23D, 0x60 }, { 0x23D, 0x63 },
	{ 0x23D, 0x66 }, { 0x23D, 0x6A }, { 0x23D, 0x6D }, { 0x23D, 0x71 },
	{ 0x23D, 0x75 }, { 0x23D, 0x79 }, { 0x23D, 0x7D }, { 0x23D, 0x81 },
	{ 0x23D, 0x85 }, { 0x23D, 0x8A }, { 0x23D, 0x8E }, { 0x23D, 0x93 },
	{ 0x23D, 0x98 }, { 0x23D, 0x9D }, { 0x23D, 0xA2 }, { 0x23D, 0xA8 },
	{ 0x2FD, 0x82 }, { 0x2FD, 0x86 }, { 0x2FD, 0x8B }, { 0x2FD, 0x8F },
	{ 0x2FD, 0x94 }, { 0x2FD, 0x99 }, { 0x2FD, 0x9E }, { 0x3BC, 0x83 },
	{ 0x3BC, 0x87 }, { 0x3BC, 0x8B }, { 0x3BC, 0x90 }, { 0x3BC, 0x95 },
	{ 0x47B, 0x80 }, { 0x47B, 0x84 }, { 0x47B, 0x89 }, { 0x47B, 0x8D },
	{ 0x47B, 0x92 }, { 0x53B, 0x81 }, { 0x53B, 0x86 }, { 0x53B, 0x8A },
	{ 0x53B, 0x8F }, { 0x5FA, 0x81 }, { 0x5FA, 0x85 }, { 0x5FA, 0x8A },
	{ 0x5FA, 0x8E }, { 0x6B9, 0x83 }, { 0x6B9, 0x87 }, { 0x6B9, 0x8C },
	{ 0x779, 0x82 }, { 0x779, 0x86 }, { 0x779, 0x8B }, { 0x838, 0x82 },
	{ 0x838, 0x87 }, { 0x838, 0x8B }, { 0x8F7, 0x84 }, { 0x8F7, 0x88 },
	{ 0x9B7, 0x82 }, { 0x9B7, 0x86 }, { 0xA76, 0x81 }, { 0xA76, 0x85 },
	{ 0xB35, 0x80 }, { 0xB35, 0x85 }, { 0xBF5, 0x80 }, { 0xBF5, 0x85 },
	{ 0xCB4, 0x81 }, { 0xCB4, 0x85 }, { 0xD73, 0x82 }, { 0xD73, 0x86 },
	{ 0xE33, 0x84 }, { 0xEF2, 0x81 }, { 0xEF2, 0x85 }, { 0xFB1, 0x83 },
};
static const struct lz24bp_exposure_vals lz24bp_60hz_tbl[] = {
	{ 0x9F,  0x40 }, { 0x9F,  0x42 }, { 0x9F,  0x44 }, { 0x9F,  0x46 },
	{ 0x9F,  0x49 }, { 0x9F,  0x4B }, { 0x9F,  0x4E }, { 0x9F,  0x50 },
	{ 0x9F,  0x53 }, { 0x9F,  0x56 }, { 0x9F,  0x59 }, { 0x9F,  0x5C },
	{ 0x9F,  0x5F }, { 0x9F,  0x62 }, { 0x9F,  0x65 }, { 0x9F,  0x68 },
	{ 0x9F,  0x6C }, { 0x9F,  0x6F }, { 0x9F,  0x73 }, { 0x9F,  0x77 },
	{ 0x9F,  0x7B }, { 0x9F,  0x7F }, { 0x13E, 0x41 }, { 0x13E, 0x44 },
	{ 0x13E, 0x46 }, { 0x13E, 0x48 }, { 0x13E, 0x4B }, { 0x13E, 0x4D },
	{ 0x13E, 0x50 }, { 0x13E, 0x52 }, { 0x13E, 0x55 }, { 0x13E, 0x58 },
	{ 0x13E, 0x5B }, { 0x13E, 0x5E }, { 0x1DE, 0x41 }, { 0x1DE, 0x43 },
	{ 0x1DE, 0x45 }, { 0x1DE, 0x47 }, { 0x1DE, 0x4A }, { 0x1DE, 0x4C },
	{ 0x1DE, 0x4F }, { 0x1DE, 0x51 }, { 0x1DE, 0x54 }, { 0x27D, 0x41 },
	{ 0x27D, 0x43 }, { 0x27D, 0x45 }, { 0x27D, 0x48 }, { 0x27D, 0x4A },
	{ 0x27D, 0x4D }, { 0x27D, 0x4F }, { 0x27D, 0x52 }, { 0x27D, 0x55 },
	{ 0x27D, 0x57 }, { 0x27D, 0x5A }, { 0x27D, 0x5D }, { 0x27D, 0x61 },
	{ 0x27D, 0x64 }, { 0x27D, 0x67 }, { 0x27D, 0x6B }, { 0x27D, 0x6E },
	{ 0x27D, 0x72 }, { 0x27D, 0x76 }, { 0x27D, 0x7A }, { 0x27D, 0x7E },
	{ 0x27D, 0x82 }, { 0x27D, 0x86 }, { 0x27D, 0x8B }, { 0x27D, 0x8F },
	{ 0x27D, 0x94 }, { 0x27D, 0x99 }, { 0x27D, 0x9E }, { 0x31D, 0x83 },
	{ 0x31D, 0x87 }, { 0x31D, 0x8B }, { 0x31D, 0x90 }, { 0x31D, 0x95 },
	{ 0x3BC, 0x80 }, { 0x3BC, 0x84 }, { 0x3BC, 0x89 }, { 0x3BC, 0x8D },
	{ 0x3BC, 0x92 }, { 0x45C, 0x81 }, { 0x45C, 0x86 }, { 0x45C, 0x8A },
	{ 0x45C, 0x8F }, { 0x4FB, 0x81 }, { 0x4FB, 0x85 }, { 0x4FB, 0x8A },
	{ 0x4FB, 0x8E }, { 0x59A, 0x83 }, { 0x59A, 0x87 }, { 0x59A, 0x8C },
	{ 0x63A, 0x82 }, { 0x63A, 0x86 }, { 0x63A, 0x8B }, { 0x6D9, 0x82 },
	{ 0x6D9, 0x87 }, { 0x6D9, 0x8B }, { 0x779, 0x84 }, { 0x779, 0x88 },
	{ 0x818, 0x82 }, { 0x818, 0x86 }, { 0x8B8, 0x81 }, { 0x8B8, 0x85 },
	{ 0x957, 0x80 }, { 0x957, 0x85 }, { 0x9F7, 0x80 }, { 0x9F7, 0x85 },
	{ 0xA96, 0x81 }, { 0xA96, 0x85 }, { 0xB35, 0x82 }, { 0xB35, 0x86 },
	{ 0xBD5, 0x84 }, { 0xC74, 0x81 }, { 0xC74, 0x85 }, { 0xD14, 0x83 },
	{ 0xDB3, 0x82 }, { 0xE53, 0x80 }, { 0xE53, 0x84 }, { 0xEF2, 0x83 }
};
static const struct {
	const struct lz24bp_exposure_vals *vals;
	int count;
} lz24bp_exposure_tables[] = {
	{ .vals = lz24bp_std_tbl,
	  .count = ARRAY_SIZE(lz24bp_std_tbl) },
	{ .vals = lz24bp_50hz_tbl,
	  .count = ARRAY_SIZE(lz24bp_50hz_tbl) },
	{ .vals = lz24bp_60hz_tbl,
	  .count = ARRAY_SIZE(lz24bp_60hz_tbl) },
};

static int lz24bp_exposure(struct sq930_dev *sdp, u8 value, int pl_mode)
{
	int ent, integclks, frameclks, gain;

	assert((pl_mode >= 0) && (pl_mode <= 2));

	ent = (value * lz24bp_exposure_tables[pl_mode].count) / 256;
	integclks = lz24bp_exposure_tables[pl_mode].vals[ent].shutter;
	frameclks = (integclks > 0x26f) ? integclks : 0x26f;
	gain = lz24bp_exposure_tables[pl_mode].vals[ent].gain;

	return sq930_exposure_set(sdp, frameclks - integclks, frameclks, gain,
				  0, NULL, 0);
}

static int lz24bp_ppl(struct sq930_dev *sdp, u16 ppl)
{
	struct sq930_ucbuswrite_cmd cmds[2] = {
		{ .bw_addr = 0xf810,
		  .bw_data = (ppl >> 8) & 0xff },
		{ .bw_addr = 0xf811,
		  .bw_data = ppl & 0xff }
	};
	assert(ppl >= 0x310);
	return sq930_ucbus_write(sdp, cmds, ARRAY_SIZE(cmds), 0, 0);
}

static int lz24bp_powerline_ctrl_get(struct usbcam_dev *udp,
				     const struct usbcam_ctrl *ctrlp,
				     struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	cp->value = sdp->sqd_powerline;
	if (!sdp->sqd_parent->ud_capturing)
		return 0;
	return lz24bp_exposure(sdp, sdp->sqd_exposure, sdp->sqd_powerline);
}

static int lz24bp_powerline_ctrl_set(struct usbcam_dev *udp,
				     const struct usbcam_ctrl *ctrlp,
				     const struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);

	sdp->sqd_powerline = cp->value;
	return 0;
}

static int lz24bp_exposure_ctrl_get(struct usbcam_dev *udp,
				    const struct usbcam_ctrl *ctrlp,
				    struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	cp->value = sdp->sqd_exposure;
	return 0;
}

static int lz24bp_exposure_ctrl_set(struct usbcam_dev *udp,
				    const struct usbcam_ctrl *ctrlp,
				    const struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);

	sdp->sqd_exposure = cp->value;
	if (!sdp->sqd_parent->ud_capturing)
		return 0;
	return lz24bp_exposure(sdp, sdp->sqd_exposure, sdp->sqd_powerline);
}

static int lz24bp_aec_ctrl_get(struct usbcam_dev *udp,
			       const struct usbcam_ctrl *ctrlp,
			       struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);
	cp->value = sdp->sqd_aec;
	return 0;
}

static int lz24bp_aec_ctrl_set(struct usbcam_dev *udp,
			       const struct usbcam_ctrl *ctrlp,
			       const struct v4l2_ext_control *cp)
{
	struct sq930_dev *sdp = udp_sq930(udp);

	sdp->sqd_aec = cp->value;
	return 0;
}


static const char *lz24bp_powerline_names[] = { "Off", "50Hz", "60Hz" };
static const struct usbcam_ctrl lz24bp_controls[] = {
	{
		.uc_v4l = { .id = V4L2_CID_SQ930_POWERLINE,
			    .type = V4L2_CTRL_TYPE_MENU,
			    .name = "Power Line Frequency",
			    .maximum = 2,
			    .default_value = 0,
			    .flags = 0 },
		.uc_menu_names = lz24bp_powerline_names,
		.get_fn = lz24bp_powerline_ctrl_get,
		.set_fn = lz24bp_powerline_ctrl_set,
	},
	{
		.uc_v4l = { .id = V4L2_CID_EXPOSURE,
			    .type = V4L2_CTRL_TYPE_INTEGER,
			    .name = "Exposure",
			    .minimum = 0,
			    .maximum = 255,
			    .step = 1,
			    .default_value = 63,
			    .flags = V4L2_CTRL_FLAG_SLIDER },
		.get_fn = lz24bp_exposure_ctrl_get,
		.set_fn = lz24bp_exposure_ctrl_set,
	},
	{
		.uc_v4l = { .id = V4L2_CID_SQ930_AUTO_EXPOSURE,
			    .type = V4L2_CTRL_TYPE_BOOLEAN,
			    .name = "Auto Exposure",
			    .default_value = 1,
			    .flags = 0 },
		.get_fn = lz24bp_aec_ctrl_get,
		.set_fn = lz24bp_aec_ctrl_set,
	},
};


static const struct sq930_ucbuswrite_cmd lz24bp_cap_start_0[] = {
	{ 0x0354, 0x00 }, { 0x03fa, 0x00 }, { 0xf800, 0x02 }, { 0xf801, 0xbe },
	{ 0xf802, 0xc6 }, { 0xf804, 0x00 }, { 0xf808, 0x00 }, { 0xf809, 0x06 },
	{ 0xf80a, 0x01 }, { 0xf80b, 0xfe }, { 0xf807, 0x84 }, { 0xf80c, 0x02 },
	{ 0xf80d, 0xf7 }, { 0xf80e, 0x03 }, { 0xf80f, 0x0b }, { 0xf81c, 0x00 },
	{ 0xf81d, 0x49 }, { 0xf81e, 0x03 }, { 0xf81f, 0x0b }, { 0xf83a, 0x00 },
	{ 0xf83b, 0x01 }, { 0xf83c, 0x00 }, { 0xf83d, 0x6b }, { 0xf810, 0x03 },
	{ 0xf811, 0x10 }, { 0xf812, 0x02 }, { 0xf813, 0x6f }, { 0xf803, 0x00 },
	{ 0xf814, 0x00 }, { 0xf815, 0x44 }, { 0xf816, 0x00 }, { 0xf817, 0x48 },
	{ 0xf818, 0x00 }, { 0xf819, 0x25 }, { 0xf81a, 0x00 }, { 0xf81b, 0x3c },
	{ 0xf82f, 0x03 }, { 0xf820, 0xff }, { 0xf821, 0x0d }, { 0xf822, 0xff },
	{ 0xf823, 0x07 }, { 0xf824, 0xfd }, { 0xf825, 0x07 }, { 0xf826, 0xf0 },
	{ 0xf827, 0x0c }, { 0xf828, 0xff }, { 0xf829, 0x03 }, { 0xf82a, 0xff },
	{ 0xf82b, 0x0c }, { 0xf82c, 0xfc }, { 0xf82d, 0x01 }, { 0xf82e, 0x00 },
	{ 0xf830, 0x00 }, { 0xf831, 0x47 }, { 0xf832, 0x00 }, { 0xf833, 0x00 },
	{ 0xf850, 0x00 }, { 0xf851, 0x00 }, { 0xf852, 0x00 }, { 0xf853, 0x24 },
	{ 0xf854, 0x00 }, { 0xf855, 0x0c }, { 0xf856, 0x00 }, { 0xf857, 0x30 },
	{ 0xf858, 0x00 }, { 0xf859, 0x18 }, { 0xf85a, 0x00 }, { 0xf85b, 0x3c },
	{ 0xf85c, 0x00 }, { 0xf85d, 0x18 }, { 0xf85e, 0x00 }, { 0xf85f, 0x3c },
	{ 0xf860, 0xff }, { 0xf861, 0x37 }, { 0xf862, 0xff }, { 0xf863, 0x1d },
	{ 0xf864, 0xff }, { 0xf865, 0x98 }, { 0xf866, 0xff }, { 0xf867, 0xc0 },
	{ 0xf868, 0x00 }, { 0xf869, 0x37 }, { 0xf86c, 0x02 }, { 0xf86d, 0x1d },
	{ 0xf86a, 0x00 }, { 0xf86b, 0x37 }, { 0xf86e, 0x02 }, { 0xf86f, 0x1d },
	{ 0xf870, 0x01 }, { 0xf871, 0xc6 }, { 0xf872, 0x02 }, { 0xf873, 0x04 },
	{ 0xf874, 0x01 }, { 0xf875, 0xc6 }, { 0xf876, 0x02 }, { 0xf877, 0x04 },
	{ 0xf878, 0x0f }, { 0xf879, 0x0f }, { 0xf87a, 0xff }, { 0xf87b, 0xff },
	{ 0xf800, 0x03 }
},

lz24bp_cap_start_1_gen[] = {
	{ 0xf5f0, 0x00 }, { 0xf5f1, 0xff }, { 0xf5f2, 0x80 }, { 0xf5f3, 0x80 },
	{ 0xf5f4, 0xb3 },
	{ 0xf5f0, 0x40 }, { 0xf5f1, 0xff }, { 0xf5f2, 0x80 }, { 0xf5f3, 0x80 },
	{ 0xf5f4, 0xb3 },
	{ 0xf5fa, 0x00 }, { 0xf5f6, 0x00 }, { 0xf5f7, 0x00 }, { 0xf5f8, 0x00 },
	{ 0xf5f9, 0x00 }
},

lz24bp_cap_start_1_wclm[] = {
	{ 0xf5f0, 0x00 }, { 0xf5f1, 0xff }, { 0xf5f2, 0x88 }, { 0xf5f3, 0x88 },
	{ 0xf5f4, 0xc0 },
	{ 0xf5f0, 0x40 }, { 0xf5f1, 0xff }, { 0xf5f2, 0x88 }, { 0xf5f3, 0x88 },
	{ 0xf5f4, 0xc0 },
	{ 0xf5fa, 0x00 }, { 0xf5f6, 0x00 }, { 0xf5f7, 0x00 }, { 0xf5f8, 0x00 },
	{ 0xf5f9, 0x00 }
},

lz24bp_cap_start_2[] = {
	{ 0xf800, 0x02 }, { 0xf807, 0xff }, { 0xf805, 0x80 }, { 0xf806, 0x00 },
	{ 0xf807, 0x7f }, { 0xf800, 0x03 },
	{ 0xf800, 0x02 }, { 0xf807, 0xff }, { 0xf805, 0x4e }, { 0xf806, 0x00 },
	{ 0xf807, 0x7f }, { 0xf800, 0x03 },
	{ 0xf800, 0x02 }, { 0xf807, 0xff }, { 0xf805, 0xc0 }, { 0xf806, 0x48 },
	{ 0xf807, 0x7f }, { 0xf800, 0x03 },
	{ 0xf800, 0x02 }, { 0xf807, 0xff }, { 0xf805, 0x00 }, { 0xf806, 0x00 },
	{ 0xf807, 0x7f }, { 0xf800, 0x03 }
};

static const struct {
	uint8_t		cc_sizeid;
	uint8_t		cc_isjpeg;
	uint8_t		cc_bytes[16];
	uint8_t		cc_jpegbytes[16];
}
lz24bp_capconfig[] = {
	/* BA81, 160x120 */
	{ 0, 0,
	  { 0x06, 0x21, 0x20, 0x0f, 0x00, 0xa1, 0x02, 0xef, 0x01, 0x00, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },

	/* BA81, 320x240 */
	{ 2, 0,
	  { 0x05, 0x21, 0x20, 0x0f, 0x00, 0xa1, 0x02, 0xef, 0x01, 0x01, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },

	/* BA81, 640x480 */
	{ 4, 0,
	  { 0x01, 0x21, 0x20, 0x0f, 0x00, 0xa1, 0x02, 0xef, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },

	/* JPEG, 160x120 */
	{ 0, 1,
	  { 0x01, 0x1f, 0x20, 0x0e, 0x00, 0x9f, 0x02, 0xee, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x02, 0x8b, 0x00, 0x8b, 0x00, 0x41, 0x01, 0x41, 0x01, 0x41, 0x01,
	    0x05, 0x40, 0x01, 0xf0, 0x00 } },

	/* JPEG, 320x240 */
	{ 2, 1,
	  { 0x01, 0x22, 0x20, 0x0e, 0x00, 0xa2, 0x02, 0xee, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x02, 0xdf, 0x01, 0x00, 0x00, 0x3f, 0x01, 0x3f, 0x01, 0x00, 0x00,
	    0x05, 0x40, 0x01, 0xf0, 0x00 } },

	/* JPEG, 640x480 */
	{ 4, 1,
	  { 0x01, 0x22, 0x20, 0x0e, 0x00, 0xa2, 0x02, 0xf0, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x07, 0xe1, 0x01, 0xe1, 0x01, 0x3f, 0x01, 0x3f, 0x01, 0x3f, 0x01,
	    0x05, 0x80, 0x02, 0xe0, 0x01 } },

#if 0
	/* OV7660 */
	{ 0, 0,
	  { 0x06, 0x03, 0xa0, 0x02, 0x00, 0x83, 0x02, 0xe2, 0x01, 0x00, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },
	{ 2, 0,
	  { 0x05, 0x03, 0xa0, 0x02, 0x00, 0x83, 0x02, 0xe2, 0x01, 0x01, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },
	{ 4, 0,
	  { 0x01, 0x03, 0xa0, 0x02, 0x00, 0x83, 0x02, 0xe2, 0x01, 0x02, 0x00,
	    0x0b, 0x1a, 0x12, 0x7b, 0xca } },
	{ 0, 1,
	  { 0x05, 0x3d, 0x20, 0x0b, 0x00, 0xbd, 0x02, 0x0b, 0x02, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x02, 0x01, 0x01, 0x01, 0x01, 0x9f, 0x00, 0x9f, 0x00, 0x9f, 0x01,
	    0x01, 0x05, 0xa0, 0x00, 0x80 } },
	{ 2, 0,
	  { 0x01, 0x03, 0xa0, 0x02, 0x00, 0x83, 0x02, 0xe2, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x02, 0xdf, 0x01, 0x00, 0x00, 0x3f, 0x01, 0x3f, 0x01, 0x00, 0x00,
	    0x05, 0x40, 0x01, 0xf0, 0x00 } },
	{ 4, 0,
	  { 0x01, 0x03, 0xa0, 0x02, 0x00, 0x83, 0x02, 0xe2, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x07, 0xe1, 0x01, 0xe1, 0x01, 0x3f, 0x01, 0x3f, 0x01, 0x3f, 0x01,
	    0x05, 0x80, 0x02, 0xe0, 0x01 } },

	/* MT9V011 */
	{ 0, 0,
	  { 0x06, 0x02, 0x20, 0x01, 0x20, 0x82, 0x02, 0xe1, 0x01, 0x00, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },
	{ 2, 0,
	  { 0x05, 0x02, 0x20, 0x01, 0x20, 0x82, 0x02, 0xe1, 0x01, 0x01, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },
	{ 4, 0,
	  { 0x01, 0x02, 0x20, 0x01, 0x20, 0x82, 0x02, 0xe1, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },
	{ 0, 1,
	  { 0x05, 0x3d, 0x20, 0x0b, 0x00, 0xbd, 0x02, 0x0b, 0x02, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x02, 0x01, 0x01, 0x01, 0x01, 0x9f, 0x00, 0x9f, 0x00, 0x9f, 0x01,
	    0x05, 0xa0, 0x00, 0x80, 0x00 } },
	{ 2, 1,
	  { 0x01, 0x02, 0x20, 0x01, 0x20, 0x82, 0x02, 0xe1, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x02, 0xdf, 0x01, 0x00, 0x00, 0x3f, 0x01, 0x3f, 0x01, 0x00, 0x00,
	    0x05, 0x40, 0x01, 0xf0, 0x00 } },
	{ 4, 1,
	  { 0x01, 0x02, 0x20, 0x01, 0x20, 0x82, 0x02, 0xe3, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x07, 0xe1, 0x01, 0xe1, 0x01, 0x3f, 0x01, 0x3f, 0x01, 0x3f, 0x01,
	    0x05, 0x80, 0x02, 0xe0, 0x01 } },

	/* OV9630 */
	{ 0, 0,
	  { 0x06, 0x06, 0xa0, 0x02, 0x20, 0x86, 0x02, 0xe2, 0xc1, 0x00, 0x00,
	    0x0c, 0x1a, 0x12, 0x7c, 0xca } },
	{ 2, 0,
	  { 0x01, 0x06, 0xa0, 0x02, 0x20, 0x86, 0x02, 0xe2, 0xc1, 0x01, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },
	{ 4, 0,
	  { 0x01, 0x06, 0xa0, 0x02, 0x20, 0x86, 0x02, 0xe2, 0xc1, 0x02, 0x00,
	    0x0c, 0x1a, 0x12, 0x7c, 0xca } },
	{ 5, 0,
	  { 0x01, 0x3c, 0x80, 0x0b, 0x20, 0xbc, 0x02, 0xeb, 0x01, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },
	{ 8, 0,
	  { 0x01, 0x06, 0xa0, 0x02, 0x20, 0x06, 0x05, 0xc2, 0xc3, 0x03, 0x00,
	    0x08, 0x18, 0x34, 0xe8, 0x78 } },
	{ 9, 0,
	  { 0x01, 0x8d, 0xa0, 0x0b, 0x20, 0x8d, 0x05, 0x0b, 0x04, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 } },

	{ 0, 1,
	  { 0x05, 0x06, 0xa0, 0x02, 0x20, 0x86, 0x02, 0xe4, 0xc1, 0x02, 0x00,
	    0x08, 0x18, 0x12, 0x78, 0xc8 },
	  { 0x02, 0x01, 0x01, 0x01, 0x01, 0x9f, 0x00, 0x9f, 0x00, 0x9f, 0x01,
	    0x05, 0xa0, 0x00, 0x80, 0x00 } },
	{ 2, 1,
	  { 0x01, 0x06, 0xa0, 0x02, 0x20, 0x86, 0x02, 0xe2, 0xc1, 0x01, 0x00,
	    0x0c, 0x1a, 0x12, 0x7c, 0xca },
	  { 0x02, 0xdf, 0x01, 0x00, 0x00, 0x3f, 0x01, 0x3f, 0x01, 0x00, 0x00,
	    0x05, 0x40, 0x01, 0xf0, 0x00 } },
	{ 4, 1,
	  { 0x01, 0x06, 0xa0, 0x02, 0x20, 0x86, 0x02, 0xe4, 0xc1, 0x02, 0x00,
	    0x0c, 0x1a, 0x12, 0x7c, 0xca },
	  { 0x07, 0xe1, 0x01, 0xe1, 0x01, 0x3f, 0x01, 0x3f, 0x01, 0x3f, 0x01,
	    0x05, 0x80, 0x02, 0xe0, 0x01 } },
#endif
};


static int lz24bp_do_init(struct sq930_dev *sdp)
{
	int ispwr, isrst, isrstbar;
	int res;

	ispwr = 25;
	isrst = 0;
	isrstbar = 10;

	if (sdp->sqd_model->sqm_gpio_ispower) {
		ispwr = sdp->sqd_model->sqm_gpio_ispower;
		if (ispwr < 0)
			ispwr = 0;
	}
	if (sdp->sqd_model->sqm_gpio_isreset)
		isrst = sdp->sqd_model->sqm_gpio_isreset;
	if (sdp->sqd_model->sqm_gpio_isresetbar) {
		isrstbar = sdp->sqd_model->sqm_gpio_isresetbar;
		if (isrstbar < 0)
			isrstbar = 0;
	}

	if (ispwr) {
		res = sq930_gpio_set_line(sdp, ispwr, 1, 0);
		if (res)
			return res;
	}

	if (isrst) {
		res = sq930_gpio_set_line(sdp, isrst, 1, 50);
		if (res)
			return res;
	}

	if (isrstbar) {
		res = sq930_gpio_set_line(sdp, isrstbar, 0, 50);
		if (res)
			return res;
	}

	res = sq930_i2c_reset(sdp);
	if (res)
		return res;

	if (isrst) {
		res = sq930_gpio_set_line(sdp, isrst, 0, 100);
		if (res)
			return res;
	}
	if (isrstbar) {
		res = sq930_gpio_set_line(sdp, isrstbar, 1, 100);
		if (res)
			return res;
	}
	return 0;
}


static int lz24bp_probe(struct sq930_dev *sdp)
{
	if (sdp->sqd_info_cmos)
		return -ENODEV;
	sq_info(sdp, "Device claims to have a CCD, assuming LZ24BP");
	return 0;
}

static int lz24bp_init(struct sq930_dev *sdp)
{
	int i, res;

	/* Run the first pre-capture init command (Creative driver does it) */
	res = sq930_ucbus_write(sdp, lz24bp_cap_start_0, 8, 0, 1);
	if (res)
		return res;

	/* Run the basic init sequence */
	res = lz24bp_do_init(sdp);
	if (res) {
		sq_err(sdp, "lz24bp_do_init failed: %d", res);
		return res;
	}

	/* Add controls */
	for (i = 0; i < ARRAY_SIZE(lz24bp_controls); i++) {
		if (!usbcam_ctrl_add_tmpl(sdp->sqd_parent,
					  &lz24bp_controls[i], 0))
			return -EINVAL;
	}

	return 0;
}

static int lz24bp_cap_start(struct sq930_dev *sdp)
{
	int res;
	int capent;

	/* Find a capture configuration that best suits the chosen fmt/res */
	for (capent = 0; capent < ARRAY_SIZE(lz24bp_capconfig); capent++) {
		if (sdp->sqd_res->sqr_sizeid !=
		    lz24bp_capconfig[capent].cc_sizeid)
			continue;
		if (sdp->sqd_fmt->base.flags & V4L2_FMT_FLAG_COMPRESSED) {
			if (!lz24bp_capconfig[capent].cc_isjpeg)
				continue;
		} else {
			if (lz24bp_capconfig[capent].cc_isjpeg)
				continue;
		}
		break;
	}

	if (capent >= ARRAY_SIZE(lz24bp_capconfig)) {
		sq_err(sdp, "lz24bp_cap_start: unsuitable configuration");
		return -EINVAL;
	}

	/* Run the basic init sequence */
	res = lz24bp_do_init(sdp);
	if (res) {
		sq_err(sdp, "lz24bp_do_init failed: %d", res);
		return res;
	}

	/* Run the pre-capture init sequence */
	if ((res = sq930_ucbus_write(sdp, lz24bp_cap_start_0,
				     ARRAY_SIZE(lz24bp_cap_start_0), 8, 1)) ||
	    (res = (sdp->sqd_model->sqm_wclm_f5f0
		    ? sq930_ucbus_write(sdp, lz24bp_cap_start_1_wclm,
					ARRAY_SIZE(lz24bp_cap_start_1_wclm),
					5, 1)
		    : sq930_ucbus_write(sdp, lz24bp_cap_start_1_gen,
					ARRAY_SIZE(lz24bp_cap_start_1_gen),
					5, 1))) ||
	    (res = sq930_ucbus_write(sdp, lz24bp_cap_start_2,
				     ARRAY_SIZE(lz24bp_cap_start_2), 6, 1))) {
		sq_err(sdp, "ucbus_write sequence failed: %d", res);
		return res;
	}

	res = lz24bp_ppl(sdp, sdp->sqd_ppl);
	if (res) {
		sq_err(sdp, "lz24bp_ppl failed: %d", res);
		return res;
	}

	/* Send the start capture command */
	res = sq930_cap_start_set(sdp,
				  lz24bp_capconfig[capent].cc_sizeid, 1,
				  sdp->sqd_quant,
				  lz24bp_capconfig[capent].cc_bytes,
				  lz24bp_capconfig[capent].cc_isjpeg
				  ? lz24bp_capconfig[capent].cc_jpegbytes
				  : NULL);

	if (res) {
		sq_err(sdp, "cap_start_set failed: %d", res);
		return res;
	}

	res = lz24bp_exposure(sdp, sdp->sqd_exposure, sdp->sqd_powerline);
	if (res) {
		sq_err(sdp, "lz24bp_exposure failed: %d", res);
		return res;
	}

	return 0;
}

const struct sq930_sensor sq930_sensor_lz24bp = {
	.sqs_name = "Sharp LZ24BP CCD",
	.sqs_cmos = 0,
	.sqs_probe = lz24bp_probe,
	.sqs_init = lz24bp_init,
	.sqs_cap_start = lz24bp_cap_start,
};
